function results=com_ieee8023_93a(varargin)
% This is NOT an official IEEE document.
%% Implementation example of annex 93A IEEE Std 802.3
%   http://www.ieee802.org/3/ck/public/adhoc/index.html
% result=com_ieee8023_93a(config_file, num_fext, num_next [, <s4p files>])
% - config_file: xls, xls, mat file which contains configuration settings
% - num_fext: number of FEXT s4p files in the listfigure(300+package_testcase_i);
% - num_next: number of NEXT s4p files  in the list
% - <s4p_files>: (1+num_fext+num_nefxt) file names. If not supplied, program
%   will ask for each of the files interactively.opupu
%
% This program is intended for the development of standard specifications
% and reflects activity of IEEE P802.3bj, .3by, .3bm, .3bs, .3cd, .3ck
% found in Annex 93A IEEE Std 802.3 and project =updates
% original proposal for COM may be found at
% http://www.ieee802.org/3/bj/public/jul12/mellitz_01_0712.pdf in July 2012
% from the following authors and affiliations in 2012.
%      Richard Mellitz, Intel Corporation
%      Charles Moore, Avago Technologies
%      Mike Dudek, QLogic Corporation
%      Mike Peng Li, Altera Corporation
%      Adee Ran, Intel Corporation
%
% Some of the authors and Contributors:
%   Adee Ran
%   Richard Mellitz
%   Yasuo Hidaka
%   John Ewen
%   Bill Kirkland
%   Adam Gregory
%   Howard Heck
%   Jingbo Li
%   Adam Healey
zzz_list_of_changes()

%% Opening Preface
% acquire parsing command string and set up OP control structure. Then read in files
close(findall(0, 'tag', 'TMWWaitbar', '-or', 'tag', 'COM'));
try %  version number at end of call string
    cmdfile=mfilename;
    hindx=strfind(mfilename,'_');
    ver=cmdfile(hindx(end)+1:end);
    output_args.code_revision = [ver(1), '.',ver(2:end)];
catch
    output_args.code_revision ='';
end
teststr='';
OP.TESTING=0;
if OP.TESTING == 1 % set to 1 or pre release
    teststr='testing';
    testmsg=sprintf('Evaluation Copy: COM%s%s\n',output_args.code_revision,teststr);
    htest = msgbox(testmsg);
    set(htest,'Color','y', 'tag', 'COM');movegui(htest,'northeast');
end
display('This is NOT an official IEEE document.')
fprintf('Revision:<strong> %s%s </strong>This is a computation example for exploring COM and ERL  \n',output_args.code_revision,teststr)
disp(' for projects like IEEE P802.3bj/b/bs/cd/ck with some exploratory extensions and is not normative or official')
t0=tic;
set(0,'defaulttextinterpreter','none') % prevents subscripting character in displayed messages
% reset to tex on exit
%% file_setup
%%
% need to see what happens for version 8
if verLessThan('matlab', '7.4.1')
    error('Matlab version 7.4 or higher required')
end

%New Command Line parser
keywords={'Legacy' 'TD'};
Remember_keyword='Legacy';
OP.TDMODE=false;
OP.GET_FD=true;
config_file='';
num_fext=[];
num_next=[];
if ~isempty(varargin)
    if ~ischar(varargin{1})
        error('First input must be a string');
    end
    keyword_idx=find(strcmpi(keywords,varargin{1}));
    if isempty(keyword_idx)
        %Legacy Mode
        [config_file,varargin]=varargin_extractor(varargin{:});
        [num_fext,varargin]=varargin_extractor(varargin{:});
        [num_next,varargin]=varargin_extractor(varargin{:});
    else
        %Keyword Mode
        my_keyword=varargin{1};
        Remember_keyword=my_keyword;
        varargin(1)=[];
        switch my_keyword
            
            case 'Legacy'
                [config_file,varargin]=varargin_extractor(varargin{:});
                [num_fext,varargin]=varargin_extractor(varargin{:});
                [num_next,varargin]=varargin_extractor(varargin{:});
            case 'TD'
                OP.TDMODE=true;
                OP.GET_FD=false;
                [config_file,varargin]=varargin_extractor(varargin{:});
                [num_fext,varargin]=varargin_extractor(varargin{:});
                [num_next,varargin]=varargin_extractor(varargin{:});
        end
    end
end


% get the first 3 arguments and allow for interactive input.
if isempty(config_file)
    config_file=input('Enter config XLS file or return will just pop a window to ask for the XLS file]:  ','s');
    if isempty(config_file)
        [config_file, config_file_path] = uigetfile([{ '*.xls;*.xlsx'} ; {'*.mat'}],'INPUT CONFIG FILE .xls');
    else
        [config_file_path,cname,cext]=fileparts(config_file);
        config_file=[cname cext];
    end
    if config_file==0
        % cancel - exit gracefully
        return;
    end
    config_file = fullfile(config_file_path, config_file);
end
[param, OP] = read_ParamConfigFile(config_file,OP);
if isempty(num_fext)
    if OP.RX_CALIBRATION
        num_fext=1;
        display('First prompt is for the measured test thru channel and following prompt is for Rx noise path channel')
    else
        if param.tfx(1) == -1 && OP.ERL_ONLY && OP.ERL == 2 % OP.ERL=2 is for package ERL, param.tfx = 1 means fixture time not set and need to be determinind for test fixture channel
            num_fext=1;
            display('First prompt is for the s2p measured data following prompt is for s4p of of the test fixtrure channel')
        elseif ~OP.ERL_ONLY
            num_fext=input('How many FEXT channels are to be entered? [return means no FEXT] ');
        else
            num_fext=0;
        end
    end
    if isempty(num_fext)==1, num_fext=0; end
end
if isempty(num_next)
    if param.tfx(1) == -1 && OP.ERL_ONLY && OP.ERL == 2 % OP.ERL=2 is for package ERL, param.tfx = 1 means fixture time not set and need to be determinind for test fixture channel
        num_next=0;
    else
        if OP.RX_CALIBRATION
            num_next=0;
        elseif ~OP.ERL_ONLY
            num_next=input('How many NEXT channels are to be entered? [return means no NEXT] ');
        else
            num_next=0;
        end
    end
    if isempty(num_next)==1, num_next=0; end
end
% Allow string inputs for running compiled version from OS command-line
if ischar(num_fext), num_fext=str2double(num_fext); end
if ischar(num_next), num_next=str2double(num_next); end
xtk=num_fext+num_next; % total number of crosstalk aggressors
param.num_next=num_next;
param.num_fext=num_fext;
param.num_s4p_files=num_fext+num_next+1;
% checking for data when running for rx compliance BBN calibration
if OP.RX_CALIBRATION == 1
    if num_fext ~=1
        h = msgbox('One and only noise path channel is required'); set(h,'Color',[1 .85 0]);
        movegui(h,'northwest')
        if OP.DEBUG ~= 1
            return
        end
    end
    h = msgbox('Please make sure the measured "sigma_RJ", A_DD, and SNR_TX" fields in the config xls file have been modified from the Tx measurement. '); set(h,'Color',[0 1 1]);
    movegui(h,'southeast')
end

if   param.tfx(1) == -1 && OP.ERL_ONLY && OP.ERL == 2
    if num_fext ~=1
        h = msgbox('One and only test channel is required'); set(h,'Color',[1 .85 0]);
        movegui(h,'northwest')
        if OP.DEBUG ~= 1
            return
        end
    end
    h = msgbox('The test fixture file is use to gate measurements '); set(h,'Color',[0 1 1]);
    movegui(h,'southeast')
end


% create result directory if needed
if ~exist(OP.RESULT_DIR,'dir'); mkdir(OP.RESULT_DIR); end
% allow finite impulse response input rather that s-parameters with
% OP.EXTERNAL = true. However the use_external_IR function is not provided
if ~isempty(varargin) % process case where file names are passed in function call
    if strfind(upper(char(varargin(1))),'EXTERNAL_IR') ~= 0
        error('External IR mode is no longer supported');
        %OP.EXTERNAL = true;
        %OP.GET_FD = 0;
        %ir1a= varargin(2);
        %ex_var = varargin(3);
        %[chdata OP param ]  = use_external_IR(param, OP ,num_fext,num_next,0,ir1a,ex_var);
    else
        if OP.TDMODE
            OP.GET_FD=false;
        end
        if length(varargin) < xtk +1 % check that number of varargin arguments passed is at least number of crosstalk files+1 (thru)
            error('files must include next + fext + a thru');
        end
        %% eveluate any extra arguments as possible modifications of parameters
        extra_args = varargin(xtk+2:end);
        for k=1:2:floor(length(extra_args)/2)*2
            try
                orig_value_is_str = 1;
                orig_value=eval(extra_args{k});
                if ~ischar(orig_value)
                    orig_value_is_str = 0;
                    orig_value=mat2str(orig_value);
                end
            catch eval_err
                if isequal(eval_err.identifier, 'MATLAB:nonExistentField')
                    % trying to modify a nonexistent parameter - probably a
                    % typo. save the user from his error.
                    error('COM:BadExtraParameter', 'Attempted override of a non-existing parameter %s.', extra_args{k});
                else
                    % unexpected condition
                    rethrow(eval_err);
                end
            end
            try
                if orig_value_is_str
                    mod_string = sprintf('%s = ''%s'';', extra_args{k}, extra_args{k+1});
                else
                    mod_string = sprintf('%s = %s;', extra_args{k}, extra_args{k+1});
                end
                eval(mod_string);
                fname=['mod_str' num2str(k)];
                % begin yasuo patch 2/11/2018
                % output_args.(fname)=mod_string;
                % If mod_string contains a comma, enclose it by double quotes to avoid misaligned column in the CSV output.
                
                % re-patch yasuo 3/18/2019
                % v2.56 if contains(mod_string,',')
                % v2.57 if isempty(strfind(mod_string,','))
                % Here, if-condition was inverted by the change of function from 'contains()' to 'isempty()'.
                % So, it is changed back by adding an '~' operator.
                % if isempty(strfind(mod_string,','))
                if ~isempty(strfind(mod_string,','))
                    output_args.(fname)=['"' mod_string '"'];
                else
                    output_args.(fname)=mod_string;
                end
                fprintf('Applied parameter modification: %s (override %s)\n', mod_string, orig_value);
            catch eval_err
                error(eval_err.identifier, 'Error evaluating "%s".', mod_string);
            end
        end
    end
end
%% Parameters computationally defined by values from the settings files
param.ui=1/param.fb;
param.sample_dt = param.ui/param.samples_per_ui;
param.sigma_X=sqrt( (param.levels^2-1)/ (3*(param.levels-1)^2) );
%% size adjust vector parameters which may be entered as one element
if numel(param.C_pkg_board)==1, param.C_pkg_board = param.C_pkg_board*[1 1]; end
if numel(param.C_diepad)==1, param.C_diepad = param.C_diepad*[1 1]; end
% [ahealey] Also check new termination model parameters.
if numel(param.L_comp) == 1, param.L_comp = param.L_comp*[1, 1]; end
if numel(param.C_bump) == 1, param.C_comp = param.C_bump*[1, 1]; end
% [ahealey] End of modifications.
if numel(param.tfx)==1, param.tfx = param.tfx*[1 1]; end
if numel(param.C_v)==1, param.C_v = param.C_v*[1 1]; end
if numel(param.C_0)==1, param.C_0 = param.C_0*[1 1]; end
if numel(param.C_1)==1, param.C_1 = param.C_1*[1 1]; end
[ncases, mele]=size(param.z_p_tx_cases); % need find the number of test cases RIM 01-08-20
if numel(param.AC_CM_RMS)==1, param.AC_CM_RMS = param.AC_CM_RMS*ones(1,ncases); end % RIM 12-23-20 expand ac_cm_rms per package
if numel(param.R_diepad)==1, param.R_diepad = param.R_diepad*[1 1]; end
if numel(param.CTLE_fp1)==1, param.CTLE_fp1 = param.CTLE_fp1*ones(size(param.ctle_gdc_values)); end
if numel(param.CTLE_fp2)==1, param.CTLE_fp2 = param.CTLE_fp2*ones(size(param.ctle_gdc_values)); end
if numel(param.CTLE_fz)==1, param.CTLE_fz = param.CTLE_fz*ones(size(param.ctle_gdc_values)); end
if numel(param.f_HP)==1, param.f_HP = param.f_HP*ones(size(param.g_DC_HP_values)); end
if numel(param.f_HP_Z)==1, param.f_HP_Z = param.f_HP_Z*ones(size(param.ctle_gdc_values)); end
if numel(param.f_HP_P)==1, param.f_HP_P = param.CTLE_fz*ones(size(param.ctle_gdc_values)); end
% need to adjust to number of cases case
if numel(param.pkg_Z_c)==1, param.pkg_Z_c = param.pkg_Z_c*[1 1]; end
if numel(param.brd_Z_c)==1, param.brd_Z_c = param.brd_Z_c*[1 1]; end
pkg_sel_vec=ones(1,max(OP.pkg_len_select));
if OP.WC_PORTZ
    if numel(param.a_thru)==1,param.a_thru=param.a_thru*[1 1];end
    if numel(param.a_fext)==1,param.a_fext=param.a_fext*[1 1];end
    if numel(param.a_next)==1,param.a_next=param.a_next*[1 1];end
    if numel(param.SNDR)==1,param.SNDR=param.SNDR*[1 1];end
else
    if numel(param.a_thru)==1,param.a_thru=param.a_thru*pkg_sel_vec;end
    if numel(param.a_fext)==1,param.a_fext=param.a_fext*pkg_sel_vec;end
    if numel(param.a_next)==1,param.a_next=param.a_next*pkg_sel_vec;end
    if numel(param.SNDR)==1,param.SNDR=param.SNDR*pkg_sel_vec;end
end

if OP.TDMODE
    OP.FIXTURE_CALIBRATION= 0;
    [chdata, param] = get_TD_files(param, OP, num_fext, num_next, varargin);
else
    OP.FIXTURE_CALIBRATION= 0;
    [chdata, param] = get_s4p_files(param, OP, num_fext, num_next, varargin);
end

OP.SAVE_CMD_STR=1;
if OP.SAVE_CMD_STR
    %cmd_str = save_cmd_line( config_file, chdata, num_fext,num_next,mfilename );
    cmd_str = save_cmd_line([Remember_keyword ''',''' config_file], chdata, num_fext,num_next,mfilename );
    setappdata(0,'cmd_str',cmd_str);
end
%% from here on, multiple package test cases are run. results will be saved separately.
results = cell(size(OP.pkg_len_select));
COM = inf;
min_COM=inf; % reset COM prior to calibration
% min_VEO = inf;
min_VEO_mV = inf;
max_VEC_dB = -inf;
threshold_DER=inf;
% begin yasuo patch 3/18/2019
threshold_DER_max = 0;	% reset worst DER
% end yasuo patch
sigma_bn=0;
DO_ONCE=true;
low_COM_found = 0;
% at this point only the impulse responses are needed. However vestiges of FD may be intermingled
while (OP.RX_CALIBRATION==1 || DO_ONCE==true)
    if ~DO_ONCE
        if abs(min_COM - param.pass_threshold)<0.1 || (sigma_bn==0 && min_COM < param.pass_threshold)
            break;
        elseif min_COM > param.pass_threshold
            % increase noise level linearly until low COM found; then perform binary search.
            if low_COM_found
                if OP.sigma_bn_STEP>0 % previous increase too small
                    OP.sigma_bn_STEP = OP.sigma_bn_STEP/2; % gearshift
                else % previously decrease too large
                    OP.sigma_bn_STEP = -OP.sigma_bn_STEP/2; % gearshift and change direction
                end
            end
        else % binary search
            low_COM_found=1;
            if OP.sigma_bn_STEP>0 % previous increase too large
                OP.sigma_bn_STEP = -OP.sigma_bn_STEP/2; % gearshift and change direction
            else % previously decrease too small
                OP.sigma_bn_STEP = OP.sigma_bn_STEP/2; % gearshift
            end
        end
        min_COM = inf; % ignore previous iterations
        min_VEO_mV = inf;
        max_VEC_dB = -inf;
        sigma_bn = sigma_bn + OP.sigma_bn_STEP;
    end
    msgctr=1;
    for package_testcase_i = 1:length(OP.pkg_len_select)
        package_testcase=OP.pkg_len_select(package_testcase_i);
        param.Pkg_len_TX = param.z_p_tx_cases(package_testcase,:);
        param.Pkg_len_NEXT = param.z_p_next_cases(package_testcase,:);
        param.Pkg_len_FEXT = param.z_p_fext_cases(package_testcase,:);
        param.Pkg_len_RX = param.z_p_rx_cases(package_testcase,:);
        param.AC_CM_RMS_TX= param.AC_CM_RMS(package_testcase);
        %         ki=package_testcase;
        % %         param.Pkg_Zc=[ param.pkg_Z_c(ki,1); param.pkg_Z_c(ki,2) ];
        %         param.Pkg_Zc=[ param.pkg_Z_c(ki,:) ];
        param.Pkg_Zc= param.pkg_Z_c;
        [cmele,centries] = size(param.Pkg_Zc);
        [mele, ncases] = size(param.Pkg_len_TX);
        if cmele ~=1 && centries ~=2 && mele ~= 1
            param.Pkg_Zc=reshape(param.Pkg_Zc,2,4);
        end
        param.package_testcase_i = package_testcase_i;
        if OP.TDMODE
            [chdata, param ] = read_PR_files(param, OP, chdata);
            [chdata, param, SDDch, SDDp2p ] = TD_FD_fillin(param, OP, chdata); % fill in fd data to keep rest of SW happy
        else
            %fill in chada with s-parameters
            [chdata, param, SDDch, SDDp2p ] = read_s4p_files(param, OP, chdata);
            [chdata, param] = process_sxp(param, OP, chdata, SDDch);
        end
        output_args.RL.f=chdata(1).faxis;  % RIM 07/19/2019 only use the first index
        output_args.RL.rl1=chdata(1).sdd11_raw; % RIM 07/19/2019 only use the first index
        if isfield(chdata(1),'sdd22_raw')% RIM 10/15/2019 only use the first index
            output_args.RL.r22=chdata(1).sdd22_raw; % RIM 07/19/2019 only use the first index
        end
        if isfield(chdata(1),'TX_RL')% RIM 10/09/2020 report Tx RL with RD
            output_args.RL.TXRL=chdata(1).TX_RL; %R IM 10/09/2020 report Tx RL with RD
        end
        if param.FLAG.S2P, OP.ERL_ONLY =1;end
        if OP.TDR && package_testcase_i == 1
            output_args.Z11est=chdata(1).TDR11.avgZport;
            if ~param.FLAG.S2P, output_args.Z22est=chdata(1).TDR22.avgZport;end
        end
        if OP.ERL && package_testcase_i == 1
            output_args.ERL11=chdata(1).TDR11.ERL;
            if ~param.FLAG.S2P,output_args.ERL22=chdata(1).TDR22.ERL;end
            %                 output_args.ERL11RMS=chdata(1).TDR11.ERLRMS;
            %                  if ~param.FLAG.S2P,output_args.ERL22RMS=chdata(1).TDR22.ERLRMS;
        end
        if OP.ERL
            
            if OP.TDR_W_TXPKG
                min_ERL=output_args.ERL22;
                ERL= [ nan output_args.ERL22 ];
            else
                if ~isfield(output_args,'ERL22')
                    min_ERL=output_args.ERL11;
                    ERL= [ output_args.ERL11 nan ];
                else
                    min_ERL=min(output_args.ERL11,output_args.ERL22);
                    ERL= [ output_args.ERL11 output_args.ERL22 ];
                end
            end
            output_args.ERL=min_ERL;
        else
            min_ERL=[];
            ERL= [];
        end
        if OP.ERL_ONLY
            if OP.BREAD_CRUMBS
                output_args.OP=OP;
                output_args.param=param;
                output_args.chdata=chdata;
                fom_result.ran=0;
            end
            output_args.config_file = config_file;
            output_args.Z_t=param.Z_t;
            fileset_str=str2csv({chdata(1).base});
            output_args.file_names=sprintf('"%s"', fileset_str);
            results = cell(1);
            results{1} = output_args;
            rt=toc(t0);
            output_args.rtmin=rt/60;
            if 1
                fprintf('run time = %g min\n',output_args.rtmin)
            end
            if OP.DISPLAY_WINDOW
                savefigs(param, OP);
            end
            if OP.CSV_REPORT ==1
                items = fieldnames(output_args);
                item_value_strings = cell(size(items));
                for field_id=1:length(items)
                    field_name=items{field_id};
                    field_value=output_args.(field_name);
                    if isstruct(output_args.(field_name))
                        field_value='struct';
                    end
                    if ischar(field_value)
                        item_value_strings{field_id}=field_value;
                    elseif numel(field_value)==1
                        item_value_strings{field_id}=num2str(field_value);
                    else
                        item_value_strings{field_id}=sprintf('"%s"', mat2str(field_value));
                    end
                end
                
                header_string = str2csv(items);
                data_string = str2csv(item_value_strings);
                fid = fopen(sprintf('%s%s_case%d_results.csv', OP.RESULT_DIR, chdata(1).base, package_testcase_i),'w');
                fprintf(fid,'%s\n', header_string);
                fprintf(fid,'%s\n', data_string);
                fclose(fid);
            end
            
            break;
        end
        number_of_s4p_files=length(chdata);
        if OP.WC_PORTZ
            A_thru = param.a_thru(param.Tx_rd_sel);
            A_fext = param.a_fext(param.Tx_rd_sel);
            if num_next ~= 0
                A_next = param.a_next(param.Tx_rd_sel);
            end
            param.SNR_TX=param.SNDR(param.Tx_rd_sel);
        else
            A_thru = param.a_thru(package_testcase);
            A_fext = param.a_fext(package_testcase);
            A_next = param.a_next(package_testcase);
            param.SNR_TX=param.SNDR(package_testcase);
        end
        %% FD processing s-parameter
        % at this point sdd21 responses and faxis (frequency) array are defined
        ICN=0;
        for i=1:number_of_s4p_files
            if isequal(chdata(i).type, 'THRU')
                chdata(i).A=A_thru;
                chdata(i).Aicn=A_thru;
            elseif isequal(chdata(i).type, 'FEXT')
                chdata(i).A=A_fext;
                chdata(i).Aicn=param.a_icn_fext;
            elseif isequal(chdata(i).type, 'NEXT')
                chdata(i).A=A_next;
                chdata(i).Aicn=param.a_icn_next;
            end
        end
        if OP.TDMODE
            for i=1:number_of_s4p_files           % freq delta for integration
                chdata(i).delta_f=chdata(i).faxis(11)-chdata(i).faxis(10);
            end
        end
        if OP.GET_FD && DO_ONCE
            f2=param.f2;
            f1=param.f1;
            MDFEXT_ICN=0; MDNEXT_ICN=0;
            for i=1:number_of_s4p_files
                if i == 2
                    PSXT(1:length(chdata(i).sdd21f))=0;
                    MDFEXT(1:length(chdata(i).sdd21f))=0;
                    MDNEXT(1:length(chdata(i).sdd21f))=0;
                end
                a=find(chdata(i).faxis(:)>=f2,1,'first');% RIM 01-12-21
                if isempty(a)
                    f2=chdata(i).faxis(end);
                    index_f2=length(chdata(i).faxis);
                else
                    index_f2=a(1);
                end
                b=find(chdata(i).faxis(:)<=f1,1,'last');% RIM 01-12-21
                if isempty(b)
                    f1=chdata(i).faxis(1);
                    index_f1=1;
                else
                    index_f1=b(1);
                end
                % R is the frequency dependent parameter for the sinc function use in the
                % PWF for ICN
                temp_angle=(param.samples_per_ui*param.sample_dt)*pi.*chdata(i).faxis;
                if(chdata(i).faxis(1)==0)
                    temp_angle(1)=1e-20;% we don't want to divide by zero
                end
                SINC = sin(temp_angle)./temp_angle;
                PWF_data=SINC.^2;
                PWF_trf=(1+(chdata(i).faxis/chdata(i).ftr).^4).^-1;
                %// bw1=2.613126; bw2=3.4142136; bw3=2.613126;
                fr=param.f_r*param.fb;
                PWF_rx=(1+(chdata(i).faxis/fr).^8).^-1;
                PWF_highpass=1;
                %% Equation 93A-57 %%
                PWF=PWF_data.*PWF_trf.*PWF_rx.*PWF_highpass; % power weight function
                % freq delta for integration
                chdata(i).delta_f=chdata(i).faxis(11)-chdata(i).faxis(10);
                % from ba spec, this is basically ICN
                faxis_GHz = chdata(i).faxis/1e9;
                if isequal(chdata(i).type, 'THRU')
                    [ILD chdata(i).fit] = get_ILN(chdata(i).sdd21f(index_f1:index_f2), chdata(i).faxis(index_f1:index_f2));
                    %nq_indx=find(round(chdata(i).faxis,4,'significant') >= round(1/param.ui/2,4,'significant'),1, 'first');
                    %finding nq_indx without using "round":
                    [tmp,nq_indx]=min(abs(chdata(i).faxis-1/param.ui/2));
                    nq_indx=min(nq_indx,length(chdata(i).fit));% freq array may not contain Fb/2 exactly RIM 06-09-21
                    chdata(i).fit_ILatNq=-chdata(i).fit(nq_indx);
                    chdata(i).ILatNq=-20*log10(abs(chdata(i).sdd21f(nq_indx)));
                    fit_loss=chdata(i).fit_ILatNq;
                    if OP.include_pcb
                        cable_loss=-20*log10(abs(chdata(i).sdd21_orig(nq_indx)));
                    end
                    total_loss=-20*log10(abs(SDDp2p(i).Parameters(2,1,nq_indx)));
                    Nq_loss=chdata(i).ILatNq;
                    [RIL_struct]= capture_RIL_RILN(chdata);
                    FOM_RILN=sqrt(chdata(i).delta_f/(param.f2-param.f1)*sum( PWF(index_f1:index_f2-1).*RIL_struct.RILN_dB(index_f1:index_f2-1)'.^2));
                    %% Equation 93A-56 %%
                    FOM_ILD=sqrt(chdata(i).delta_f/(param.f2-param.f1)*sum( PWF(index_f1:index_f2).*ILD.^2));
                    if OP.DEBUG
                        if OP.DISPLAY_WINDOW
                            
                            figure(300+package_testcase_i);
                            set(gcf,'Tag','COM')
                            screen_size=get(0,'ScreenSize');
                            pos = get(gcf, 'OuterPosition');
                            set(gcf, 'Name', [sprintf('%.3gdB IL Channel: ',Nq_loss) 'Raw frequency-domain data'], 'OuterPosition', ...
                                screen_size([3 4 3 4]).*[0 1 0 0] + pos([3 4 3 4]).*[0 -2 1 2] ...
                                - (package_testcase_i-1)*[0 20 0 0]);
                            subplot(3,1,1)
                            title('Losses')
                            plot(faxis_GHz, 20*log10(abs(squeeze(chdata(i).sdd21f))), 'b', 'LineWidth', 3, 'Disp','IL w/o packages')
                            hold on
                            plot(faxis_GHz, 20*log10(abs(squeeze(SDDp2p(i).Parameters(2,1,:)))), 'b--', 'Disp','IL with pkgs')
                            ylim(get(gca, 'ylim'));
                            %                             plot(faxis_GHz(index_f1:index_f2),chdata(i).fit(index_f1:index_f2),'g','Disp','IL fit')
                            plot(faxis_GHz(index_f1:index_f2),chdata(i).fit,'g','Disp','IL fit')
                            plot(faxis_GHz, 20*log10(abs(squeeze(chdata(i).sdd11))),'c','Disp','RL11')
                            plot(faxis_GHz, 20*log10(abs(squeeze(chdata(i).sdd22))),'m','Disp','RL22')
                            subplot(3,1,3)
                            plot(faxis_GHz(index_f1:index_f2), ILD,'Disp','ILD')
                            if OP.PLOT_CM
                                if package_testcase_i ==1
                                    h350=figure(350);set(gcf,'Tag','COM')
                                    screen_size=get(0,'ScreenSize');
                                    pos = get(gcf, 'OuterPosition');
                                    set(gcf, 'OuterPosition',screen_size([3 4 3 4]).*[1  1 0 0] + pos([3 4 3 4]).*[-2 -2 2 1])
                                    movegui(gcf,'center');
                                    htabgroup350 = uitabgroup(h350);
                                    htab1 = uitab(htabgroup350, 'Title', 'CM Through Losses');
                                    hax1 = axes('Parent', htab1);
                                    set(h350,'CurrentAxes',hax1)
                                    hold on
                                    set(gcf,'Tag','COM')
                                    screen_size=get(0,'ScreenSize');
                                    pos = get(gcf, 'OuterPosition');
                                    title('IL & CM Losses')
                                    base=strrep(chdata(i).base,'_',' ');
                                    plot(faxis_GHz, 20*log10(abs(squeeze(chdata(i).sdd21f))), 'b', 'LineWidth', 3, 'Disp',  [ 'sdd21(IL) TP0-TP5 ' base])
                                    plot(faxis_GHz, 20*log10(abs(squeeze(chdata(i).sdc21_raw))),   'LineWidth', 2, 'Disp',[ 'sdc21 TP0-TP5 '  base])
                                    ylabel('dB')
                                    xlabel('GHz')
                                    legend show
                                    legend('Location','eastoutside')
                                    hold on
                                    grid on
                                    if number_of_s4p_files > 1
                                        htab2 = uitab(htabgroup350, 'Title', 'CM Crosstalk Losses');
                                        hax2 = axes('Parent', htab2);
                                        htab3 = uitab(htabgroup350, 'Title', 'Crosstalk Losses');
                                        hax3 = axes('Parent', htab3);
                                    end
                                    
                                end
                            end
                        else
                            display(['Insertion Loss at Nyquist = ', num2str(chdata(i).ILatNq)])
                        end
                    end
                else % NEXT or FEXT
                    if isequal(chdata(i).type, 'FEXT')
                        MDFEXT=sqrt(abs(chdata(i).sdd21f).^2+MDFEXT.^2); % power sum xtk
                        MDFEXT_ICN=sqrt(2*chdata(i).delta_f/param.f2*sum( chdata(i).Aicn^2*PWF(index_f1:index_f2).*abs(MDFEXT(index_f1:index_f2)).^2)); %eq 46
                    elseif isequal(chdata(i).type, 'NEXT')
                        MDNEXT=sqrt(abs(chdata(i).sdd21f).^2+MDNEXT.^2); % power sum xtk
                        MDNEXT_ICN=sqrt(2*chdata(i).delta_f/param.f2*sum( chdata(i).Aicn^2*PWF(index_f1:index_f2).*abs(MDNEXT(index_f1:index_f2)).^2)); %eq 47
                    end
                    PSXT=sqrt((abs(chdata(i).sdd21f)*chdata(i).Aicn).^2+PSXT.^2); % power sum xtk
                    ICN=sqrt(2*chdata(i).delta_f/param.f2*sum( PWF(index_f1:index_f2).*abs(PSXT(index_f1:index_f2)).^2));
                    ICN_test=norm([MDFEXT_ICN MDNEXT_ICN]);
                    if  OP.PLOT_CM && OP.DISPLAY_WINDOW
                        if package_testcase_i ==1
                            %                             htab2 = uitab(htabgroup350, 'Title', 'CM Crosstalk Losses');
                            %                             hax2 = axes('Parent', htab2);
                            set(h350,'CurrentAxes',hax2)
                            hold on
                            title('CM Losses')
                            base=strrep(chdata(i).base,'_',' ');
                            plot(faxis_GHz, 20*log10(abs(squeeze(chdata(i).sdc21_raw))), 'LineWidth', 2, 'Disp',[ 'sdc21 TP0-TP5 ' base ])
                            legend('Location','eastoutside')
                            hold on
                            grid on
                            set(h350,'CurrentAxes',hax3)
                            plot(faxis_GHz, 20*log10(abs(squeeze(chdata(i).sdd21_raw))), 'LineWidth', 2, 'Disp',[ 'sdd21 TP0-TP5 ' base ])
                            legend('Location','eastoutside')
                            hold on
                            grid on
                        end
                    end
                end
            end  % for loop
            ICN_test=norm([MDFEXT_ICN MDNEXT_ICN]);
            if OP.DEBUG && OP.DISPLAY_WINDOW
                figure(300+package_testcase_i);set(gcf,'Tag','COM');
                if number_of_s4p_files > 1
                    %                     scale=scale=A_fext/A_thru;% added for crosstalk scaling
                    scale=1/chdata(2).Aicn; %chdata(i).sdd21f not scalled
                    subplot(3,1,1)
                    hold on
                    plot(faxis_GHz, 20*log10(abs(PSXT*scale)),'r','Disp','PSXTK')
                    icrxi=find(chdata(i).faxis >=param.fb/2,1,'first');
                    subplot(3,1,2)
                    grid on
                    ILtemp=20*log10(abs(chdata(1).sdd21f));
                    IL4ICR=interp1(chdata(1).faxis,ILtemp,chdata(i).faxis);
                    %                     scale=scale=A_fext/A_thru;% added for crosstalk scaling
                    scale=1/chdata(2).Aicn; % chdata(i).sdd21f not scalled
                    ICR=-20*log10(abs(PSXT*scale))+IL4ICR;
                    semilogx(faxis_GHz, ICR,'Disp', 'ICR')
                    hold on
                    stem(faxis_GHz(icrxi), ICR(icrxi),'g', 'disp', 'f_{Baud}/2')
                end
                subplot(3,1,1)
                title([param.base ' Losses']); ylabel('dB'); xlabel('GHz')
                grid on;  legend show
                subplot(3,1,2)
                title([param.base ' ICR']); ylabel('dB'); xlabel('GHz')
                ylim([0 80])
                xlim([.1 100])
                grid on; %legend show
                subplot(3,1,3)
                title([param.base ' ILD']); ylabel('dB'); xlabel('GHz')
                ylim([-3 3])
                grid on; legend show
            end
        end % get_FD && DO_ONCE
        if DO_ONCE
            % get impulse responses which in interim step between equation for X(f) and
            % H^(k)(t) without TX FFE or CTLE. These will we added later.
            for i=1:number_of_s4p_files
                if ~OP.TDMODE
                    if OP.INCLUDE_FILTER % apply RX filter
                        %% Equation 93A-20 %%
                        %                     H_r = 1./polyval([1 2.613126 3.414214 2.613126 1], 1i*chdata(i).faxis./(param.f_r*param.fb));
                        f=chdata(i).faxis;
                        %
                        param.fb_BW_cutoff=param.f_r;
                        factor_3db=0.473037;
                        param.fb_BT_cutof=factor_3db*param.f_r;
                        if OP.Bessel_Thomson
                            a = bessel( param.BTorder );
                            acoef=fliplr( a );
                            H_bt =a(1)./ polyval(acoef, (1i*f./(param.fb_BT_cutoff*param.fb)));
                        else
                            H_bt=ones(1,length(f));
                        end
                        if OP.Butterworth
                            H_bw = 1./polyval([1 2.613126 3.414214 2.613126 1], 1i*f./(param.fb_BW_cutoff*param.fb));
                        else
                            H_bw=ones(1,length(f));
                        end
                        H_r=H_bw.*H_bt;
                        chdata(i).sdd21=chdata(i).sdd21.*H_r;
                        
                    end
                    %                 if param.Tukey_Window ~= 0
                    %                     L=length(chdata(i).faxis);
                    %                     H_tw= Tukey_Window(chdata(i).faxis ,param);
                    %                     chdata(i).sdd21=chdata(i).sdd21.*H_tw;
                    %                 end
                    % if i==1 and else are the same. just do it
                    
                    [chdata(i).uneq_imp_response, ...
                        chdata(i).t, ...
                        chdata(i).causality_correction_dB, ...
                        chdata(i).truncation_dB] = s21_to_impulse_DC(chdata(i).sdd21 ,chdata(i).faxis, param.sample_dt, OP) ;
                    if ~OP.RX_CALIBRATION || i==1 % DC (common to differentail model is not good used for RX_Calibrataion channel
                        chdata(i).uneq_imp_response=chdata(i).uneq_imp_response*chdata(i).A; % adjust IRx for amplitude
                        [chdata(i).uneq_DC_imp_response, ...
                            chdata(i).t_DC, ...
                            chdata(i).causality_correction_DC_dB, ...
                            chdata(i).truncation__DC_dB] = s21_to_impulse_DC(chdata(i).sdc21 ,chdata(i).faxis, param.sample_dt, OP) ;
                    end
                    % adjust voltage derive here once it's decided what to use
                    %------------------------------------------------------------
                    % next find Pulse response (SBR) for each channel h^(k)(t)
                    if ~OP.DISPLAY_WINDOW && i==1, fprintf('processing COM PDF '); end
                    
                    chdata(i).uneq_pulse_response=filter(ones(1, param.samples_per_ui), 1, chdata(i).uneq_imp_response);
                    chdata(i).uneq_pulse_DC_response=filter(ones(1, param.samples_per_ui), 1, chdata(i).uneq_DC_imp_response);
                end
                if OP.DEBUG
                    if OP.DISPLAY_WINDOW && ~OP.RX_CALIBRATION
                        figure(150+package_testcase_i);set(gcf,'Tag','COM');
                        screen_size=get(0,'ScreenSize');
                        pos = get(gcf, 'OuterPosition');
                        set(gcf, 'OuterPosition', ...
                            screen_size([3 4 3 4]).*[1 1 0 0] + pos([3 4 3 4]).*[-1 -1 1 1] ...
                            - (package_testcase_i-1)*[0 20 0 0]);
                        %movegui(gcf,'northeast')
                        
                        set(gcf, 'Name', sprintf('Case %d PR & PDF - %s', package_testcase_i, chdata(i).base));
                        subplot(2,1,1);  hold on; % all plots on the same axes
                        hp=plot(chdata(i).t, chdata(i).uneq_pulse_response,'Disp', chdata(i).base);
                        if ~OP.TDMODE
                            hp1=plot(chdata(i).t_DC, chdata(i).uneq_pulse_DC_response,'Disp', [ 'DC ' chdata(i).base ]) ;
                        end
                        % hide thru PR in order to show xtalk in a reasonable
                        % scale. thru is shown in another plot.
                        if isequal(chdata(i).type, 'THRU' ) || OP.RX_CALIBRATION
                            set(hp, 'visible', 'off');
                            set(get(get(hp,'Annotation'),'LegendInformation'), 'IconDisplayStyle','off');
                        end
                        title('Crosstalk and DC Conversion Pulse Responses')
                        ylabel('Volts')
                        xlabel('seconds')
                        
                        recolor_plots(gca);
                    else
                        if param.ndfe~=0
                            fprintf('%s\tUnequalized pulse peak = %.1f mV\n', chdata(i).base, 1000*max(abs(chdata(i).uneq_pulse_response)));
                        end
                    end
                    if ~OP.TDMODE
                        if OP.ENFORCE_CAUSALITY
                            fprintf('%s\tCausality correction = %.1f dB\n', chdata(i).base, chdata(i).causality_correction_dB);
                        else
                            fprintf('%s\tCausality correction = %.1f dB (not applied)\n', chdata(i).base, chdata(i).causality_correction_dB);
                        end
                        fprintf('%s\tTruncation ratio = %.1f dB\n', chdata(i).base, chdata(i).truncation_dB);
                        
                    end
                end
            end
        end %DO_ONCE
        
        %% Determine equalization settings
        %---------------------
        if param.T_O == 0
            fom_result = optimize_fom(OP,param, chdata, sigma_bn,0);
        else
            if param.Min_VEO_Test == 0
                fom_result = optimize_fom(OP,param, chdata, sigma_bn,0);
            else
                fom_result = optimize_fom(OP,param, chdata, sigma_bn,1);
            end
        end
        %%
        %--------------------
        A_s=abs(fom_result.A_s); % this is the "s" in SNR (PAM4 gain in handled in last sections
        param.A_s=A_s;
        param.use_bmax=fom_result.best_bmax.';
        %AJG021820
        param.use_bmin=fom_result.best_bmin.';
        %% Recommended Delta_y no larger than As/1000 or 0.01 mV
        param.current_ffegain=fom_result.best_current_ffegain;
        if OP.force_pdf_bin_size
            delta_y = OP.BinSize;
        else
            delta_y = min(A_s/1000, OP.BinSize);
        end
        % the pdf for PAM4 uses the full swing SBR but assigns voltage for the PDF accordingly
        if OP.RX_CALIBRATION, number_of_s4p_files=1; end
        for i=1:number_of_s4p_files
            % get quick PDF  results but only for THRU when in Rx calibration
            if OP.INCLUDE_CTLE==1
                switch param.CTLE_type
                    case 'CL93'
                        chdata(i).eq_imp_response = TD_CTLE(chdata(i).uneq_imp_response, param.fb ...
                            , param.CTLE_fz(fom_result.ctle), param.CTLE_fp1(fom_result.ctle) ...
                            , param.CTLE_fp2(fom_result.ctle), param.ctle_gdc_values(fom_result.ctle) ...
                            , param.samples_per_ui);
                        
                    case 'CL120d'
                        chdata(i).eq_imp_response = TD_CTLE(chdata(i).uneq_imp_response, param.fb ...
                            , param.CTLE_fz(fom_result.ctle), param.CTLE_fp1(fom_result.ctle) ...
                            , param.CTLE_fp2(fom_result.ctle), param.ctle_gdc_values(fom_result.ctle) ...
                            , param.samples_per_ui);
                        
                        chdata(i).eq_imp_response = TD_CTLE(chdata(i).eq_imp_response, param.fb ...
                            , param.f_HP(fom_result.best_G_high_pass), param.f_HP(fom_result.best_G_high_pass) ...
                            , 100e100 , param.g_DC_HP_values(fom_result.best_G_high_pass) ...
                            , param.samples_per_ui);
                    case 'CL120e' % z has been adjusted  for gain
                        chdata(i).eq_imp_response = TD_CTLE(chdata(i).uneq_imp_response, param.fb ...
                            , param.CTLE_fz(fom_result.ctle), param.CTLE_fp1(fom_result.ctle) ...
                            , param.CTLE_fp2(fom_result.ctle), param.ctle_gdc_values(fom_result.ctle) ...
                            , param.samples_per_ui);
                        
                        chdata(i).eq_imp_response =  TD_CTLE(chdata(i).eq_imp_response,param.fb, ...
                            param.f_HP_Z(fom_result.ctle), param.f_HP_P(fom_result.ctle),1e99, 0, param.samples_per_ui);
                end
            else
                chdata(i).eq_imp_response=chdata(i).uneq_imp_response;
            end
            chdata(i).eq_pulse_response=filter(ones(1, param.samples_per_ui), 1, chdata(i).eq_imp_response);
            if isequal(chdata(i).type, 'FEXT') || isequal(chdata(i).type, 'THRU')
                chdata(i).eq_pulse_response = FFE( fom_result.txffe ,fom_result.cur-1 , param.samples_per_ui,  chdata(i).eq_pulse_response );
            end
            if OP.RxFFE
                [ chdata(i).eq_pulse_response, C]=force(chdata(i).eq_pulse_response,param,OP,fom_result.t_s,fom_result.RxFFE);
                output_args.RxFFE=fom_result.RxFFE;
                output_args.RxFFEgain=param.current_ffegain;
            else % Yasou Hidaka 11/20/2018 help to align csv file columns
                output_args.RxFFE='';
                output_args.RxFFEgain='';
            end
            if ~OP.DISPLAY_WINDOW, fprintf('%d ', i); end
            pdf = get_pdf(chdata(i), delta_y, fom_result.t_s, param, OP) ;
            if OP.DEBUG && OP.DISPLAY_WINDOW && ~OP.RX_CALIBRATION
                figure(150+package_testcase_i);set(gcf,'Tag','COM');
                subplot(2,1,2);
                pdf0.x=pdf.x(pdf.y~=0);
                pdf0.y=pdf.y(pdf.y~=0);
                semilogy(pdf0.x, pdf0.y,'disp', chdata(i).base);
                current_ylim=ylim; ylim([param.specBER/100, current_ylim(2)]);
                hold on; title('PDF')
                recolor_plots(gca);
            end
            
            chdata(i).pdfr=pdf;
            % reporting
            a=find(cumsum(chdata(i).pdfr.y) >1e-12,1,'first');
            chdata(i).maxquickpdf=(chdata(i).pdfr.y(a));
            
        end
        if ~OP.DISPLAY_WINDOW, fprintf('\n'); end
        
        if OP.RX_CALIBRATION
            ctle_gain2 = (10.^(param.ctle_gdc_values(fom_result.ctle)/20) + 1i*chdata(2).faxis/param.CTLE_fz(fom_result.ctle)) ./ ...
                ((1+1i*chdata(2).faxis/param.CTLE_fp1(fom_result.ctle)).*(1+1i*chdata(2).faxis/param.CTLE_fp2(fom_result.ctle)));
            switch param.CTLE_type
                case 'CL93'
                    H_low2=1;
                case 'CL120d' % this clause uses two gain indexes
                    H_low2=(10.^(param.g_DC_HP_values(fom_result.best_G_high_pass)/20) +  1i*chdata(2).faxis/param.f_HP(fom_result.best_G_high_pass))./(1 + 1i*chdata(2).faxis/param.f_HP(fom_result.best_G_high_pass));
                case 'CL120e' % Z1 has been adjusted
                    H_low2=(1 +  1i*chdata(2).faxis/f_HP_P(fom_result.ctle))./(1 + 1i*chdata(2).faxis/f_HP_Z(fom_result.ctle));
            end
            H_ctf2=H_low2.*ctle_gain2;
            sigma_ne = get_sigma_noise( H_ctf2,  param, chdata, sigma_bn );
        else
            sigma_ne=0;
        end
        sigma_N = fom_result.sigma_N; % eta zero noise
        %% Equation 93A-30 %%
        sigma_TX = (param.levels-1)*A_s/param.R_LM*10^(-param.SNR_TX/20); % SNR_Tx and RLM
        
        %% Equation 93A-41 %%
        sigma_G  = norm([param.sigma_RJ*param.sigma_X*norm(fom_result.h_J), sigma_N, sigma_TX]);
        sigma_rjit= param.sigma_RJ*param.sigma_X*norm(fom_result.h_J);
        
        %% Equation 93A-42 %%
        % number of sigmas needed depends on the required BER.
        if param.Noise_Crest_Factor == 0
            ber_q = sqrt(2)*erfcinv(2*param.specBER);
        else
            ber_q=param.Noise_Crest_Factor;
        end
        gaussian_noise_pdf = normal_dist(sigma_G, ber_q, delta_y);
        % enable overriding the Q factor of the BBN instrument.
        if OP.force_BBN_Q_factor
            ne_noise_pdf = normal_dist(sigma_ne, OP.BBN_Q_factor, delta_y);
        else
            ne_noise_pdf = normal_dist(sigma_ne, ber_q, delta_y);
        end
        gaussian_noise_pdf = conv_fct(gaussian_noise_pdf, ne_noise_pdf);
        
        %% p_DD is computed using the procedure defined in 93A.1.7.1 with h(n)=A_DD*h_J(n)
        p_DD = get_pdf_from_sampled_signal(param.A_DD*fom_result.h_J, param.levels, delta_y);
        
        %% Equation 93A-43
        noise_pdf=conv_fct(gaussian_noise_pdf, p_DD);
        
        gaussian_rjitt_pdf = normal_dist(sigma_rjit, ber_q, delta_y);
        jitt_pdf=conv_fct(gaussian_rjitt_pdf, p_DD);
        
        %% Implementation of 93A.1.7.3 combination procedure
        %%  (effectively Equation 93A-44) %%
        
        % Self-Channel Interference is thru residual result
        sci_pdf = chdata(1).pdfr;
        sci_mxi=find(cumsum(sci_pdf.y)>=param.specBER, 1, 'first');
        thru_peak_interference_at_BER=abs(sci_pdf.x(sci_mxi));
        sci_msi=find(cumsum(sci_pdf.y)>=param.specBER, 1, 'first');
        sci_sigma=abs(sci_pdf.x(sci_msi)/(erfcinv(2*param.specBER)*sqrt(2)));
        if OP.RX_CALIBRATION ==0
            % Co-Channel Interference PDFs (for information only):
            % initialize to deltas
            MDNEXT_cci_pdf = d_cpdf(delta_y, 0, 1);
            MDFEXT_cci_pdf = d_cpdf(delta_y, 0, 1);
            % serially convolve FEXT/NEXT PDFs
            for k=2:number_of_s4p_files
                if isequal(chdata(k).type, 'NEXT')
                    MDNEXT_cci_pdf = conv_fct(MDNEXT_cci_pdf, chdata(k).pdfr);
                else % ... must be FEXT
                    MDFEXT_cci_pdf = conv_fct(MDFEXT_cci_pdf, chdata(k).pdfr);
                end
            end
            
            % find "peaks" of MDNEXT/MDFEXT for reporting
            mdnxi=find(cumsum(MDNEXT_cci_pdf.y)>=param.specBER, 1, 'first');
            MDNEXT_peak_interference=abs(MDNEXT_cci_pdf.x(mdnxi));
            mdfxi=find(cumsum(MDFEXT_cci_pdf.y)>=param.specBER, 1, 'first');
            MDFEXT_peak_interference=abs(MDFEXT_cci_pdf.x(mdfxi));
            
            % Combined crosstalk effect
            cci_pdf = conv_fct(MDFEXT_cci_pdf, MDNEXT_cci_pdf);
            cci_mxi=find(cumsum(cci_pdf.y)>=param.specBER, 1, 'first');
            cci_msi=find(cumsum(cci_pdf.y)>=param.specBER, 1, 'first');
            cci_sigma=abs(cci_pdf.x(cci_msi)/(erfcinv(2*param.specBER)*sqrt(2)));
            crosstalk_peak_interference_at_BER=abs(cci_pdf.x(cci_mxi));
            % combine cci and sci
            isi_and_xtalk_pdf = conv_fct(sci_pdf, cci_pdf);
        else
            % for calibration there is no cci
            isi_and_xtalk_pdf=sci_pdf;
        end
        
        mxi=find(cumsum(isi_and_xtalk_pdf.y)>=param.specBER, 1, 'first');
        peak_interference_at_BER=abs(isi_and_xtalk_pdf.x(mxi));
        
        
        %% Equation 93A-45
        combined_interference_and_noise_pdf = conv_fct(isi_and_xtalk_pdf, noise_pdf);
        
        %% Equation 93A-37
        combined_interference_and_noise_cdf=cumsum(combined_interference_and_noise_pdf.y);
        
        % begin yasuo patch 12/11/2018
        % collect sigma values to report
        % pdf2sgm() is a function to calculate sigma value from PDF
        % It is added at the end of this file code.
        % I am not sure if an equivalent function already exists.
        sgm_Ani__isi_xt_noise = pdf2sgm(combined_interference_and_noise_pdf);
        sgm_isi_xt = pdf2sgm(isi_and_xtalk_pdf);
        sgm_noise__gaussian_noise_p_DD  = pdf2sgm(noise_pdf);
        sgm_p_DD = pdf2sgm(p_DD);
        sgm_gaussian_noise = pdf2sgm(gaussian_noise_pdf);
        sgm_G = sigma_G;
        sgm_rjit = sigma_rjit;
        sgm_N = sigma_N;
        sgm_TX = sigma_TX;
        sgm_isi = pdf2sgm(sci_pdf);
        if exist('cci_pdf','var') ,sgm_xt  = pdf2sgm(cci_pdf);end
        % end yasuo patch
        
        %% The noise and interference amplitude, A_ni, is the magnitude of the value of y0
        %% that satisfies the relationship P(y0) = DER_0
        A_ni_ix=find(combined_interference_and_noise_cdf>param.specBER, 1, 'first');
        A_ni = abs(combined_interference_and_noise_pdf.x(A_ni_ix));
        
        % begin yasuo patch 3/18/2019
        % estimate DER at threshold COM
        threshold_ix=find(combined_interference_and_noise_pdf.x>-A_s/(10^(param.pass_threshold/20)),1);
        threshold_DER=combined_interference_and_noise_cdf(threshold_ix);
        threshold_DER_max = max(threshold_DER_max, threshold_DER);
        % end yasuo patch
        
        %         OP.EW=1;
        if OP.RX_CALIBRATION ==0 && OP.EW == 1
            
            Struct_Noise.sigma_N=sigma_N;
            Struct_Noise.sigma_TX=sigma_TX;
            Struct_Noise.cci_pdf=cci_pdf;
            Struct_Noise.ber_q=ber_q;
            Struct_Noise.ne_noise_pdf=ne_noise_pdf;
            [Left_EW,Right_EW,eye_contour,EH_T_C2M,EH_B_C2M]=COM_eye_width(chdata,delta_y,fom_result,param,OP,Struct_Noise,0);
            EW_UI=floor((Left_EW+Right_EW))/param.samples_for_C2M;
        else
            EW_UI=0;
        end
        
        if param.T_O ~=0
            eye_opening=EH_T_C2M-EH_B_C2M;
            A_ni=2*A_s-eye_opening;
            %eq 124E-4
            vec_arg=2*A_s/eye_opening;
            if vec_arg<eps
                vec_arg=eps;
            end
            VEC_dB = 20*log10(vec_arg);
            COM=20*log10(2*A_s/A_ni);
            VEO_mV=eye_opening*1000;
            min_COM = min(min_COM, COM);
            min_VEO_mV = min(min_VEO_mV,VEO_mV);
            max_VEC_dB = max(max_VEC_dB, VEC_dB);
        else
            VEO_mV = 1000*(A_s-A_ni)*2;
            vec_arg=(A_s-A_ni)/A_s;
            if vec_arg<eps
                vec_arg=eps;
            end
            VEC_dB = -20*log10(vec_arg);
            COM=20*log10(A_s/A_ni);
            min_COM = min(min_COM, COM);
            min_VEO_mV = min(min_VEO_mV,VEO_mV);
            max_VEC_dB = max(max_VEC_dB, VEC_dB);
        end
        
        
        %% Calculation of error propagation and burst probability
        if OP.nburst>0
            % an error burst of length N will cause each of the first N taps tap to mis-correct and create a PAM (2 or
            % 4) noise term - depending on the N'th previous symbol, with double the tap voltage. From this we calculate
            % the probability of staying in error state, i.e. burst of length N+1.
            
            % initialize loop with uncorrelated noise and BER
            error_propagation_noise_pdf{1}=combined_interference_and_noise_pdf; %#ok<AGROW> % PDF for burst of length 1 is the uncorrelated PDF
            
            % Assume an error will occur if the noise excceds the available signal
            % reduced by some dB. reduction is by COM threshold minus Error
            % propagation margin (a positive EP margin reduces uncorrelated error probability
            % below target BER).
            error_threshold =  A_s./10^((param.pass_threshold-OP.COM_EP_margin)/20);
            % Find the probability of this event by integration of the PDF. Use 1e-20 as a floor probabilty if noise PDF isn't wide enough.
            x_error_propagation = find(error_propagation_noise_pdf{1}.x >= error_threshold, 1, 'first');
            if isempty(x_error_propagation)
                p_error_propagation(1) = 1e-20;
            else
                p_error_propagation(1) = sum(error_propagation_noise_pdf{1}.y(x_error_propagation:end));  % uncorrelated BER
            end
            
            sorted_abs_dfe_taps = sort(abs(result.DFE_taps), 'descend');
            for k=2:min(param.ndfe, OP.nburst)
                % (arrays kept to allow tracking during development, though not really needed)
                if OP.use_simple_EP_model
                    post_error_dfe_noise_pdf{k} = get_pdf_from_sampled_signal( 2*A_s*max(sorted_abs_dfe_taps), param.levels, delta_y ); %#ok<AGROW>
                    error_propagation_noise_pdf{k} = conv_fct(error_propagation_noise_pdf{1}, post_error_dfe_noise_pdf{k}); %#ok<AGROW>
                else
                    post_error_dfe_noise_pdf{k} = get_pdf_from_sampled_signal( 2*A_s*sorted_abs_dfe_taps(k-1), param.levels, delta_y ); %#ok<AGROW>
                    error_propagation_noise_pdf{k} = conv_fct(error_propagation_noise_pdf{k-1}, post_error_dfe_noise_pdf{k}); %#ok<AGROW>
                end
                
                % Assume an error will propagate if this noise exceeds the threshold defined above
                x_error_propagation = find(error_propagation_noise_pdf{k}.x >= error_threshold, 1, 'first');
                if isempty(x_error_propagation)
                    p_error_propagation(k) = 1e-20; %#ok<AGROW>
                else
                    p_error_propagation(k) = sum(error_propagation_noise_pdf{k}.y(x_error_propagation:end)); %#ok<AGROW>
                end
            end
            
            % Assume an uncorrelated error will occur if the original noise exceeds
            % the available signal reduced by pass_threhsold dB. Find the probability
            % of this event by partial sum of the PDF.
            % p_uncorrelated_error_i = find(combined_interference_and_noise_pdf.x >= A_s./10^(param.pass_threshold/20), 1, 'first');
            % p_uncorrelated_error = sum(combined_interference_and_noise_pdf.y(p_uncorrelated_error_i:end));
            
            % probability of bursts of different lengths
            p_burst = cumprod(p_error_propagation);
        end
        if OP.SAVE_TD
            sbr=timeseries(fom_result.sbr,fom_result.t);
            if ~OP.TDMODE
                fir=timeseries(fom_result.IR,fom_result.t);
            end
            for i=1:number_of_s4p_files
                Pulses(i).uneq_responce= timeseries(chdata(i).uneq_pulse_response, chdata(i).t );
                Pulses(i).eq_responce= timeseries(chdata(i).eq_pulse_response, chdata(i).t );
                if ~OP.TDMODE
                    FIR(i).uneq_imp_response=  timeseries(chdata(i).uneq_imp_response, chdata(i).t );
                    FIR(i).eq_imp_response=  timeseries(chdata(i).eq_imp_response, chdata(i).t );
                end
            end
            if OP.TDMODE
                save( [OP.RESULT_DIR 'sbr_fir_' param.base '.mat'],'sbr','Pulses');
            else
                save( [OP.RESULT_DIR 'sbr_fir_' param.base '.mat'],'sbr','fir','Pulses','FIR')
            end
        end
        switch lower(OP.TDECQ)
            case { false 'none' } % should be the default
            case 'vma'
                est_vma=vma(fom_result.sbr,param.samples_per_ui);
                output_args.VMA=est_vma.VMA;
            otherwise
                error('%s not recognized for Histogram_Window_Weigh this feature is limited',OP.TDECQ)
        end
        output_args.config_file = config_file;
        
        fileset_str=str2csv({chdata.base});
        output_args.file_names=sprintf('"%s"', fileset_str);
        % [ahealey] Echo the termination parameters in the output arguments..
        for odt_param = {'R_diepad', 'C_diepad', 'L_comp', 'C_bump'}
            output_args.(odt_param{:}) = param.(odt_param{:});
        end
        % [ahealey] End of modifications.
        for pkg_params = {'levels', 'Pkg_len_TX', 'Pkg_len_NEXT', 'Pkg_len_FEXT', 'Pkg_len_RX','R_diepad','pkg_Z_c','C_v'}
            output_args.(pkg_params{:})= param.(pkg_params{:});
        end
        output_args.baud_rate_GHz=param.fb/1e9;
        output_args.f_Nyquist_GHz = param.fb/2e9;
        output_args.BER=param.specBER;
        output_args.FOM = fom_result.FOM;
        output_args.sigma_N=sigma_N;
        output_args.DFE4_RSS=norm(fom_result.DFE_taps(4:end));
        output_args.DFE2_RSS=norm(fom_result.DFE_taps(2:end));
        output_args.tail_RSS=fom_result.tail_RSS;
        output_args.channel_operating_margin_dB=COM;
        output_args.available_signal_after_eq_mV=1000*A_s;
        output_args.peak_uneq_pulse_mV=1000*max(abs(chdata(1).uneq_pulse_response));
        try
            output_args.uneq_FIR_peak_time=chdata(1).t(chdata(1).uneq_imp_response==max(chdata(1).uneq_imp_response));
        catch
        end
        output_args.steady_state_voltage_mV = 1000*fom_result.A_f; % RIM 7/03/2019 use peak from optimize_FOM
        its=find(chdata(1).eq_pulse_response>=max(chdata(1).eq_pulse_response),1,'first');
        output_args.steady_state_voltage_weq_mV = sum(chdata(1).eq_pulse_response(1:its+param.N_v*param.samples_per_ui) )/param.samples_per_ui;
        if OP.GET_FD
            output_args.FOM_ILD=FOM_ILD;
            output_args.FOM_RILN=FOM_RILN;
        end
        if OP.RX_CALIBRATION== 1, output_args.sigma_bn=sigma_bn; end
        output_args.Peak_ISI_XTK_and_Noise_interference_at_BER_mV=1000*A_ni;
        output_args.peak_ISI_XTK_interference_at_BER_mV=1000*peak_interference_at_BER;
        output_args.peak_ISI_interference_at_BER_mV=1000*thru_peak_interference_at_BER;
        output_args.equivalent_ICI_sigma_assuming_PDF_is_Gaussian_mV=sci_sigma*1000;
        
        if OP.RX_CALIBRATION == 0
            output_args.peak_MDXTK_interference_at_BER_mV=1000*crosstalk_peak_interference_at_BER;
            output_args.peak_MDNEXT_interference_at_BER_mV=1000*MDNEXT_peak_interference;
            output_args.peak_MDFEXT_interference_at_BER_mV=1000*MDFEXT_peak_interference;
        end
        output_args.ICN_mV=ICN*1000;
        %             output_args.ICN_test_mV=ICN_test*1000;
        if xtk>0 && OP.RX_CALIBRATION ==0 && OP.TDMODE==0
            output_args.MDNEXT_ICN_92_46_mV=MDNEXT_ICN*1000;
            output_args.MDFEXT_ICN_92_47_mV=MDFEXT_ICN*1000;
            output_args.equivalent_ICN_assuming_Gaussian_PDF_mV=cci_sigma*1000;
        else
            output_args.MDNEXT_ICN_92_46_mV=0;
            output_args.MDFEXT_ICN_92_47_mV=0;
            output_args.equivalent_ICN_assuming_PDF_is_Gaussian_mV=0;
        end
        %output_args.SNR_ISI_XTK_normalized_1_sigma=20*log10(A_s/(peak_interference_at_BER/qfuncinv(param.specBER))); modified by Yasuo Hidaka, 8/7/17
        if 1
            output_args.SNR_ISI_XTK_normalized_1_sigma=20*log10(A_s/(peak_interference_at_BER/sqrt(2)/erfcinv(2*param.specBER)));
            output_args.SNR_ISI_est=fom_result.SNR_ISI;
            output_args.Pmax_by_Vf_est=fom_result.Pmax_by_Vf;
        end
        
        if ~exist('msg','var')
            msg=[];
        end
        if OP.DEBUG
            [ncases, mele]=size(param.z_p_tx_cases);
            switch param.flex
                case 4
                    msg = sprintf('%s: Case %g: z_p=(%g:%g:%g:%g, %g:%g:%g:%g, %g:%g:%g:%g, %g:%g:%g:%g) (TX, RX, NEXT, FEXT):\n', ...
                        msg,package_testcase_i, param.Pkg_len_TX, param.Pkg_len_RX, param.Pkg_len_NEXT, param.Pkg_len_FEXT ...
                        );
                case 2
                    msg = sprintf('%s: Case %g: z_p=(%g:%g, %g:%g, %g:%g, %g:%g) (TX, RX, NEXT, FEXT):\n', ...
                        msg,package_testcase_i, param.Pkg_len_TX(1:2), param.Pkg_len_RX(1:2), param.Pkg_len_NEXT(1:2), param.Pkg_len_FEXT(1:2) ...
                        );
                otherwise
                    msg = sprintf('%s: Case %g: z_p=(%g, %g, %g, %g) (TX, RX, NEXT, FEXT):', ...
                        msg, package_testcase_i, param.Pkg_len_TX, param.Pkg_len_RX, param.Pkg_len_NEXT, param.Pkg_len_FEXT ...
                        );
                    
            end
        else
            msg = sprintf('Case %d:', package_testcase_i );
        end
        
        
        if OP.DISPLAY_WINDOW && ~OP.RX_CALIBRATION
            % display bathtub curves in one axis per test case.
            if ~OP.COM_CONTRIBUTION_CURVES
                figure_name =  'Voltage bathtub curves';
                fig=findobj('Name', figure_name);
                if isempty(fig), fig=figure('Name', figure_name); end
                figure(fig);set(gcf,'Tag','COM');
                movegui(fig,'south')
                hax = subplot(length(OP.pkg_len_select), 1, package_testcase_i);
                plot_bathtub_curves( hax ...
                    , A_s ...
                    , sci_pdf ...
                    , cci_pdf ...
                    , isi_and_xtalk_pdf ...
                    , noise_pdf ...
                    , jitt_pdf ...
                    , combined_interference_and_noise_pdf ...
                    , delta_y ...
                    );
                set(hax, 'tag', 'BTC');
                title(hax, sprintf('case %d VBC: %s ', package_testcase_i, regexprep([chdata(1).base,' '],'_',' ')));
                ylim(hax, [param.specBER/10 1]);
                % show BER target line
                hp=plot(get(hax, 'xlim'), param.specBER*[1 1], 'r:');
                set(get(get(hp,'Annotation'),'LegendInformation'), 'IconDisplayStyle','off');
            else
                figure_name =  'COM Contributions (Rough Allocations)';
                fig=findobj('Name', figure_name);
                if isempty(fig), fig=figure('Name', figure_name); end
                figure(fig);set(gcf,'Tag','COM');
                movegui(fig,'south')
                hax = subplot(length(OP.pkg_len_select), 1, package_testcase_i);
                
                plot_pie_com( hax ...
                    , A_s ...
                    , sci_pdf ...
                    , cci_pdf ...
                    , isi_and_xtalk_pdf ...
                    , noise_pdf ...
                    , combined_interference_and_noise_pdf ...
                    , delta_y, param...
                    );
                set(hax, 'tag', 'BTC');
                title(hax, sprintf('case %d rough COM impact: %s ', package_testcase_i, regexprep([chdata(1).base,' '],'_',' ')));
            end
        end
        
        if ~exist('msg','var')
            msg=[];
        end
        if OP.TDMODE
            min_ERL=inf;
            ERL=[inf inf];
        end
        [msg] = end_display_control(msg,param,OP,output_args,COM,min_ERL,ERL, VEO_mV,VEC_dB,threshold_DER,OP.DISPLAY_WINDOW); % {} forces no ERL print
        switch param.CTLE_type
            case 'CL93'
                output_args.CTLE_zero_poles=[param.CTLE_fz(fom_result.ctle) param.CTLE_fp2(fom_result.ctle) param.CTLE_fp1(fom_result.ctle)];
                output_args.CTLE_DC_gain_dB=param.ctle_gdc_values(fom_result.ctle);
            case 'CL120d'
                output_args.CTLE_zero_poles=[param.CTLE_fz(fom_result.ctle) param.CTLE_fp2(fom_result.ctle) param.CTLE_fp1(fom_result.ctle)];
                output_args.CTLE_DC_gain_dB=param.ctle_gdc_values(fom_result.ctle);
                output_args.g_DC_HP=param.g_DC_HP_values(fom_result.best_G_high_pass);
                output_args.HP_poles_zero=param.f_HP(fom_result.best_G_high_pass);
            case 'CL120e'
                output_args.CTLE_zero_poles=[param.CTLE_fz(fom_result.ctle) param.f_HP_Z(fom_result.ctle) param.CTLE_fp2(fom_result.ctle) param.CTLE_fp1(fom_result.ctle) param.f_HP_P(fom_result.ctle)];
                output_args.CTLE_DC_gain_dB=param.ctle_gdc_values(fom_result.ctle);
        end
        output_args.TXLE_taps=fom_result.txffe;
        if  length(output_args.TXLE_taps) >= 3
            output_args.Pre2Pmax = -output_args.TXLE_taps(end-2)/output_args.TXLE_taps(end-1);
        end
        output_args.DFE_taps=fom_result.DFE_taps;
        if param.N_bg ~=0
            output_args.floating_tap_locations=fom_result.floating_tap_locations;
        end
        if OP.nburst>0
            output_args.error_propagation_probability = p_error_propagation;
            output_args.burst_probabilities = p_burst;
        end
        %% making mat file
        %if ~OP.EXTERNAL
        if(OP.DEBUG)
            save (sprintf('%s%s_case%d_results.mat', OP.RESULT_DIR, chdata(1).base, package_testcase_i), ...
                'output_args','param','OP');
        end
        %else %??
        %output_args.baud_rate_GHz=param.fb/1e9;
        
        %Question on these 2 lines.  Any need for them?
        %output_args.sigma_rj=param.sigma_RJ;
        %output_args.Add_noise=param.A_DD;
        %end
        
        % begin yasuo patch 12/11/2018
        % store sigma values to the output argument
        output_args.sgm_Ani__isi_xt_noise = sgm_Ani__isi_xt_noise;
        output_args.sgm_isi_xt = sgm_isi_xt;
        output_args.sgm_isi = sgm_isi;
        if exist('sgm_xt','var')    output_args.sgm_xt = sgm_xt; end
        output_args.sgm_noise__gaussian_noise_p_DD = sgm_noise__gaussian_noise_p_DD;
        output_args.sgm_p_DD = sgm_p_DD;
        output_args.sgm_gaussian_noise = sgm_gaussian_noise;
        output_args.sgm_G = sgm_G;
        output_args.sgm_rjit = sgm_rjit;
        output_args.sgm_N = sgm_N;
        output_args.sgm_TX = sgm_TX;
        % r259 putting COM, VEO and loss last in report
        if OP.GET_FD
            output_args.total_IL_wpkgs_dB_at_Fnq=total_loss;
            output_args.IL_dB_channel_only_at_Fnq=Nq_loss;
            output_args.fitted_IL_dB_at_Fnq = fit_loss;
            if OP.include_pcb
                output_args.cable__assembley_loss=cable_loss;
            end
        end
        %         output_args.VEO_normalized = (A_s-A_ni)/A_s;
        output_args.VEC_dB = VEC_dB;
        output_args.VEO_mV = VEO_mV;
        if OP.RX_CALIBRATION ==0 && OP.EW == 1
            output_args.EW_UI_est=EW_UI;
            output_args.eye_contour=eye_contour;
            output_args.VEO_window_mUI= param.T_O;
        end
        
        if sum(param.AC_CM_RMS) ~= 0
            output_args.sigma_ACCM_at_tp0_mV=param.sigma_ACCM_at_tp0*1000;
            fprintf(' AC RMS at TP0 = %.3g mV \n',output_args.sigma_ACCM_at_tp0_mV)
        end
        output_args.COM_dB=COM;
        
        
        % end yasuo patch
        % begin yasuo patch 3/18/2019
        output_args.DER_thresh = threshold_DER;
        % end yasuo patch
        %% making csv file
        if OP.CSV_REPORT ==1
            items = fieldnames(output_args);
            item_value_strings = cell(size(items));
            for field_id=1:length(items)
                field_name=items{field_id};
                field_value=output_args.(field_name);
                if isstruct(output_args.(field_name))
                    field_value='struct';
                end
                if ischar(field_value)
                    item_value_strings{field_id}=field_value;
                elseif numel(field_value)==1
                    item_value_strings{field_id}=num2str(field_value);
                else
                    item_value_strings{field_id}=sprintf('"%s"', mat2str(field_value));
                end
            end
            
            header_string = str2csv(items);
            data_string = str2csv(item_value_strings);
            fid = fopen(sprintf('%s%s_case%d_results.csv', OP.RESULT_DIR, chdata(1).base, package_testcase_i),'w');
            fprintf(fid,'%s\n', header_string);
            fprintf(fid,'%s\n', data_string);
            fclose(fid);
        end
        rt=toc(t0);
        output_args.rtmin=rt/60;
        if 1
            fprintf('run time = %g min \n',output_args.rtmin)
        end
        if OP.BREAD_CRUMBS
            output_args.OP=OP;
            output_args.param=param;
            output_args.chdata=chdata;
            output_args.fom_result = fom_result;
        end
        results{package_testcase_i} = output_args;
        if nargout==0
            fprintf('<strong>--- Testcase %d results ---</strong>\n', package_testcase_i);
            disp(output_args)
        end
        % package test cases
        
        if OP.DEBUG && OP.DISPLAY_WINDOW && OP.RX_CALIBRATION==0
            btc_axes = findobj('tag', 'BTC');
            if ~isempty(btc_axes), linkaxes(btc_axes, 'x'); end
            eqe_axes = findobj('tag', 'EQE');
            if ~isempty(eqe_axes), linkaxes(eqe_axes, 'xy'); end
        end
        if OP.BREAD_CRUMBS
            [my_path,rootname]=fileparts(chdata(1).filename);
            save (sprintf('%s%s_case%d_results.mat', OP.RESULT_DIR, rootname, package_testcase), ...
                'output_args','param','OP');
        end
    end
    [tmp] = end_display_control('WC All cases',param,OP,output_args,min_COM,min_ERL,ERL,min_VEO_mV,max_VEC_dB,threshold_DER,0);
    %%
    if OP.RX_CALIBRATION ==1
        display ([' LOOP with sigma_bn = ' num2str(sigma_bn) ' performed with COM = ' num2str(min_COM) ])
    end
    DO_ONCE=false;
end

if OP.DISPLAY_WINDOW
    savefigs(param, OP);
    set(0,'defaulttextinterpreter','tex'); % reset defaut text interpreter to tex
end

if OP.RX_CALIBRATION==1
    fprintf ('Set Rx calibration noise rms voltage to %g mV\n', sigma_bn*1000);
    if OP.DISPLAY_WINDOW
        message=sprintf('Set Rx calibration noise rms voltage to %g mV.',sigma_bn*1000);
        hlast = msgbox(message,'sigma_bn','help');
        set(hlast,'Color','y', 'tag', 'COM');
    end
end
% if OP.BREAD_CRUMBS
%     save( [OP.RESULT_DIR 'param_n_OP_' param.base '.mat'],'param','OP','chdata','fom_result','output_args')
% end
if length(results)==1, results = results{1}; end
redo_cmd_str=' redo string is: eval([''My_var_0 = '' getappdata(0,''cmd_str'')])';
disp(redo_cmd_str);
if isdeployed
    if OP.exit_if_deployed
        quit
    end
end
%%
%--------------------------------------------------------------------------
%--------------- Helper functions -----------------------------------------
%--------------------------------------------------------------------------
function [Left_EW,Right_EW,eye_contour,out_VT,out_VB]=COM_eye_width(chdata,delta_y,fom_result,param,OP,Struct_Noise,pdf_range_flag)


debug_plot=0;


samp_UI=param.samples_for_C2M;
half_UI=get_center_of_UI(samp_UI);
T_O=floor((param.T_O/1000)*param.samples_for_C2M);
start_sample=half_UI-T_O;
end_sample=half_UI+T_O;


%pdf_range is a placeholder to allow fractional UI calculation for optimize
%C2M speedup.  For regular COM_eye_width calls, pdf_range will be empty
%which will force pdf_range=1:samp_UI for stanard full UI calculation.
if pdf_range_flag
    pdf_range=[start_sample end_sample];
else
    pdf_range=[];
end

%pdf_full is self ISI pdf for each sample point
%h_j_full is the v/t calculation for each sample point
%the center vector for each should be identical to the standard COM variables
[pdf_full_1,h_j_full,A_s_vec] = get_pdf_full(chdata(1), delta_y, fom_result.t_s, param, OP,pdf_range) ;



if isempty(pdf_range)
    pdf_range=1:samp_UI;
else
    pdf_range=min(pdf_range):max(pdf_range);
end

%Test doing Level PDFs
Levels = 2*(0:param.levels-1)/(param.levels-1)-1;
%A_s_vec=A_s_vec*(param.levels-1)/param.R_LM;
A_s_vec=A_s_vec*(param.levels-1);

%add signal vector into pdf
for n=1:param.levels
    pdf_full{n}=pdf_full_1;
    for j=pdf_range
        pdf_full{n}(j).x=pdf_full{n}(j).x+A_s_vec(j)*Levels(n);
        pdf_full{n}(j).Min=pdf_full{n}(j).x(1)/pdf_full{n}(j).BinSize;
    end
end


% figure;
% hold on;
%This loop builds the same PDF/CDF structures from regular COM, but it is
%computed for every sample point
for n=1:param.levels
    for j=pdf_range
        sigma_G_full(j)  = norm([param.sigma_RJ*param.sigma_X*norm(h_j_full(:,j)), Struct_Noise.sigma_N, Struct_Noise.sigma_TX]);
        gaussian_noise_pdf_full(j) = normal_dist(sigma_G_full(j), Struct_Noise.ber_q, delta_y);
        gaussian_noise_pdf_full(j) = conv_fct(gaussian_noise_pdf_full(j), Struct_Noise.ne_noise_pdf);
        p_DD_full(j) = get_pdf_from_sampled_signal(param.A_DD*h_j_full(:,j), param.levels, delta_y);
        noise_pdf_full(j)=conv_fct(gaussian_noise_pdf_full(j), p_DD_full(j));
        isi_and_xtalk_pdf_full(j) = conv_fct_MeanNotZero(pdf_full{n}(j), Struct_Noise.cci_pdf);
        % change from adam gregory to include crosstalk
        %     combined_interference_and_noise_pdf_full = conv_fct(pdf_full(j), noise_pdf_full(j));
        combined_interference_and_noise_pdf_full{n}(j) = conv_fct_MeanNotZero(isi_and_xtalk_pdf_full(j), noise_pdf_full(j));
        
        %PDF to CDF
        combined_interference_and_noise_cdf_full{n}(j)=pdf_to_cdf(combined_interference_and_noise_pdf_full{n}(j));
        
    end
end
%hold off;


%For the given BER, find the top & bottom voltage level in the CDF
for n=1:param.levels
    A_ni_bottom{n}=zeros(1,samp_UI);
    A_ni_top{n}=zeros(1,samp_UI);
    for j=pdf_range
        [A_ni_top{n}(j),A_ni_bottom{n}(j)]=cdf_to_ber_contour(combined_interference_and_noise_cdf_full{n}(j),param.specBER);
    end
end
%plot(1:samp_UI,cursor_vector-A_ni_top,1:samp_UI,-cursor_vector+A_ni_bottom)

for n=1:param.levels-1
    eye_contour{n}(:,1)=A_ni_top{n+1};
    eye_contour{n}(:,2)=A_ni_bottom{n};
end


for n=1:param.levels-1
    %eye_contour holds the top eye in the 1st column & bottom eye in the 2nd column
    %define vref as middle of top eye height and bottom eye height.  Now
    %that all eyes are created, vref is non-zero except for middle eye
    EH_top=eye_contour{n}(half_UI,1);
    EH_bot=eye_contour{n}(half_UI,2);
    EH=EH_top-EH_bot;
    vref=EH_top/2+EH_bot/2;
    %This function finds left/right eye width by finding the vref crossings of
    %the top and bottom eye contours
    [Left_EW(n),Right_EW(n)]=find_eye_width(eye_contour{n},half_UI,samp_UI,vref);
end

%For reporting to .csv, need eye contour to be a matrix instead of cell
eye_contour_tmp=eye_contour;
eye_contour=[];
for n=1:param.levels-1
    eye_contour(:,(n-1)*2+1:n*2)=eye_contour_tmp{n};
end


%Find VEC eye height
out_VT=[];
out_VB=[];
if param.T_O ~=0
    
    switch lower(OP.Histogram_Window_Weight)
        case {'gaussian' 'norm' 'normal' 'guassian'}
            %build a gaussian window of weights that are multiplied by each pdf in the T_O range
            %Sigma = T_O/QL.  Default QL=2.5.  This gives a nice descent to near 0 at the edge of the window
            QL_sigma=T_O/param.QL;
            weights=exp(-1/2 * ([-T_O:T_O]/QL_sigma).^2);
        case 'triangle'
            %triangle window. linear slope from 0 to 1 and back down to 0
            %for the weights
            t_slope=1/(T_O);
            weights=[0:t_slope:1 1-t_slope:-t_slope:0];
        case 'rectangle'
            %default = rectangle.  all weights = 1
            weights(1:2*T_O+1)=1;
        case 'dual_rayleigh'
            QL_sigma=T_O/param.QL;
            X=-T_O:T_O;
            weights=(X+T_O)/QL_sigma^2.*exp(-1/2 * ((X+T_O)/QL_sigma).^2)...
                -(X-T_O)/QL_sigma^2.*exp(-1/2 * ((X-T_O)/QL_sigma).^2);
            weights=weights/max(weights);
        otherwise
            error('%s not recognized for Histogram_Window_Weight',OP.Histogram_Window_Weight)
    end
    
    for n=1:param.levels
        out_pdf{n}=[];
        for j=start_sample:end_sample
            target_pdf=combined_interference_and_noise_pdf_full{n}(j);
            target_pdf.y=target_pdf.y*weights(j-start_sample+1);
            if isempty(out_pdf{n})
                out_pdf{n}=target_pdf;
            else
                out_pdf{n} = combine_pdf_same_voltage_axis(out_pdf{n}, target_pdf);
            end
        end
        out_pdf{n}.y=out_pdf{n}.y/sum(out_pdf{n}.y);
    end
    
    for n=1:param.levels
        out_cdf{n}=pdf_to_cdf(out_pdf{n});
    end
    
    for n=1:param.levels
        [A_ni_top_O(n),A_ni_bottom_O(n)]=cdf_to_ber_contour(out_cdf{n},param.specBER);
    end
    
    for n=1:param.levels-1
        OUT_VT_L(n,1)=A_ni_top_O(n+1);
        OUT_VT_L(n,2)=A_ni_bottom_O(n);
    end
    
    %Report the top/bottom eye height of the worst eye
    EH_VT=OUT_VT_L(:,1)-OUT_VT_L(:,2);
    [mineh,min_idx]=min(EH_VT);
    out_VT=OUT_VT_L(min_idx,1);
    out_VB=OUT_VT_L(min_idx,2);
    
    %     CDF_Mean=mean(A_s_vec(start_sample:end_sample));
    %     out_VT=2*CDF_Mean-A_ni_top_O;
    %     out_VB=-1*A_ni_bottom_O;
    
    if debug_plot
        figure;
        hold on;
        for j=start_sample:end_sample
            plot(combined_interference_and_noise_pdf_full(j).x,combined_interference_and_noise_pdf_full(j).y);
        end
        plot(out_pdf.x,out_pdf.y,'color','k','LineWidth',2);
        hold off;
    end
end




function [hctf] = FD_CTLE(freq, fb, f_z, f_p1, f_p2, kacdc_dB)
hctf = ( 10.^(kacdc_dB/20) + 1i*freq/f_z ) ./ ( (1+1i*freq/f_p1) .* (1+1i*freq/f_p2)) ;

function [ V0 ] = FFE( C , cmx,spui, V )
% C      FFE taps
% cmx    number of precursors taps
% spui   samples per ui
% V      input signal
V0=0;
if iscolumn(V); V=V.';end
for i=1:length(C)
    ishift=(i-1)*spui;
    V0=circshift(V',[ishift,0])*C(i)+V0;
end
V0=circshift(V0,[(-cmx)*spui,0]);


% begin yasuo patch 12/11/2018
% calculate sigma (standard deviation) value of PDF
function pdf=Init_PDF_Fast( EmptyPDF, values, probs)
%  p=cpdf(type, ...)
%
% CPDF is a probability mass function for discrete distributions or an
% approxmation of a PDF for continuous distributions.
%
% cpdf is internally normalized so that the sum of probabilities is 1
% (regardless of bin size).

% Internal fields:
% Min: *bin number* of minimum value.
% BinSize: size of PDF bins. Bin center is the representative value.
% Vec: vector of probabilities per bin.

pdf=EmptyPDF;

rounded_values_div_binsize=round(values/pdf.BinSize);
%values=pdf.BinSize*rounded_values_div_binsize;

% %speed up for small values round to 0 (because they are all much smaller than binsize)
% if all(values==0)
%     return;
% end
%
% %speed up for all values rounded to the same bin
% %The output pdf is the same as the
% %empty pdf, but the x value is non-zero (but still scalar)
% if all(values==values(1))
%     pdf.Min=rounded_values_div_binsize(1);
%     pdf.x=values(1);
%     return;
% end
%
% %The code below requires that values is
% %sorted.  Generally this should be true, but check to be sure
% if ~issorted(values)
%     [values,si]=sort(values);
%     rounded_values_div_binsize=rounded_values_div_binsize(si);
%     probs=probs(si);
% end


%pdf.x=values(1):pdf.BinSize:values(end);
pdf.x=pdf.BinSize*rounded_values_div_binsize(1):pdf.BinSize:pdf.BinSize*rounded_values_div_binsize(end);
pdf.Min=rounded_values_div_binsize(1);

pdf.y=zeros(size(pdf.x));
%The rounded values divided by binsize will reveal the bin number if
%pdf.Min is subtracted from it
bin_placement=rounded_values_div_binsize-pdf.Min+1;
%Can avoid one addition by inserting the first probability
%actually helps when calling this 2 million times
pdf.y(bin_placement(1))=probs(1);
for k=2:length(values)
    pdf.y(bin_placement(k)) = pdf.y(bin_placement(k))+probs(k);
end


%Have already ensured that sum(pdf.y)=1
%pdf.y=pdf.y/sum(pdf.y);

% if any(~isreal(pdf.y)) || any(pdf.y<0)
%     error('PDF must be real and nonnegative');
% end

% pMax=pdf.Min+length(pdf.y)-1;
% pdf.x = values(1):pdf.BinSize:pMax*pdf.BinSize;
function [ seq syms syms_nrz ] = PRBS13Q( )
%UNTITLED Summary of this function goes here
%   Detailed explanation goes here


taps = ([13 12 2  1]);
seed =([0 0 0 0 0 1 0 1 0 1 0 1 1]);
[seq_nrz c] =LFSR(seed,taps);
seq_nrz=2*(seq_nrz-0.5);
seq=pam(seq_nrz);
% syms=round(2*(seq+1));
syms((round(2*(seq+1))/2==2))=3;
syms((round(2*(seq+1))/2==1.5))=2;
syms((round(2*(seq+1))/2==.5))=1;
syms((round(2*(seq+1))/2==0))=0;

% syms_nrz=((seq_nrz+1)/2);

syms_nrz=seq_nrz;


function[seq c]=LFSR(s,t)
%s=initial state of LFSR, you can choose any lenght of LFSR
%Instruction:==========
%Save LFSR.m in your current directory and type following
%on Command window for simulating 5 bit LFSR with tap [5 2]
%---------------------
%>>s=[1 1 0 0 1]
%>>t=[5 2]
%>>[seq c] =LFSR(s,t)
%---------------------------
%seq = generated sequence
%c will be matrix containing the states of LFSR raw wise
%
%-----------------------------------------------------------
%If any doubt, confusion or feedback please contact me
% NIKESH BAJAJ
% bajaj.nikkey@gmail.com (+91-9915522564)
% Asst. Professor at Lovely Profesional University
% Masters from Aligarh Muslim University,INDIA
%--------------------------------------------------
n=length(s);
c(1,:)=s;
m=length(t);
for k=1:2^n-2;
    b(1)=xor(s(t(1)), s(t(2)));
    if m>2;
        for i=1:m-2;
            b(i+1)=xor(s(t(i+2)), b(i));
        end
    end
    j=1:n-1;
    s(n+1-j)=s(n-j);
    s(1)=b(m-1);
    c(k+1,:)=s;
end
seq=c(:,n)';

function [ dataout ] = pam( data )
% mapping data usng Grey Coding
for i=1:2:floor(length(data)/2)*2
    if data(i:i+1)==[ -1 -1 ]
        dataout(ceil(i/2)) = -1;
    elseif data(i:i+1)==[ -1 1 ]
        dataout(ceil(i/2)) = -1/3;
    elseif data(i:i+1)==[ 1 1 ]
        dataout(ceil(i/2)) = 1/3;
    elseif data(i:i+1)==[ 1 -1 ]
        dataout(ceil(i/2)) = 1;
    end
end
function S =R_series2(zref,f,R)
r=ones(1,length(f))*R;
S.Parameters(1,1,:) =  r./(r + 2*zref);
S.Parameters(2,2,:) =  r./(r + 2*zref);
S.Parameters(2,1,:) = (2*zref)./(r + 2*zref);
S.Parameters(1,2,:) =  (2*zref)./(r + 2*zref);
Sm=sparameters(S.Parameters,f,zref);

function SLD=SL(S,f,R)
% source load impact return loss add to S21
% S and SLD are the same structure
% S.Parameters
% S.Impedance
% S.NumPorts
% S.Frequencies
SLD=S; % assign the fields
zref=100;
if R==0
    warndlg('Termination should not be set to zero');
    SLD=S;
    return
end

if R > zref
    spr =R_series2(zref,f,(R-zref)); % make series sparameter
    %     SLD=sparameters(cascadesparams(S.Parameters,spr.Parameters),f,S.Impedance); % Casdeade
    [ SLD.Parameters(1,1,:), SLD.Parameters(1,2,:), SLD.Parameters(2,1,:), SLD.Parameters(2,2,:)]  = ...
        combines4p( ...
        spr.Parameters(1,1,:), spr.Parameters(1,2,:), spr.Parameters(2,1,:), spr.Parameters(2,2,:),...
        S.Parameters(1,1,:),   S.Parameters(1,2,:),   S.Parameters(2,1,:),   S.Parameters(2,2,:) ...
        );
elseif R < zref
    spr =r_parrelell2(zref,f,-R*zref/(R-zref)); % make parrellel sparameter
    %     SLD=sparameters(cascadesparams(S.Parameters,spr.Parameters),f,S.Impedance);
    [ SLD.Parameters(1,1,:), SLD.Parameters(1,2,:), SLD.Parameters(2,1,:), SLD.Parameters(2,2,:)]  = ...
        combines4p( ...
        spr.Parameters(1,1,:),  spr.Parameters(1,2,:), spr.Parameters(2,1,:), spr.Parameters(2,2,:),...
        S.Parameters(1,1,:),   S.Parameters(1,2,:),   S.Parameters(2,1,:),   S.Parameters(2,2,:) ...
        );
else
    SLD=S;
end

%%

function [impulse_response, p1_ctle, p2_ctle, z_ctle] = TD_CTLE(ir_in, fb, f_z, f_p1, f_p2, kacdc_dB, oversampling)
%% Equation 93A-22 implemented in z-space and applied to the impulse response.
p1_ctle = -2*pi*f_p1;
p2_ctle = -2*pi*f_p2;
z_ctle = -2*pi*f_z*10^(kacdc_dB/20);
k_ctle = -p2_ctle;
bilinear_fs = 2*fb*oversampling;
p2d = (1+p2_ctle/bilinear_fs)./(1-p2_ctle/bilinear_fs);
p1d = (1+p1_ctle/bilinear_fs)./(1-p1_ctle/bilinear_fs);
zd = (1+z_ctle/bilinear_fs)./(1-z_ctle/bilinear_fs);
% kd = (bilinear_fs-z_ctle)/((bilinear_fs-p1_ctle)*(bilinear_fs-p2_ctle));
% allow for different pole zeros RIM 9-29-2015
kd = (bilinear_fs-z_ctle)./((bilinear_fs-p1_ctle)*(bilinear_fs-p2_ctle))*f_p1/f_z;
B_filt =k_ctle*kd*poly([zd, -1]);
A_filt=poly([p1d, p2d]);
impulse_response=filter(B_filt,A_filt,ir_in);

function  [chdata, param, SDDch,SDDp2p] = TD_FD_fillin(param, OP, chdata)
Over_sample=2;
num_files=length(chdata);
for i=1:num_files
    V=chdata(i).uneq_pulse_response;
    T=chdata(i).t;
    dt=T(2)-T(1);
    f=0:1/max(T):1/dt;
    chdata(i).faxis=f;
    f75=find(f >= param.fb*.75,1,'first');
    fnq=find(f >= param.fb*.5,1,'first');
    chdata(i).fmaxi = length(f);
    chdata(i).faxis = f;
    UI=param.ui; % unit interval
    M=param.samples_per_ui; % sample per UI
    N_v=param.N_v; % number of UI for Vf determination
    % filters
    param.fb_BW_cutoff=param.f_r;
    factor_3db=0.473037;
    param.fb_BT_cutoff=factor_3db*param.f_r;
    if OP.Bessel_Thomson
        a = bessel( param.BTorder );
        acoef=fliplr( a );
        H_bt =a(1)./ polyval(acoef, (1i*f./(param.fb_BT_cutoff*param.fb)));
    else
        H_bt=ones(1,length(f));
    end
    if OP.Butterworth
        H_bw = 1./polyval([1 2.613126 3.414214 2.613126 1], 1i*f./(param.fb_BW_cutoff*param.fb));
    else
        H_bw=ones(1,length(f));
    end
    H_ftr=H_bw.*H_bt;
    H_ftr=H_ftr(:);
    % fd of PR
    prr=sin(f*UI*pi)./(f*UI*pi); %nremoving sinc function to avoid using sig proc toolbox
    prr = prr(:);
    fd=fft(V);
    fd=fd(1:floor(length(fd)/2)); % un process freq domain response
    
    %% get Vf
    shifting_vector=kron(ones(1,200) ,[ 1  zeros(1,M-1) ]) ;
    step_response=filter(V,1, shifting_vector);
    Vf=step_response(end);
    STEP=timeseries(step_response(1:length(shifting_vector)),T(1:length(shifting_vector)));
    %%
    % ILest=20.*log10(abs(fd(1:f75)/Vf/M/Over_sample))  -   20.*log10(abs(prr(1:f75))) - 20*log10(abs(squeeze(H_ftr(1:f75)))) ;  %removing db function to avoid using sig proc toolbox
    % figure
    % plot(f(1:f75),ILest)
    
    IL_conv=fd(1:f75)/Vf/M/Over_sample  ./  prr(1:f75) ./ H_ftr(1:f75) ;  %removing db function to avoid using sig proc toolbox
    % set same variables as get_s4p_files
    chdata(i).sdd12_raw =IL_conv;
    chdata(i).sdd21_raw = IL_conv;
    chdata(i).sdd22_raw = zeros(length(IL_conv),1);
    chdata(i).sdd11_raw = zeros(length(IL_conv),1);
    % mode conversion
    chdata(i).sdc12_raw = zeros(length(IL_conv),1);
    chdata(i).sdc21_raw = zeros(length(IL_conv),1);
    chdata(i).sdc22_raw = zeros(length(IL_conv),1);
    chdata(i).sdc11_raw = zeros(length(IL_conv),1);
    chdata(i).sdd12_orig=chdata(i).sdd12_raw;
    chdata(i).sdd21_orig=chdata(i).sdd21_raw;
    chdata(i).sdd21_orig=chdata(i).sdd11_raw;
    chdata(i).sdd22_orig=chdata(i).sdd22_raw;
    
    chdata(i).sdd12=chdata(i).sdd12_raw;
    chdata(i).sdd21=chdata(i).sdd21_raw;
    chdata(i).sdd21=chdata(i).sdd11_raw;
    chdata(i).sdd22=chdata(i).sdd22_raw;
    
    chdata(i).sdc12=chdata(i).sdc12_raw;
    chdata(i).sdc21=chdata(i).sdc21_raw;
    chdata(i).sdc21_=chdata(i).sdc11_raw;
    chdata(i).sdc22=chdata(i).sdc22_raw;
    
    chdata(i).sdd21p=chdata(i).sdd21_raw;
    chdata(i).sdc21p=chdata(i).sdc21_raw;
    chdata(i).sdd21f=chdata(i).sdd21_raw;
    if i==1
        SDDch(:,1,2)=chdata.sdd12_raw;
        SDDch(:,2,1)=chdata.sdd21_raw;
        SDDch(:,1,1)=chdata.sdd11_raw;
        SDDch(:,2,2)=chdata.sdd22_raw;
        SDDp2p= zeros(length(IL_conv),1);
    end
    chdata(i).TX_RL=[];
    chdata(i).TDR11=[];
    chdata(i).PDTR11=[];
    chdata(i).TDR22=[];
    chdata(i).PDTR22=[];
end


function H_tw=Tukey_Window(f,param)
fb=param.fb;
fr=param.f_r*param.fb;
fperiod=2*(fb-fr);
H_tw = [ ones(1,length(f(f<fr))) ...
    0.5*cos(2*pi*(f( f>=fr & f<=fb)  -fb ) /fperiod-pi)+.5 ...
    zeros(1,length(f(f>fb)) )];
H_tw=H_tw(1:length(f));
if 0
    plot(f/1e9,H_tw)
end


%% moved output control to functions
function [ s11out, s12out, s21out, s22out ] = add_brd(chdata, param, OP)
%% Used in Clause 92 for adding board trace between TP0 and TP2

switch chdata.type
    case 'THRU'
        z_bp_tx = param.z_bp_tx;
        z_bp_rx = param.z_bp_rx;
    case 'NEXT'
        z_bp_tx = param.z_bp_rx;
        z_bp_rx = param.z_bp_next;
    case 'FEXT'
        z_bp_tx = param.z_bp_fext;
        z_bp_rx = param.z_bp_rx;
end
% Same cap on each tx and rx three is a data stratue for bifrucation but
% logic no implemented here RIM 06/28/2019
zref=param.Z0;
c1=param.C_0;
c2=param.C_1;
f=chdata.faxis;
f(f<eps)=eps;
s11pad1t= -1i*2*pi.*f*c1(1)*zref./(2+1i*2*pi.*f*c1(1)*zref);
s21pad1t= 2./(2+1i*2*pi.*f*c1(1)*zref);
s11pad2t= -1i*2*pi.*f*c2(1)*zref./(2+1i*2*pi.*f*c2(1)*zref);
s21pad2t= 2./(2+1i*2*pi.*f*c2(1)*zref);
[ s11tx, s12tx, s21tx, s22tx ] = synth_tline(chdata.faxis, param.brd_Z_c(1), param.Z0, param.brd_gamma0_a1_a2, param.brd_tau, z_bp_tx);
% add Tx caps
[s11tx, s12tx, s21tx, s22tx  ]= ...
    combines4p(  s11pad1t, s21pad1t, s21pad1t, s11pad1t, s11tx, s12tx, s21tx, s22tx   );
[s11tx, s12tx, s21tx, s22tx  ]= ...
    combines4p(   s11tx, s12tx, s21tx, s22tx,s11pad2t, s21pad2t, s21pad2t, s11pad2t  );


s11pad1r= -1i*2*pi.*f*c1(2)*zref./(2+1i*2*pi.*f*c1(2)*zref);
s21pad1r= 2./(2+1i*2*pi.*f*c1(2)*zref);
s11pad2r= -1i*2*pi.*f*c2(2)*zref./(2+1i*2*pi.*f*c2(2)*zref);
s21pad2r= 2./(2+1i*2*pi.*f*c2(2)*zref);
[ s11rx, s12rx, s21rx, s22rx ] = synth_tline(chdata.faxis, param.brd_Z_c(2), param.Z0, param.brd_gamma0_a1_a2, param.brd_tau, z_bp_rx);
% add Rx caps
[s11rx, s12rx, s21rx, s22rx  ]= ...
    combines4p(  s11pad2r, s21pad2r, s21pad2r, s11pad2r, s11rx, s12rx, s21rx, s22rx   );
[s11rx, s12rx, s21rx, s22rx  ]= ...
    combines4p(   s11rx, s12rx, s21rx, s22rx,s11pad1r, s21pad1r, s21pad1r, s11pad1r  );


switch OP.include_pcb
    case 1
        [ s11out1, s12out1, s21out1, s22out1 ]=combines4p(  s11tx, s12tx, s21tx, s22tx, chdata.sdd11_raw, chdata.sdd21_raw, chdata.sdd12_raw, chdata.sdd22_raw );
        [ s11out, s12out, s21out, s22out ]=combines4p(  s11out1, s12out1, s21out1, s22out1,  s11rx, s12rx, s21rx, s22rx);
    case 2
        [ s11out, s12out, s21out, s22out ]=combines4p( chdata.sdd11_raw, chdata.sdd21_raw, chdata.sdd12_raw, chdata.sdd22_raw, s11rx, s12rx, s21rx, s22rx);
end

function [ s11out, s12out, s21out, s22out ] = add_brdorig(chdata, param, OP)
% Used in Clause 92 for adding board trace between TP0 and TP2
switch chdata.type
    case 'THRU'
        z_bp_tx = param.z_bp_tx;
    case 'NEXT'
        z_bp_tx = param.z_bp_next;
    case 'FEXT'
        z_bp_tx = param.z_bp_fext;
end

[ s11tx, s12tx, s21tx, s22tx ] = synth_tline(chdata.faxis, param.brd_Z_c(1), param.Z0, param.brd_gamma0_a1_a2, param.brd_tau, z_bp_tx);
[ s11rx, s12rx, s21rx, s22rx ] = synth_tline(chdata.faxis, param.brd_Z_c(2), param.Z0, param.brd_gamma0_a1_a2, param.brd_tau, param.z_bp_rx);

switch OP.include_pcb
    case 1
        [ s11out1, s12out1, s21out1, s22out1 ]=combines4p(  s11tx, s12tx, s21tx, s22tx, chdata.sdd11_raw, chdata.sdd21_raw, chdata.sdd12_raw, chdata.sdd22_raw );
        [ s11out, s12out, s21out, s22out ]=combines4p(  s11out1, s12out1, s21out1, s22out1,  s11rx, s12rx, s21rx, s22rx);
    case 2
        [ s11out, s12out, s21out, s22out ]=combines4p( chdata.sdd11_raw, chdata.sdd21_raw, chdata.sdd12_raw, chdata.sdd22_raw, s11rx, s12rx, s21rx, s22rx);
end

function [hisi,tap_coef,hisi_ref]=applyDFEbk(hisi,hisi_ref,idx,tap_bk,curval,bmaxg,dfe_delta)
% hrem=applyDFEbk(hsis,idx,tap_bk,bmaxg)
% applying a bank of DFE at desired location
% hisi: waveform with cursor values;
% idx: starting index of the bank;
% tap_bk: number of taps in bank;
% bmaxg: maxmum coefficient in bank;

if nargin<6
    dfe_delta=0;
end

rng=idx:idx+tap_bk-1;
flt_curval=hisi(rng);

%floor(abs(floatingcursors/sbr(cursor_i))./param.dfe_delta).*param.dfe_delta.*sign(floatingcursors)*sbr(cursor_i);

if dfe_delta~=0
    flt_curval_q=floor(abs(flt_curval/curval)./dfe_delta) .* ...
        dfe_delta.*sign(flt_curval)*curval;
else
    flt_curval_q=hisi(rng);
end

tap_coef=min(abs(flt_curval_q/curval),bmaxg).*sign(flt_curval_q);
hisi(rng)= hisi(rng) - curval*tap_coef;
hisi_ref(rng)=0;

%AJG021820
function [ a] = bessel( n )
% bessel polynomial
for ii= 0:n
    a(ii+1) = factorial(2*n-ii) /  (2^(n-ii)*factorial(ii)*factorial(n-ii));
end


function [return_struct]= capture_RIL_RILN(chdata)
%Definition:
%This function captures the reflectionless insertion loss (RIL) and reflective insertion loss nois (RILN) for any arbitary S-parameter

%Author:
%Hansel Dsilva (dsilvahansel@gmail.com or hanseldsilva@achronix.com)

%This function has been shared by Hansel for others to evaluate for reflectionless insertion loss (RIL) and reflective insertion loss nois (RILN) for any arbitary S-parameter.

%INPUT: Struct passes to the function has the following,
% SCH.Frequencies= faxis;
% SCH.Parameters(1,1,:)= sdd11;
% SCH.Parameters(2,2,:)= sdd22;
% SCH.Parameters(1,2,:)= sdd12;
% SCH.Parameters(2,1,:)= sdd21;
% SCH.NumPorts= 2;
% SCH.Impedance= 100;

%OUTPUT: Struct returned has the following,
%return_struct.RIL_dB		%Reflectionless Insertion Loss in decibel
%return_struct.RILN_dB		%Reflective Insertion Loss Noise in decibel
%return_struct.Z_port1		%Frequency dependent, complex values of impedance for termination of port 1
%return_struct.Z_port2		%Frequency dependent, complex values of impedance for termination of port 2
%return_struct.rho_port1		%Frequency dependent, complex values of the reflection coefficient of port 1
%return_struct.rho_port2		%Frequency dependent, complex values of reflection coefficient of port 2
%return_struct.freq		%Frequency axis


%History,
%Rev0.5 (12/04/2019)
%- Initial release.


%---start. allowed is only a 2-port network having a transmitter and receiver

SCH.Parameters(1,1,:)=chdata(1).sdd11_orig;
SCH.Parameters(2,2,:)=chdata(1).sdd22_orig;
SCH.Parameters(1,2,:)=chdata(1).sdd12_orig;
SCH.Parameters(2,1,:)=chdata(1).sdd21_orig;
SCH.Frequencies=chdata(1).faxis;
SCH.Impedance=100;
SCH.NumPorts= 2;

if size(SCH.Parameters, 1)~=2
    fprintf('Size of given S matrix is %d.', size(SCH.Parameters, 3) );
    error('Allowed is only a 2-port network having a transmitter and receiver.');
end
%---end. allowed is only a 2-port network having a transmitter and receiver

%---start. do not include the DC point given sinusoidals at DC are not
%defined
if SCH.Frequencies(1)==0
    idx_start= 2;
else
    idx_start= 1;
end
%---end. do not include the DC point given sinusoidals at DC are not
%defined

Sdd11= squeeze(SCH.Parameters(1,1,idx_start:end));
Sdd12= squeeze(SCH.Parameters(1,2,idx_start:end));
Sdd21= squeeze(SCH.Parameters(2,1,idx_start:end));
Sdd22= squeeze(SCH.Parameters(2,2,idx_start:end));

a= -Sdd22+ Sdd11.*Sdd22.*conj(Sdd11)- Sdd21.*Sdd12.*conj(Sdd11);
b= 1+ Sdd22.*conj(Sdd22)+...
    Sdd11.*Sdd22.*conj(Sdd21).*conj(Sdd12)-...
    Sdd21.*Sdd12.*conj(Sdd21).*conj(Sdd12)-...
    Sdd11.*Sdd22.*conj(Sdd11).*conj(Sdd22)+...
    Sdd12.*Sdd21.*conj(Sdd11).*conj(Sdd22)-...
    Sdd11.*conj(Sdd11);
c= -conj(Sdd22)-...
    Sdd11.*conj(Sdd21).*conj(Sdd12)+...
    Sdd11.*conj(Sdd11).*conj(Sdd22);

solution_1= (-b+sqrt((b.^2)-(4*a.*c)))./(2*a);
solution_2= (-b-sqrt((b.^2)-(4*a.*c)))./(2*a);

clear a b c


%---start. Given the real part of the impedance is to be positive
if all(abs(solution_1(idx_start:end))< 1) && ~all(abs(solution_2(idx_start:end))< 1)
    rho_port2= solution_1;
elseif all(abs(solution_2(idx_start:end))< 1) && ~all(abs(solution_1(idx_start:end))< 1)
    rho_port2= solution_2;
else
    rho_port2= solution_1;
    %     error('Please contact the tool developer. It appears a odd case has appeared.');
end
%---end. For passive networks the reflection coefficient should be less
%than one


rho_port1= conj(Sdd11+ ((rho_port2.*Sdd21.*Sdd12)./(1-(rho_port2.*Sdd22))));


%---start. calculate the equivalent port impedance
Z_port1= (rho_port1*conj(SCH.Impedance)+ SCH.Impedance)./(1-rho_port1);
Z_port2= (rho_port2*conj(SCH.Impedance)+ SCH.Impedance)./(1-rho_port2);
%---end. calculate the equivalent port impedance


%---start. The reflectionless insertion loss is the insertion loss corresponding
%to zero reflections.
RIL= conj((1- rho_port1).*(sqrt(1- rho_port1.*conj(rho_port1))./abs(1-rho_port1)  )).*(  Sdd21.*(1- rho_port2.*Sdd22)+ (Sdd22- conj(rho_port2)).*rho_port2.*Sdd21  )./ ( ((1- rho_port2).*(sqrt(1- rho_port2.*conj(rho_port2))./abs(1-rho_port2)  )) .* (  -rho_port1.*rho_port2.*Sdd21.*Sdd12+  (1- rho_port2.*Sdd22).*(1- rho_port1.*Sdd11)  ) );
%---end. The reflectionless insertion loss is the insertion loss corresponding
%to zero reflections.


%---start. Calculate RILN (Reflective Insertion Loss Noise)
RILN= 20*log10(abs(Sdd21))- 20*log10(abs(RIL));
%---end. Calculate RILN (Reflective Insertion Loss Noise)
if SCH.Frequencies(1)==0 % added by Hansel 4/21/2021
    RIL= [NaN; RIL];
    RILN= [NaN; RILN];
    Z_port1= [NaN; Z_port1];
    Z_port2= [NaN; Z_port2];
    rho_port1= [NaN; rho_port1];
    rho_port2= [NaN; rho_port2];
end


%---start.  preparing returns struct
return_struct.RIL= RIL;		%Reflectionless Insertion Loss
return_struct.RIL_dB= 20*log10(abs(RIL));		%Reflectionless Insertion Loss
return_struct.RILN_dB= RILN;		%Reflective Insertion Loss Noise
return_struct.Z_port1= Z_port1;		%Frequency dependent, complex values of impedance for termination of port 1
return_struct.Z_port2= Z_port2;		%Frequency dependent, complex values of impedance for termination of port 2
return_struct.rho_port1= rho_port1;		%Frequency dependent, complex values of the reflection coefficient of port 1
return_struct.rho_port2= rho_port2;		%Frequency dependent, complex values of reflection coefficient of port 2
return_struct.freq= SCH.Frequencies(1:end);		%Frequency axis change index to start a 1 by Hansel 4/21/2021
%---end. preparing returns struct

function [noise_bottom,noise_top]=cdf_to_ber_contour(cdf,specBER)


%For the given BER, find the top & bottom voltage level in the CDF

%for the top, just find the first index at the spec BER
nidx=find(cdf.y>specBER, 1, 'first');
noise_bottom = cdf.x(nidx);
%for top, flip the cdf.  need to operate on row vector for fliplr:  cdf.y(:)'
nidx=find(fliplr(cdf.y(:)')>specBER, 1, 'first');
%the true index without flipping
nidx=length(cdf.y)-nidx+1;
noise_top = cdf.x(nidx);
function p=comb_fct(p1, p2)
if p1.BinSize ~= p2.BinSize
    error('bin size must be equal')
end

p=p1;
p.BinSize=p1.BinSize;
%p.Min=p1.Min+p2.Min;
p.Min=min(p1.Min,p2.Min);
difsz=abs(p1.Min-p2.Min);
lp1=length(p1.y);
lp2=length(p2.y);
if p1.Min == p.Min
    p2.y(difsz+1:lp2+difsz)=p2.y;
    p2.y(1:difsz)=0;
    p2.y(lp2+difsz+1:lp1)=0;
elseif p2.Min == p.Min
    p1.y(difsz+1:lp1+difsz)=p1.y;
    p1.y(1:difsz)=0;
    p1.y(lp1+difsz+1:lp2)=0;
end
% p.y=conv(p1.y, p2.y);
p.y=(p1.y+p2.y);
% p.y=p.y/sum(p.y);
% p.x =p.Min*p.BinSize:p.BinSize:-p.Min*p.BinSize;
p.x =(p.Min:-p.Min)*p.BinSize;	% modified by Yasuo Hidaka, 9/4/2016


function out_pdf=combine_pdf_same_voltage_axis(pdf1,pdf2)

if pdf1.BinSize ~= pdf2.BinSize
    error('bin size must be equal')
end

x1=pdf1.x;
y1=pdf1.y;
x2=pdf2.x;
y2=pdf2.y;
%find the pdf with a larger min, force it to have the same min, and insert
%probability = 0
min1=pdf1.x(1);
min2=pdf2.x(1);
shift_amount=round(abs(min1-min2)/pdf1.BinSize);
if min1<min2
    x2=[pdf1.x(1:shift_amount) pdf2.x];
    y2=[zeros(1,shift_amount) pdf2.y];
else
    x1=[pdf2.x(1:shift_amount) pdf1.x];
    y1=[zeros(1,shift_amount) pdf1.y];
end
%find the pdf with smaller max, force it to have the same max, and insert
%probability=0
L1=length(x1);
L2=length(x2);
Ldiff=abs(L1-L2);
if L1>L2
    out_x=x1;
    y2=[y2 zeros(1,Ldiff)];
else
    out_x=x2;
    y1=[y1 zeros(1,Ldiff)];
end
%now the 2 pdfs have the same voltage axis, add probabilities together
%renormalization is not handled here, so the output pdf will not have sum=1
%It is the responsibility of the calling function to handle renormalization
%if needed
out_y=y1+y2;
out_pdf.x=out_x;
out_pdf.y=out_y;
out_pdf.BinSize=pdf1.BinSize;
function [ s11out, s12out, s21out, s22out ] = combines4p( s11in1, s12in1, s21in1, s22in1, s11in2, s12in2, s21in2, s22in2)

s1=zeros(2,2,length(s11in1)); s2=s1; t3=s1;
for i=1:length(s11in1)
    s1(:,:,i)=[s11in1(i) s12in1(i); s21in1(i) s22in1(i) ];
    s2(:,:,i)=[s11in2(i) s12in2(i); s21in2(i) s22in2(i) ];
end
t1=stot(s1);
t2=stot(s2);
for i=1:length(s11in1)
    t3(:,:,i)=t1(:,:,i)*t2(:,:,i);
end
s3=ttos(t3);
s11out=s3(1,1,:);
s11out=transpose(s11out(:));
s12out=s3(1,2,:);
s12out=transpose(s12out(:));
s21out=s3(2,1,:);
s21out=transpose(s21out(:));
s22out=s3(2,2,:);
s22out=transpose(s22out(:));

function p=conv_fct(p1, p2)
if p1.BinSize ~= p2.BinSize
    error('bin size must be equal')
end

p=p1;
%p.BinSize=p1.BinSize;
%p.Min=p1.Min+p2.Min;
p.Min=round(p1.Min+p2.Min);	% modified by Yasuo Hidaka, 9/4/2016
p.y=conv2(p1.y, p2.y);
%p.x =p.Min*p.BinSize:p.BinSize:-p.Min*p.BinSize;
%p.x =(p.Min:-p.Min)*p.BinSize;	% modified by Yasuo Hidaka, 9/4/2016
pMax=p.Min+length(p.y)-1;
p.x =(p.Min*p.BinSize:p.BinSize:pMax*p.BinSize);




function p=conv_fct_MeanNotZero(p1, p2)

if p1.BinSize ~= p2.BinSize
    error('bin size must be equal')
end

p=p1;
%p.BinSize=p1.BinSize;
%p.Min=p1.Min+p2.Min;
p.Min=round(p1.Min+p2.Min);	% modified by Yasuo Hidaka, 9/4/2016
p.y=conv2(p1.y, p2.y);

%This is equivalent to (p.Min:p.Min+length(p.y)-1)*p.BinSize
%But it is faster to pre-multiply BinSize instead of multiplying the entire
%vector by BinSize
pMax=p.Min+length(p.y)-1;
p.x =(p.Min*p.BinSize:p.BinSize:pMax*p.BinSize);
function pdf=d_cpdf( binsize, values, probs)
%  p=cpdf(type, ...)
%
% CPDF is a probability mass function for discrete distributions or an
% approxmation of a PDF for continuous distributions.
%
% cpdf is internally normalized so that the sum of probabilities is 1
% (regardless of bin size).

% Internal fields:
% Min: *bin number* of minimum value.
% BinSize: size of PDF bins. Bin center is the representative value.
% Vec: vector of probabilities per bin.

%speed up for initializing empty pdf
if all(values==0)
    pdf.BinSize=binsize;
    pdf.Min=0;
    pdf.y=1;
    pdf.x=0;
    return;
end

if ~issorted(values)
    [values,si]=sort(values);
    probs=probs(si);
end
values=binsize*round(values/binsize);
t=(values(1):binsize:values(end));
pdf.Min=values(1)/binsize;
pdf.y=zeros(size(t));
for k=1:length(values)
    if k==1
        bin=1;
    elseif k==length(values)
        bin=length(t);
    else
        [UNUSED_OUTPUT, bin]=min(abs(t-values(k))); %#ok<ASGLU>
    end
    pdf.y(bin) = pdf.y(bin)+probs(k);
end

pdf.BinSize=binsize;
pdf.y=pdf.y/sum(pdf.y);
if any(~isreal(pdf.y)) || any(pdf.y<0)
    error('PDF must be real and nonnegative');
end
support=find(pdf.y);
pdf.y=pdf.y(support(1):support(end));
pdf.Min=pdf.Min+(support(1)-1);
pdf.x=(pdf.Min:-pdf.Min)*binsize;
function clip_output=dfe_clipper(input,max_threshold,min_threshold)

if isrow(input)
    max_threshold=max_threshold(:).';
    min_threshold=min_threshold(:).';
else
    max_threshold=max_threshold(:);
    min_threshold=min_threshold(:);
end

clip_output=input;
clip_output(input>max_threshold)=max_threshold(input>max_threshold);
clip_output(input<min_threshold)=min_threshold(input<min_threshold);


function [msg] = end_display_control(msg,param,OP,output_args,COM,min_ERL,ERL,VEO_mV,VEC_dB,threshold_DER,DISPLAY_WINDOW)
[ncases, mele]=size(param.z_p_tx_cases);
if mele ==2
    param.flex=2;
elseif mele==4
    param.flex=4;
elseif mele==1
    param.flex=1;
else
    error(springf('config file syntax error'))
end


if DISPLAY_WINDOW && ~OP.RX_CALIBRATION
    % display bathtub curves in one axis per test case.
    %             h=findall(0, 'Name', 'COM results');
    if ~exist('h','var')
        msgtext = cell(1, length(OP.pkg_len_select));
        msgcolor = 'g';
    else
        msgtext=get(findobj(h, 'type', 'text'), 'string');
        msgcolor = get(h, 'color');
        close(h); % will be recreated
    end
    msgctr=size(msgtext,1)+1;
    if ~OP.ERL_ONLY
        switch OP.PHY
            case 'C2M'
                if VEO_mV >= param.Min_VEO && VEO_mV <= param.Max_VEO
                    msg=sprintf('%s: EH = %.3f mV (pass)\n', ...
                        msg, VEO_mV);
                else
                    msg=sprintf('%s: EH = %.3f mV (FAIL)\n', ...
                        msg, VEO_mV);
                    msgcolor = 'r';
                end
                
                if VEC_dB <= param.VEC_pass_threshold
                    msg=sprintf('%s: VEC = %.3f dB (pass)\n', ...
                        (msg), VEC_dB);
                else
                    msg=sprintf('%s: VEC = %.3f dB (FAIL)\n', ...
                        (msg), VEC_dB);
                    msgcolor = 'r';
                end
            case 'C2C'
                if COM >= param.pass_threshold
                    %                         msgtext{package_testcase_i}=sprintf('%s: COM = %.3f dB (pass)\n', ...
                    %                             msg, COM);
                    msg=sprintf('%s: COM = %.3f dB (pass)\n', ...
                        msg, COM);
                else
                    %                         msgtext{package_testcase_i}=sprintf('%s: COM = %.3f dB (FAIL)\n', ...
                    %                             msg, COM);
                    msg=sprintf('%s: COM = %.3f dB (FAIL)\n', ...
                        msg, COM);
                    msgcolor = 'r';
                end
                % begin yasuo patch 3/18/2019
                msg=sprintf('%s: DER = %.3e at COM threshold \n', msg, threshold_DER);
                % end yasuo patch
            case 'C2Mcom'
                if VEO_mV >= param.Min_VEO && VEO_mV <= param.Max_VEO
                    msg=sprintf('%s: EH = %.3f mV (pass)\n', ...
                        msg, VEO_mV);
                else
                    msg=sprintf('%s: EH = %.3f mV (FAIL)\n', ...
                        msg, VEO_mV);
                    msgcolor = 'r';
                end
                
                if VEC_dB <= param.VEC_pass_threshold
                    msg=sprintf('%s: VEC = %.3f dB (pass)\n', ...
                        (msg), VEC_dB);
                else
                    msg=sprintf('%s: VEC = %.3f dB (FAIL)\n', ...
                        (msg), VEC_dB);
                    msgcolor = 'r';
                end
                if COM >= param.pass_threshold
                    %                         msgtext{package_testcase_i}=sprintf('%s: COM = %.3f dB (pass)\n', ...
                    %                             msg, COM);
                    msg=sprintf('%s: COM = %.3f dB (pass)\n', ...
                        msg, COM);
                else
                    %                         msgtext{package_testcase_i}=sprintf('%s: COM = %.3f dB (FAIL)\n', ...
                    %                             msg, COM);
                    msg=sprintf('%s: COM = %.3f dB (FAIL)\n', ...
                        msg, COM);
                    msgcolor = 'r';
                end
                % begin yasuo patch 3/18/2019
                msg=sprintf('%s: DER = %.3e at COM threshold \n', msg, threshold_DER);
                % end yasuo patch
        end
    end
    if OP.ERL
        if ~isempty(ERL)
            if min_ERL >= param.ERL_pass_threshold
                %                             msgtext{package_testcase_i+1}=[sprintf(' PASS ... ERL = %.3f dB\n', min_ERL)];
                msg=[sprintf('%s: PASS ... ERL = %.3f dB (%.3f dB,%.3f dB) \n',msg, min_ERL, ERL)];
            else
                %                             msgtext{package_testcase_i+1}=[ sprintf(' FAIL ... ERL = %.3f dB\n', min_ERL) ];
                msg=[ sprintf('%s: FAIL ... ERL = %.3f dB (%.3f dB,%.3f dB) \n',msg, min_ERL, ERL) ];
                msgcolor = 'r';
            end
        end
    end
    h=msgbox(msg, ['COM r' output_args.code_revision ' results']);
    set(h, 'color', msgcolor, 'tag', 'COM');
    movegui(h, 'center');
else % no windows
    %     display(['max noise at BER = ' num2str(peak_interference_at_BER)])
    %     display(['signal after eq = ' num2str(A_s/(param.levels-1))])
    if ~OP.ERL_ONLY
        switch OP.PHY
            case 'C2C'
                if COM >= param.pass_threshold
                    fprintf('%s <strong> PASS ... COM = %.3f dB</strong>\n', msg, COM);
                else
                    fprintf(2,'%s <strong> FAIL ... COM = %.3f dB</strong>\n',  msg, COM);
                end
                % begin yasuo patch 3/18/2019
                fprintf('%s DER = %.3e at COM threshold \n', msg, threshold_DER);
                % end yasuo patch
            case 'C2Mcom'
                if VEC_dB <= param.VEC_pass_threshold
                    fprintf('%s <strong> PASS ... VEC = %.3f dB</strong>\n', msg, VEC_dB);
                else
                    fprintf(2,'%s <strong> FAIL ... VEC = %.3f dB</strong>\n',  msg, VEC_dB);
                end
                
                if VEO_mV >= param.Min_VEO && VEO_mV <= param.Max_VEO
                    fprintf('%s <strong> PASS ... EH = %.3f mV</strong>\n', msg, VEO_mV);
                else
                    fprintf(2,'%s <strong> FAIL ... EH = %.3f mV</strong>\n',  msg, VEO_mV);
                end
                if COM >= param.pass_threshold
                    fprintf('%s <strong> PASS ... COM = %.3f dB</strong>\n', msg, COM);
                else
                    fprintf(2,'%s <strong> FAIL ... COM = %.3f dB</strong>\n',  msg, COM);
                end
                % begin yasuo patch 3/18/2019
                fprintf('%s DER = %.3e at COM threshold \n', msg, threshold_DER);
                % end yasuo patch
            case 'C2M'
                if VEC_dB <= param.VEC_pass_threshold
                    fprintf('%s <strong> PASS ... VEC = %.3f dB</strong>\n', msg, VEC_dB);
                else
                    fprintf(2,'%s <strong> FAIL ... VEC = %.3f dB</strong>\n',  msg, VEC_dB);
                end
                
                if VEO_mV >= param.Min_VEO && VEO_mV <= param.Max_VEO
                    fprintf('%s <strong> PASS ... EH = %.3f mV</strong>\n', msg, VEO_mV);
                else
                    fprintf(2,'%s <strong> FAIL ... EH = %.3f mV</strong>\n',  msg, VEO_mV);
                end
        end
    end
    if OP.ERL
        if ~isempty(ERL)
            if min_ERL >= param.ERL_pass_threshold
                %                             msgtext{package_testcase_i+1}=[sprintf(' PASS ... ERL = %.3f dB\n', min_ERL)];
                fprintf('%s: <strong> PASS ... ERL = %.3f dB (%.3f dB, %.3f dB)</strong>\n',msg, min_ERL, ERL );
            else
                %                             msgtext{package_testcase_i+1}=[ sprintf(' FAIL ... ERL = %.3f dB\n', min_ERL) ];
                fprintf(2,'%s:  <strong> FAIL ... ERL = %.3f dB (%.3f dB, %.3f dB)</strong>\n',msg, min_ERL, ERL) ;
            end
        end
    end
end

function [Left_EW,Right_EW]=find_eye_width(eye_contour,half_UI,samples_per_UI,vref)

%Left eye Width (Top Eye)
left_top=eye_contour(half_UI:-1:1,1);
%vref_crossing is the first point less than vref (usually first point < 0)
vref_crossing=find(left_top<vref,1,'first');
if isempty(vref_crossing)
    %this case handles completely open eye
    L1=half_UI;
elseif vref_crossing==1
    %this case handles completely closed eye
    L1=0;
else
    %this case handles the normal eye
    %INT is a linear interpolation between the 2 points on either side of
    %vref to determine where the vref crossing occurred.  In systems with
    %a small number of samples_per_UI, interpolation improves accuracy over
    %just using the integer sample point
    INT=vref_intersect(eye_contour(1:half_UI,1),half_UI-vref_crossing+1+1,vref);
    L1=half_UI-INT;
end
%Left eye Width (Bottom Eye)
left_bot=eye_contour(half_UI:-1:1,2);
vref_crossing=find(left_bot>vref,1,'first');
if isempty(vref_crossing)
    L0=half_UI;
elseif vref_crossing==1
    L0=0;
else
    INT=vref_intersect(eye_contour(1:half_UI,2),half_UI-vref_crossing+1+1,vref);
    L0=half_UI-INT;
end
%Right eye Width (Top Eye)
right_top=eye_contour(half_UI:end,1);
vref_crossing=find(right_top<vref,1,'first');
if isempty(vref_crossing)
    R1=samples_per_UI-half_UI;
elseif vref_crossing==1
    R1=0;
else
    INT=vref_intersect(eye_contour(half_UI:end,1),vref_crossing,vref)+half_UI-1;
    R1=INT-half_UI;
end
%Right eye Width (Bottom Eye)
right_bot=eye_contour(half_UI:end,2);
vref_crossing=find(right_bot>vref,1,'first');
if isempty(vref_crossing)
    R0=samples_per_UI-half_UI;
elseif vref_crossing==1
    R0=0;
else
    INT=vref_intersect(eye_contour(half_UI:end,2),vref_crossing,vref)+half_UI-1;
    R0=INT-half_UI;
end

%L1 = top left eye width
%L0 = bottom left eye width
%Left eye width is the minimum
%R1 = top right eye width
%R0 = bottom right eye width
%Right eye width is the minimum
Left_EW=min([L1 L0]);
Right_EW=min([R1 R0]);

function [idx]=findbankloc(hisi,idx_st,idx_en,tap_bk,curval,bmaxg)
% [idx]=findbankloc(hisi,idx_st,idx_en,tap_bk)
% find the location of the DFE bank
% hisi: waveform with cursor values;
% idx_st: starting index;
% idx_en: ending index ;
% tap_bk: number of taps per bank;
% bmaxg: maximum coefficient;

hisi=hisi(:);
len=idx_en-idx_st+1;
h0=abs(hisi(idx_st:idx_en));
h1=max(0,h0-bmaxg*curval);

h0n=zeros(len-tap_bk+1,1);
h1n=h0n;

for ii=1:tap_bk
    h0tmp=h0(ii:ii+len-tap_bk);
    h0n=h0n+h0tmp.^2;
    h1tmp=h1(ii:ii+len-tap_bk);
    h1n=h1n+h1tmp.^2;
end

ndiff=h0n-h1n;
[M,idx] = max(ndiff);
idx=idx+idx_st-1;




function [tap_loc,tap_coef,hisi,b]=floatingDFE( hisi,  N_b,      N_bf,   N_bg, N_bmax,       bmaxg,  curval,  dfe_delta  )

% Adam Gregroy modifications of floatingDFE, Jingbo Li 5/16/2019
% COOP = 1 co-optimize banks , -0 sequenatial optmization ; remove

% idx_st=20;    % N_b number of fixed dfe taps
% nb_bk=4;      % N_bg number of floating tap groups. 1 2 or 3 right now
% tap_bk=6;     % N_bf number of floating taps per group
% curval=1;     % hsis(cursor_i)
% bmaxg=0.3;    % bmaxg limit for the floating taps
% dfe_delta=0;  % param.dfe_delta need to pass
% idx_en=90;    % N_bmax number of ui for the max reach of the floating taps

if nargin<8, dfe_delta=0;end

%new code making the entire vector of dfe taps'
tap_loc=zeros(1,N_bg*N_bf);
bank_loc=zeros(1,N_bg);
tap_coef=zeros(1,length(hisi));
b=zeros(1,length(hisi));
%hisi_ref is used for finding dfe taps.  whenever a bank is applied, those
%locations in hisi_ref are set to exactly 0.  This is to fix an issue where
%small bmaxg can create a situation where the same location may be used
%more than once even after DFE is applied to hisi.
hisi_ref=hisi;
for ii=1:N_bg
    % [ahealey] The start index for the floating tap search should be the
    % tap *following* the last fixed tap.
    bank_loc(ii)=findbankloc(hisi_ref,N_b+1,N_bmax,N_bf,curval,bmaxg);
    % bank_loc(ii)=findbankloc(hisi_ref,N_b,N_bmax,N_bf,curval,bmaxg);
    % [ahealey] End of modification.
    %don't need tap_loc anymore, but still populating it
    tap_loc((ii-1)*N_bf+1:ii*N_bf)=bank_loc(ii):bank_loc(ii)+N_bf-1;
    
    [hisi,coef,hisi_ref]=applyDFEbk(hisi,hisi_ref,bank_loc(ii),N_bf,curval,bmaxg,dfe_delta);
    tap_coef(bank_loc(ii):bank_loc(ii)+N_bf-1)=coef;
end
tap_loc=sort(tap_loc,'ascend');
b(tap_loc)=bmaxg;



function [ bmax floating_tap_locations] = floating_taps_1sttest( hisi,N_b,N_bf,N_bg,N_bmax, bmaxg, COOP )
% Richard Mellitz:  04/23/2019
% hisi is the isi 1 ui/sample
% N_b number of fixed dfe taps
% N_bf number of floating taps per group
% N_bg number of floating tap groups. 1 2 or 3 right now
% N_bmax number of ui for the max reach of the floating taps
% bmaxg limit for the floating taps
% COOP = 1 co-optimize banks , -0 sequenatial optmization
%
%
% function to remove isi or add noise above bmaxg
if ~exist('COOP','var'), COOP=0;end
if iscolumn(hisi); hisi=hisi.';end
hsis_in=hisi;
% find all the reduction group taken N_bf at a time
% we are looking for the group when when remove yield the miminim isi, h, power
best_sigma=inf;best_ig1=-1;best_ig2=-1;best_ig3=-1;
% add on switch and loop for each potential group
switch N_bg
    case 0
        bmax=0;
        return
    case 1
        end1=N_bmax-N_bf;
        end2=N_b+1;
        end3=N_b+1;
    case 2
        end1=N_bmax-N_bf;
        end2=N_bmax-N_bf;
        end3=N_b+1;
    case 3
        end1=N_bmax-N_bf;
        end2=N_bmax-N_bf;
        end3=N_bmax-N_bf;
end
if COOP
    for ig1= N_b+1:end1 % now remove the 2nd group
        hcap=  hrem(hisi,ig1,N_bf,bmaxg)  ;
        % loop for 2rd group
        for ig2= N_b+1: end2
            hcap2= hrem(hcap,ig2,N_bf,bmaxg)  ;
            if N_bg < 2; hcap2 =hcap; end
            for ig3= N_b+1: end3
                hcap3= hrem(hcap2,ig3,N_bf,bmaxg)  ;
                if N_bg < 3 ; hcap3=hcap2 ; end
                sigma=norm( hcap3 );
                if sigma < best_sigma
                    best_sigma=sigma;
                    best_ig1=ig1;
                    best_ig2=ig2;
                    best_ig3=ig3;
                    best_hcap=hcap3;
                end
            end
        end
    end
else % sequentail
    for ig1= N_b+1:end1 % now remove the 1st group
        hcap=  hrem(hisi,ig1,N_bf,bmaxg)  ;
        sigma=norm( hcap );
        if sigma < best_sigma
            best_sigma=sigma;
            best_ig1=ig1;
            best_hcap=hcap;
        end
    end
    % loop for 2rd group
    hisi=best_hcap;
    for ig2= N_b+1: end2
        hcap= hrem(hisi,ig2,N_bf,bmaxg)  ;
        sigma=norm( hcap );
        if sigma < best_sigma
            best_sigma=sigma;
            best_ig2=ig2;
            best_hcap=hcap;
        end
    end
    hisi=best_hcap;
    % loop for 3rd group
    for ig3= N_b+1: end3
        hcap= hrem(hisi, ig3,N_bf,bmaxg)  ;
        sigma=norm( hcap );
        if sigma < best_sigma
            best_sigma=sigma;
            best_ig3=ig3;
            best_hcap=hcap;
        end
    end
    
end
bmax(N_b+1:N_bmax)=zeros(1,N_bmax-N_b);
switch N_bg
    case 1
        bmax(best_ig1:best_ig1+N_bf-1)=ones(1,N_bf)*bmaxg;
        floating_tap_locations= [best_ig1:best_ig1+N_bf-1];
    case 2
        bmax(best_ig1:best_ig1+N_bf-1)=ones(1,N_bf)*bmaxg;
        bmax(best_ig2:best_ig2+N_bf-1)=ones(1,N_bf)*bmaxg;
        floating_tap_locations= [best_ig1:best_ig1+N_bf-1 best_ig2:best_ig2+N_bf-1 ];
    case 3
        bmax(best_ig1:best_ig1+N_bf-1)=ones(1,N_bf)*bmaxg;
        bmax(best_ig2:best_ig2+N_bf-1)=ones(1,N_bf)*bmaxg;
        bmax(best_ig3:best_ig3+N_bf-1)=ones(1,N_bf)*bmaxg;
        floating_tap_locations= [best_ig1:best_ig1+N_bf-1 best_ig2:best_ig2+N_bf-1 best_ig3:best_ig3+N_bf-1 ];
end
floating_tap_locations=sort(floating_tap_locations);
if 0 % for code debug
    close force all
    stem(best_hcap,'disp','hcap')
    hold on
    stem(bmax,'-k','disp','bmax')
    stem(hisi,'disp','hisi')
    hold off
end

% function [ bmax floating_tap_locations] = floating_taps( hisi,   N_b,     N_bf,  N_bg,  N_bmax,       bmaxg. COO))
function [ Vfiltered, Cmod ] = force( V ,param, OP , ix, C)
% Vfilter is vector forced filtered sbr
% Cmod is the ffe tap co-efficient vector
% if C is passed, just process V with C else compute C
% cmx=param.rx_cmx; number of pre cursor taps
% cpx=param.rx_cps; number of post cursor taps
% V=sbr; pass pulse response
% ix the sample point in the passed pulse response
% the sample point is recomputed by optimize_fom
% OP not used for now
% test with load('SBR_FIR_resp.mydata','-mat')
%% set up parameters ------------------------------------------------------
if ~exist('cursor_gain','var'), cursor_gain=0; end
csm = @(b,n) [ b((length(b)-n)+1:length(b)) b(1:(length(b)-n))]; % minus circshift
csp = @(b,n) [ b(n:length(b)) b(1:n-1) ]; %  plus circshift
db = @(x) 20*log10(abs(x));
if ~exist('ix','var'),ix=find(V==max(V),1,'first');end
cmx=param.RxFFE_cmx;
cpx=param.RxFFE_cpx;
cstep=param.RxFFE_stepz;
ndfe=param.ndfe;
spui=param.samples_per_ui;
%% resample to align with sample poont ------------------------------------
%% only consider sample aligned with sample point
% if ix < length(V);
%     if isrow(V)
%         vsampled_raw = [V(mod(ix,spui)+1:spui:(mod(ix,spui)+spui*(floor(ix/spui)-1)))'; V(ix:spui:end)'];
%     else
%         vsampled_raw = [V(mod(ix,spui)+1:spui:(mod(ix,spui)+spui*(floor(ix/spui)-1))); V(ix:spui:end)];
%     end
% else
%     if isrow(V)
%         vsampled_raw = V(mod(ix,spui)+1:spui:end)';
%     else
%         vsampled_raw = V(mod(ix,spui)+1:spui:end) ;
%     end
% end
if ix < length(V);
    if isrow(V)
        %         vsampled_raw = [V(mod(ix,spui):spui:(mod(ix,spui)+spui*(floor(ix/spui)-1)))'; V(ix:spui:end)'];
        if mod(ix,spui) == 0
            vsampled_raw = [V(spui+mod(ix,spui):spui:(mod(ix,spui)+spui*(floor(ix/spui)-1)))'; V(ix:spui:end)'];
        else
            vsampled_raw = [V(mod(ix,spui):spui:(mod(ix,spui)+spui*(floor(ix/spui)-1)))'; V(ix:spui:end)'];
        end
        
    else
        %         vsampled_raw = [V(mod(ix,spui):spui:(mod(ix,spui)+spui*(floor(ix/spui)-1))); V(ix:spui:end)];
        if mod(ix,spui) == 0
            vsampled_raw = [V(spui+mod(ix,spui):spui:(mod(ix,spui)+spui*(floor(ix/spui)-1))); V(ix:spui:end)];
        else
            vsampled_raw = [V(mod(ix,spui):spui:(mod(ix,spui)+spui*(floor(ix/spui)-1))); V(ix:spui:end)];
        end
    end
else
    if isrow(V)
        %        vsampled_raw = V(mod(ix,spui):spui:end)'; %Yasou Hidaka 11/16/2018
        if mod(ix,spui) == 0%Yasou Hidaka 11/16/2018
            vsampled_raw = V(spui+mod(ix,spui):spui:end)';%Yasou Hidaka 11/16/2018
        else
            vsampled_raw = V(mod(ix,spui):spui:end)';%Yasou Hidaka 11/16/2018
        end
    else
        %        vsampled_raw = V(mod(ix,spui):spui:end) ;%Yasou Hidaka 11/16/2018
        if mod(ix,spui) == 0%Yasou Hidaka 11/16/2018
            vsampled_raw = V(spui+mod(ix,spui):spui:end) ;
        else
            vsampled_raw = V(mod(ix,spui):spui:end) ;%Yasou Hidaka 11/16/2018
        end
    end
    
    
end

vsampled=[zeros(1,cmx) vsampled_raw' zeros(1,cpx)];% pad for pre and post cursor prior to shifting
%% create VV matrix of shifted UI spaced sample of the pulse response
ishift=cmx+1;
for i=1:cmx
    ishift=ishift-1;
    VV(i,:)=circshift(vsampled,[0,-ishift]);
end
VV(cmx+1,:)=vsampled;
ishift=0;
for i=1:cpx
    ishift=ishift+1;
    VV(i+cmx+1,:)=circshift(vsampled,[0,ishift]);
end
% find the index for the sample point but in the UI resample vector, vsampled
% ivs=find(vsampled==V(ix),1,'first');% ivs is the sample point for V
%% Upen Kareti suggested fix for indexing 11/04/18
if ix < length(V)
    ivs=find(vsampled==V(ix),1,'first');% ivs is the sample point for V
else
    ivs=find(vsampled == max(vsampled),1,'first');
end
if ~exist('C','var')
    % cmx+1 is the cursor or sample point
    VV=VV(:,ivs-cmx:ivs+cpx); % only consider the VV matrix that correstonds to the FFE taps
    FV=zeros(1,cmx+cpx+1); % zero the forceing veotor, FV first
    FV(cmx+1)=vsampled(ivs)*10^(param.current_ffegain/20); % force the voltage at sample point
    if param.ndfe~=0 && (cpx > 0) % Yasuo Hidaka suggest fix for no postC 11/11/18
        %          FV(cmx+2)=param.bmax(1)*FV(cmx+1);   % force the post cursor to bmax if dfe exists
        FV(cmx+2)=min(param.bmax(1)*FV(cmx+1),abs(vsampled(ivs+1)))*sign(vsampled(ivs+1));
    end
    C=((VV'*VV)^-1*VV')'*FV'; % sikve for FFE taps, C
    if cstep ~= 0
        %         C=C/sum(C); % r240 constrain taps to sum of taps = 1
        C=C/C(cmx+1); % r241 make cursor tap 1
        Cmod=floor(abs(C/cstep)).*sign(C)*cstep;% r250 quantize with floor ad sign(C)
    else
        %         C=C/sum(C); % r240
        Cmod=C/C(cmx+1); % r241 constrain taps to sum of taps = 1
    end
    Cmod=Cmod(1:cmx+1+cpx);
else
    Cmod=C;%just us the FFE taps, C, passed for filtering
end
%% The DNH routing (do no hard) for long FFE i.e. see if zeroing taps helps
fom_ffe=0;
Cmod1=Cmod;
% for i=0:param.ffe_backoff
for i=0:min(param.ffe_backoff,cpx)
    Cmodtest=[ Cmod(1:1+cpx+cmx-i).' zeros( 1,i) ].';
    Vtest=FFE(Cmodtest , param.RxFFE_cmx,1, vsampled );
    Vs=Vtest(ivs);
    Vtest(ivs)=0;
    fom_ffe_test= db( Vs/norm(Vtest));
    if abs(fom_ffe_test) > abs(fom_ffe)
        fom_ffe=fom_ffe_test;
        Cmod1=Cmodtest; % modify Cmod1
    end
end
Cmod=Cmod1;
%%
%% filter the pulse response with the solved FFE
Vfiltered=FFE( Cmod , param.RxFFE_cmx,spui, V );






% figure(101010)
% plot(Vfiltered)
% hold on
% plot(V)
% hold off

function [ILN, efit]= get_ILN(sdd21,faxis_f2)
% used for FD IL fitting
% sdd21 us a complex insertion loss
db = @(x) 20*log10(abs(x));
sdd21=squeeze(sdd21);
if  iscolumn(sdd21)
    sdd21=sdd21.';
end
fmbg=[ones(length(faxis_f2),1).*transpose(abs(sdd21))  transpose(sqrt(faxis_f2)).*transpose(abs(sdd21))  transpose(faxis_f2).*transpose(abs(sdd21)) transpose(faxis_f2.^2).*transpose(abs(sdd21)) ];
warning('off','MATLAB:nearlySingularMatrix');
LGw=transpose(abs(sdd21).*db(sdd21));
alpha = ((fmbg'*fmbg)^-1)*fmbg'*LGw;
efit=(alpha(1)+alpha(2).*sqrt(faxis_f2)+alpha(3).*faxis_f2 +faxis_f2.^2.*alpha(4)   );
ILN = db(sdd21)-efit;

function result=get_PulseR(ir,param,cb_step,ZT)
%ir = impulse response
%t_base=time array with equal time steps
%samp_UI = number of samples per UI for ir

% t for debug
t=(1/param.fb)/param.samples_per_ui*(0:length(ir)-1);

if cb_step
    Ag=1;
    dt=1/param.fb/param.samples_per_ui;
    edge_time=param.TR_TDR*1e-9;
    fedge=1/edge_time;
    tedge=0:dt:edge_time*2;
    %
    edge=Ag*(2*cos(2*pi*(tedge)*fedge/16-pi/4).^2-1);
    drive_pulse=[edge ones(1,param.samples_per_ui)];
    %pulse=filter(UI_ones,1,ir);
    % t for debug
    t=(1/param.fb)/param.samples_per_ui*(0:length(ir)-1);
    
    pulse=filter(drive_pulse,1,ir);
else
    pulse=filter( ones(1,param.samples_per_ui),1,ir);
end
PDR_response=(1+pulse)./(1-pulse).*ZT*2;
result.PDR=PDR_response;
result.pulse=pulse;



function [ FIR t] =get_RAW_FIR(H,f,OP,param)
H_r = 1./polyval([1 2.613126 3.414214 2.613126 1], 1i*f./(0.75*param.fb));
if ~iscolumn(H), H=H.';end
if ~iscolumn(H_r), H_r=H_r.';end
H=H(:).*H_r;
[FIR, t, ~,~] = s21_to_impulse_DC(H ,f, param.sample_dt, OP) ;
% SBR=filter(ones(1, param.samples_per_ui), 1, FIR);

function result=get_StepR(ir,param,cb_step,ZT)
%ir = impulse response
%t_base=time array with equal time steps
%samp_UI = number of samples per UI for ir
%   result.SBR
% t for debug
t=(1/param.fb)/param.samples_per_ui*(0:length(ir)-1);

if cb_step
    Ag=1;
    dt=1/param.fb/param.samples_per_ui;
    edge_time=param.TR_TDR*1e-9;
    fedge=1/edge_time;
    tedge=0:dt:edge_time*2;
    %
    edge=Ag*(2*cos(2*pi*(tedge)*fedge/16-pi/4).^2-1);
    drive_pulse=[edge ones(1,param.samples_per_ui)];
    %pulse=filter(UI_ones,1,ir);
    
    pulse=filter(drive_pulse,1,ir);
else
    pulse=filter( ones(1,length(ir)),1,ir);
end
TDR_response=(1+pulse)./(1-pulse)*ZT*2;
result.ZSR=TDR_response;
result.pulse=pulse;


function TDR_results = get_TDR(sdd, OP, param,ZT,np)
% sdd is differential s-parameters structure (2 port assumed)
% input parameter structure for s parameters sdd--> sdd.Impedance, sdd.Frequencies, sdd.Parameters, sdd.NumPorts
% TDR_results.delay             pre t=0 delay for TDR... help with time domain responce quaility
% TDR_results.tdr               the TDR responce (ohms vs  TDR_results.t
% TDR_results.t                 starting at t=0
% TDR_results.tx_filter         transmitter filter vs TDR_results.f
% TDR_results.Rx_filter         receiver filter vs TDR_results.f
% TDR_results.f                 frequency for filter and s parameters
% TDR_results.ptdr_RL           reflection waveform from the pulse
% TDR_results.WC_ptdr_samples_t worst case time sample of the reflection pulse
% TDR_results.WC_ptdr_samples   worst case reflection samples of the reflection pulse
% TDR_results.ERL               reported effective return loss
%
db = @(x) 20*log10(abs(x));
rms =@(x) norm(x)/sqrt(length(x));
if isfield(OP,'TDR_duration')
    TDR_duration=OP.TDR_duration; % approximate transit time multipler
else
    TDR_duration=5;
end
if ~isfield(OP,'DISPLAY_WINDOW')
    OP.DISPLAY_WINDOW=1; % approximate transit time multipler
end
f=sdd.Frequencies;
TDR_results.f=f;
% OP.Zt_adj=2;
if param.FLAG.S2P == 0
    
    % re-normalize reference of s-parameterss: this seems correct for a s4p input file
    TDR_RL =@(Zin,Zout,s11,s12,s21,s22)(Zin.^2.*s11+Zin.^2.*s22+Zout.^2.*s11+Zout.^2.*s22+Zin.^2-Zout.^2+Zin.*Zout.*s11.*2.0-Zin.*Zout.*s22.*2.0+Zin.^2.*s11.*s22-Zin.^2.*s12.*s21-Zout.^2.*s11.*s22+Zout.^2.*s12.*s21)./(Zin.*Zout.*2.0+Zin.^2.*s11+Zin.^2.*s22-Zout.^2.*s11-Zout.^2.*s22+Zin.^2+Zout.^2+Zin.^2.*s11.*s22-Zin.^2.*s12.*s21+Zout.^2.*s11.*s22-Zout.^2.*s12.*s21-Zin.*Zout.*s11.*s22.*2.0+Zin.*Zout.*s12.*s21.*2.0);
    
    if param.RL_sel==1, other_port=2;end
    if param.RL_sel==2, other_port=1;end
    for i = 1:length(sdd.Frequencies)
        if size(sdd.Parameters,2) ==1 % for s2p files
            RL(i)=TDR_RL(sdd.Impedance,2*ZT,sdd.Parameters(param.RL_sel,param.RL_sel,i) ,1, 1 ,sdd.Parameters(param.RL_sel,param.RL_sel,i) );
        else
            RL(i)=TDR_RL(sdd.Impedance,2*ZT,sdd.Parameters(param.RL_sel,param.RL_sel,i) ,sdd.Parameters( 1,2 ,i), sdd.Parameters( 2,1 ,i),sdd.Parameters( other_port,other_port ,i) );
        end
    end
    % elseif OP.Zt_adj ==2 % only adjust z_t drive impedance
    %     RL=squeeze((sdd.Parameters(param.RL_sel,param.RL_sel,:)));
    %     Z_t=ZT;
    %     zref=sdd.Impedance/2;
    %     if Z_t > zref
    %         radjust=  (zref-Z_t);
    %         S11adjust= radjust./(radjust + 2*zref);
    %         RL=RL  +S11adjust;
    %     elseif Z_t < zref
    %         rpad=-Z_t*zref/(Z_t-zref);
    %         S11adjust=zref/(rpad*(zref/rpad + 2));
    %         RL=RL   + S11adjust;
    %     else
    %         RL=RL;
    %     end
else
    for i = 1:length(sdd.Frequencies)
        rho= (2*ZT - sdd.Impedance) / (2*ZT + sdd.Impedance);
        interim =  sqrt(1-abs(rho)^2) * (1 - rho) / abs(1-rho);
        RL(i) = interim \ (sdd.Parameters(param.RL_sel,param.RL_sel,i) - rho) / ...
            (1 - rho *sdd.Parameters(param.RL_sel,param.RL_sel,i) ) * interim;
    end
end

% end
RL=squeeze(RL);
f9=f/1e9;
tr=param.TR_TDR;
TDR_results.delay=500e-12 ;
% determine max time from thue
% if sdd.NumPorts==1
%     try
%         maxtime=OP.N*param.ui;
%     catch
%         maxtime=2e-9;
%     end
%     pix=1;
% else
%     [ fir4del, tu] =get_RAW_FIR(squeeze(sdd.Parameters(2,1,:)),f,OP,param);
%     pix=find(fir4del==max(fir4del),1);
%     maxtime=tu(pix)*TDR_duration+TDR_results.delay;
%     if maxtime > tu(end); maxtime=tu(end);end
% end

try
    maxtime=OP.N*param.ui;
catch
    maxtime=2e-9;
end
if OP.N==0
    if sdd.NumPorts==1
        fprintf('<strong> Warning for s2p files N must not be zero<\strong> ');
    else
        [ fir4del, tu] =get_RAW_FIR(squeeze(sdd.Parameters(2,1,:)),f,OP,param);
        pix=find(fir4del==max(fir4del),1);
        maxtime=tu(pix)*TDR_duration+TDR_results.delay;
        if maxtime > tu(end); maxtime=tu(end);end
    end
end


% add delay 500 ps for TDR and 3 times Gaussnan transtion time
% (makes gausian edge somewhat causal)
H_t = exp( -2*(pi*f9*(tr)/1.6832).^2 ).*exp(-(1j)*2*pi*f9*TDR_results.delay/1e-9).*exp(-1j*2*pi*f9*tr*3);
if ~isfield(OP,'cb_Guassian')
    Use_gaussian=1;
else
    Use_gaussian=OP.cb_Guassian;
end
if Use_gaussian
    if iscolumn(H_t), H_t=H_t.'; end
    RLf=RL(:).'.*H_t;
else % add extra 3x tr delay for causality
    RLf=RL.*exp(-(1j)*2*pi*f9*TDR_results.delay/1e-9);
end

if isfield(param,'f_r')
    param.fb_BW_cutoff=param.f_r;
end
% if OP.Bessel_Thomson
if 0
    factor_3db=0.473037;
    param.fb_BT_cutoff=factor_3db*param.f_r;
    a = bessel( param.BTorder );
    acoef=fliplr( a );
    H_bt =a(1)./ polyval(acoef, (1i*f./(param.fb_BT_cutoff*param.fb)));
else
    H_bt=ones(1,length(f));
end

if isfield(OP,'TDR_Butterworth')
    if OP.TDR_Butterworth
        H_bw = 1./polyval([1 2.613126 3.414214 2.613126 1], 1i*f./(param.fb_BW_cutoff*param.fb));
    else
        H_bw=ones(1,length(f));
    end
else
    H_bw=ones(1,length(f));
end


if param.Tukey_Window ~= 0
    H_tw= Tukey_Window(f,param);
else
    H_tw=ones(1,length(f));
end


if iscolumn(H_tw), H_tw=H_tw.';end
if iscolumn(H_bt), H_bt=H_bt.';end
if iscolumn(H_bw), H_bw=H_bw.';end
if iscolumn(RLf), RLf=RLf.';end

TDR_results.Rx_filter=H_bt.*H_bw.*H_tw;
RLf=RLf.*TDR_results.Rx_filter;
TDR_results.tx_filter=H_t;


[IR, t, causality_correction_dB, truncation_dB] = ...
    s21_to_impulse_DC(RLf,  sdd.Frequencies(:), param.sample_dt,OP);


%
% param.tfx =4.2e-10; % need to put in xls file and comment this out
tfx=param.tfx(np); % use fixture delay for port (np)

%  IR(abs(IR)<=OP.impulse_response_truncation_threshold)=0; % noise filter

t = t-TDR_results.delay;
tend=find(t>=maxtime+tfx,1); % n starts at tfx
if isempty(tend), tend=length(t); end
IR=IR(1:tend);
t=t(1:tend);
if isempty(tend), tend=length(t); end
tstart=find(t>=tr*1e-9,1); % account for Gaussian precursor
if isempty(tstart), tstart=1;end
if isempty(tend) || tstart >= tend
    if isempty(tend) || tstart >= tend
        %         warndlg('TDR compuation not valid, try decreasing truncation tolerance, increasing samples, or adding a transmisson line','WrnTDR');
    end
    tend=length(t);
    tstart=1;
end
OP.cb_step=0; % step is a basically a cos^2 form -pi/4 to 0. not activated.
ch=get_StepR(IR(tstart:tend),param,OP.cb_step,ZT);
TDR_results.tdr= ch.ZSR;
TDR_results.t = t(tstart:tend);

PTDR=get_PulseR(IR(tstart:tend),param,OP.cb_step,ZT);
if OP.TDR || OP.PTDR % determin average impededance with
    try
        tfstart=find(t>=3*tr*1e-9,1);
        x=squeeze(TDR_results.t(tfstart:end));TDR_results.x=TDR_results.tdr(:);
        y=squeeze(TDR_results.tdr(tfstart:end));TDR_results.y=TDR_results.t(:);
        w= exp(-(x-x(1))/OP.T_k ) ; % weighting function
        TDR_results.avgZport=mean(y.*w.')/mean(w.');
    catch
        TDR_results.avgZport=0;
        fit=zeros(1,1);
        p=[0 0 0 0  ];
    end
    TDR_results.RL=RL;
end
if OP.PTDR
    %     param.N_bx=param.ndfe;
    RL_equiv=-inf;
    L=param.levels;
    BinSize=OP.BinSize;
    %     param.specBER=1e-5;
    if OP.DISPLAY_WINDOW
        hwaitbar=waitbar(0);
    else
        fprintf('Worst ERL searching');
    end
    % adjust PTDR for NDFE
    % ---------------------- 2.7 code
    %     ntx=find(TDR_results.t >= tfx,1,'first');
    %     %     gatestartt=TDR_results.t(ntx);
    %     %     gatestartV=PTDR.pulse(ntx);
    %     ndfex=find(TDR_results.t > (param.N_bx+1)*param.ui+tfx,1,'first');
    %     tk=param.ui*1*(param.N_bx+1)+tfx;
    % -------------------
    % [ahealey 09/06/2019] Need to account for the 3*TR_TDR delay included in the rise
    % time filter.
    % ntx=find(TDR_results.t >= tfx,1,'first');
    ntx=find(TDR_results.t >= tfx+3*tr*1e-9,1,'first');
    %     gatestartt=TDR_results.t(ntx);
    %     gatestartV=PTDR.pulse(ntx);
    % ndfex=find(TDR_results.t > (param.N_bx+1)*param.ui+tfx,1,'first');
    % tk=param.ui*1*(param.N_bx+1)+tfx;
    ndfex=find(TDR_results.t > (param.N_bx+1)*param.ui+tfx+3*tr*1e-9,1,'first');
    tk=param.ui*1*(param.N_bx+1)+tfx+3*tr*1e-9;
    % [ahealey] End of modifications.
    if  isempty(ndfex), ndfex=length(TDR_results.t); end
    PTDR.pulse_orig=PTDR.pulse;
    
    switch param.Grr
        case 0 % pre .3cd release
            fctrx(1:length(PTDR.pulse_orig))=(1+param.rho_x)*param.rho_x;
        case 1 % .3cd release
            fctrx(1:length(PTDR.pulse_orig))=1;
        case 2 % .3ck working
            fctrx(1:length(PTDR.pulse_orig))=1;
    end
    Gloss(1:length(TDR_results.t))=1;
    Grr(1:length(TDR_results.t))=1;
    fctrx(1:ntx)=0; % moved out of loop for beta x and n_bx
    
    for ii=ntx:ndfex
        % adjust for near end loss
        if param.N_bx>0 && param.beta_x~=0;
            Gloss(ii)= 10.^(param.beta_x*(TDR_results.t(ii)-tk)/20);
        else
            Gloss(ii)=1;
        end
        % ---------------------- 2.7 code
        %         x=(TDR_results.t(ii)-tfx)/param.ui;
        % ----------------------
        % [ahealey9/06/2019] Need to account for the 3*TR_TDR delay included in the
        % rise time filter.
        % x=(TDR_results.t(ii)-tfx)/param.ui;
        x=(TDR_results.t(ii)-tfx-3*tr*1e-9)/param.ui;
        % determine how much of the return loss to use base on expected
        % missing reflections
        switch param.Grr
            case 0 % pre .3cd release
                Grr(ii)= (1+param.rho_x)*param.rho_x*exp(-(x-1*param.N_bx-1).^2/(1+param.N_bx)^2);
            case 1 % .3cd release
                Grr(ii)= (1+param.rho_x)*param.rho_x*exp(-(x-1*param.N_bx-1).^2/(1+param.N_bx)^2);
            case 2 % .3ck working
                Grr(ii)= param.rho_x ;
        end
        fctrx(ii)=Gloss(ii).*Grr(ii);
    end
    
    if isrow(fctrx), fctrx=fctrx(:);end
    PTDR.pulse=PTDR.pulse.*fctrx;
    if 0
        figure(10101+param.RL_sel);set(gcf,'Tag','COM');
        s1=subplot(2,1,1);
        plot((TDR_results.t(ntx:end)- TDR_results.t(ntx))/param.ui,fctrx(ntx:end),'disp','G_l_o_s_s*G_r_r');
        hold on
        plot((TDR_results.t(ntx:end)- TDR_results.t(ntx))/param.ui,Gloss(ntx:end),'disp','G_l_o_s_s');
        plot((TDR_results.t(ntx:end)- TDR_results.t(ntx))/param.ui,Grr(ntx:end),'disp','G_r_r');
        grid on
        ylim([ 0 1.2])
        s2=subplot(2,1,2);
        plot((TDR_results.t(ntx:end)- TDR_results.t(ntx))/param.ui,PTDR.pulse(ntx:end)./fctrx(ntx:end),'disp','PTDR');
        grid on
        linkaxes([s1,s2],'x')
        xlabel 'UI'
        xlim ([ 1 200])
    end
    
    FAST_NOISE_CONV=0;
    ERLRMS=rms(PTDR.pulse);
    for ki=1:param.samples_per_ui
        progress = ki/param.samples_per_ui;
        if OP.DISPLAY_WINDOW
            waitbar(progress, hwaitbar, 'ERL computing'); figure(hwaitbar); drawnow;
        else
            if ~mod(progress*100,1), fprintf('%i%% ', progress*100 );end
        end
        tps=PTDR.pulse(ki:param.samples_per_ui:end);
        if OP.RL_norm_test
            rl_fom=(norm(tps));
        else
            testpdf=get_pdf_from_sampled_signal( tps,L, BinSize*10, FAST_NOISE_CONV );
            cdf_test=cumsum(testpdf.y);
            rl_test=(-testpdf.x(find(cdf_test>=param.specBER,1,'first')));
            rl_fom=rl_test;
        end
        if rl_fom > RL_equiv
            RL_equiv=rl_fom;
            best_ki=ki;
        end
        if ~OP.RL_norm_test
            best_erl=rl_test;
            best_pdf=testpdf;
            best_cdf=cdf_test;
        end
        
    end
    if OP.RL_norm_test
        tps=PTDR.pulse(best_ki:param.samples_per_ui:end);
        testpdf=get_pdf_from_sampled_signal( tps,L, BinSize*10, FAST_NOISE_CONV );
        cdf_test=cumsum(testpdf.y);
        best_erl=(-testpdf.x(find(cdf_test>=param.specBER,1,'first')));
    end
    
    fprintf('\n');
    try
        close(hwaitbar)
    catch
    end
    if ~exist('best_ki','var'),best_ki=1;end
    TDR_results.ptdr_RL=PTDR.pulse;  % reflection waveform from the pulse
    TDR_results.WC_ptdr_samples_t=TDR_results.t(best_ki:param.samples_per_ui:end);
    TDR_results.WC_ptdr_samples=PTDR.pulse(best_ki:param.samples_per_ui:end);
    TDR_results.ERL=-db(best_erl);
    TDR_results.ERLRMS=-db(ERLRMS);
    
end


% end get TDR
%%




function [chdata, param] = get_TD_files(param, OP, num_fext, num_next, file_list)
% filename parsing and acquisition
%------------------------------------------------------------------
%----------put files names into chdata structure ---------
% The thru file has the index of 1
% crosstalk file are indexed from 2
% nxi is incremented each time a file is read in  so that nxi will end
filepath=[]; % path name for file
nxi=0; % file index
% get the THRU file
if size(file_list,2) ~= 0
    file_list(1)=strrep(file_list(1),'\', filesep); % OS file convention conversion
    [filepath, basename, fileext]=fileparts(file_list{1});
    
else
    if OP.RX_CALIBRATION == 1 || OP.FIXTURE_CALIBRATION
        h = msgbox('enter test channel s-parameter file'); set(h,'Color',[1 .85 0]);
        movegui(h,'northeast')
    end
    dir=fullfile(filepath, '*.csv');
    [basename,filepath]=uigetfile(dir,'input thru channel response .cav ');
    if filepath == 0
        error('No Thru file')
    end
    [UNUSED_OUTPUT, basename, fileext]=fileparts(basename); %#ok<ASGLU>
end
nxi=nxi+1;
chdata(nxi).filename = fullfile(filepath,  [basename fileext]);
chdata(nxi).ext = fileext;
[UNUSED_OUTPUT, dirname]=fileparts(filepath); %#ok<ASGLU>
%    chdata(nxi).base=[pth(max(strfind(pth,filesep))+1:end) '--' basename]; % add 1 directory back to basename
chdata(nxi).base=[OP.RUNTAG ' ' dirname  '--' basename]; % add 1 directory back to basename
% chdata(nxi).A=param.A_thru; % pam encoding amplitude reduction is do in reporting but is comprehending in crosstalk PDF
chdata(nxi).type='THRU';
chdata(nxi).ftr=param.fb*param.f_v;
param.base=chdata(nxi).base; %for print out function that don't pass chdata

% now get FEXT file names into chdata structure
kxi=nxi;
for nxi=kxi+1:num_fext+kxi
    lastfilepath=filepath;
    if size(file_list,2) ~= 0
        [filepath, basename, fileext]=fileparts(file_list{nxi});
    else
        if OP.RX_CALIBRATION == 1 || OP.FIXTURE_CALIBRATION
            h = msgbox('enter noise channel s-parameter file'); set(h,'Color',[1 .85 0]);
            movegui(h,'northeast')
        end
        if  param.tfx(1) == -1 && OP.ERL_ONLY && OP.ERL == 2;
            dir=fullfile(filepath, '*.csv');
            [basename,filepath]=uigetfile(dir,'input noise channel response .csv');
            if filepath==0
                error('Not enough NEXT files')
            end
        else
            dir=fullfile(filepath, '*.csv');
            if OP.RX_CALIBRATION == 1 || OP.FIXTURE_CALIBRATION
                [basename,filepath]=uigetfile(dir,'input noise channel response .csv');
            else
                [basename,filepath]=uigetfile(dir,['input fext channel response .csv #', num2str(nxi-kxi)]);
            end
            if filepath==0
                error('Not enough NEXT files')
            end
        end
        [UNUSED_OUTPUT, basename, fileext]=fileparts(basename); %#ok<ASGLU>
    end
    if isempty( filepath), filepath=lastfilepath; end
    chdata(nxi).filename = fullfile(filepath,  [basename fileext]);
    chdata(nxi).ext = fileext;
    [UNUSED_OUTPUT, dirname]=fileparts(filepath); %#ok<ASGLU>
    chdata(nxi).base=[OP.RUNTAG ' ' dirname '--' basename ];
    %     chdata(nxi).A=param.a_fext;
    chdata(nxi).ftr=param.fb*param.f_f;
    chdata(nxi).type='FEXT';
end
% now get NEXT file names into chdata structure
kxi=num_fext+kxi;
for nxi=kxi+1:num_next+kxi
    lastfilepath=filepath;
    if size(file_list,2) ~= 0
        [filepath, basename, fileext]=fileparts(file_list{nxi});
    else
        dir=fullfile(filepath, '*.csv');
        [basename,filepath]=uigetfile(dir,['input next channel response .csv ', num2str(nxi-kxi)]);
        if filepath==0
            error('Not enough NEXT files')
        end
        [UNUSED_OUTPUT, basename, fileext]=fileparts(basename); %#ok<ASGLU>
    end
    if isempty( filepath), filepath=lastfilepath; end
    chdata(nxi).filename = fullfile(filepath,  [basename fileext]);
    chdata(nxi).ext = fileext;
    [UNUSED_OUTPUT, dirname]=fileparts(filepath); %#ok<ASGLU>
    chdata(nxi).base=[OP.RUNTAG ' ' dirname '--' basename ];
    %     chdata(nxi).A=param.A_next;
    chdata(nxi).ftr=param.fb*param.f_n;
    chdata(nxi).type='NEXT';
end
function half_UI=get_center_of_UI(samples_per_UI)

%half_UI reveals which value to use for the center of the UI.  For eye
%width calculations, it is necessary to place the cursor in the center of the
%UI window to ensure a 0 crossing on both left/right inside the window.
%This function was written in order to support even and odd samples_per_UI
%and to prevent the ambiguity of using samples_per_UI/2 vs. samples_per_UI/2+1

%The UI window goes from 0 to 1 with 1/samples_per_UI steps
UI_window=0:1/samples_per_UI:1-1/samples_per_UI;
%the center of the UI is sample closest to 0.5
[temp_diff,half_UI]=min(abs(UI_window-0.5));
function pdf=get_pdf(chdata, delta_y, t_s, param, OP)
SBR=chdata.eq_pulse_response(:)'; % row vector
type=chdata.type;
samp_UI=param.samples_per_ui;
residual_response = SBR;

if isequal(type, 'THRU')
    % for thru pulse response:
    % remove the cursor and the DFE postcursors (up to their limit), since
    % we only care about the residuals.
    
    if param.N_bg ==0
        ideal_cancelled_cursors = SBR(t_s+param.samples_per_ui*(0:param.ndfe));
    else
        ideal_cancelled_cursors = SBR(t_s+param.samples_per_ui*(0:param.N_bmax));
    end
    if param.dfe_delta ~= 0
        ideal_cancelled_cursors_q=floor(abs( ideal_cancelled_cursors/(residual_response(t_s).*param.dfe_delta) )).*residual_response(t_s)*param.dfe_delta.*sign(ideal_cancelled_cursors);
    else
        ideal_cancelled_cursors_q=ideal_cancelled_cursors;
    end
    
    %AJG021820
    if param.N_bg ==0
        bmax_vec=residual_response(t_s)*[1,param.bmax];
        bmin_vec=residual_response(t_s)*[1,param.bmin];
    else
        bmax_vec=residual_response(t_s)*[1,param.use_bmax];
        bmin_vec=residual_response(t_s)*[1,param.use_bmin];
    end
    effective_cancelled_cursors=dfe_clipper(ideal_cancelled_cursors_q,bmax_vec,bmin_vec);
    
    
    effective_cancellation_samples = kron(effective_cancelled_cursors, ones(1, param.samples_per_ui));
    dfetaps=effective_cancelled_cursors/SBR(t_s);
    
    % Apply a constant DFE coefficient 1/2 UI before and after each postcursor. Not
    % really needed for COM, but helps debugging. May be factored out in future revisions.
    start_cancel = t_s-param.samples_per_ui/2;
    if param.N_bg ==0
        end_cancel = t_s+(1/2+param.ndfe)*param.samples_per_ui - 1;
    else
        end_cancel = t_s+(1/2+param.N_bmax)*param.samples_per_ui - 1;
    end
    residual_response(start_cancel:end_cancel) = ...
        residual_response(start_cancel:end_cancel) - effective_cancellation_samples;
    %else
    % for crosstalk pulse responses, nothing is cancelled, and all phases
    % are equally important.
end

nui=round(length(residual_response)/param.samples_per_ui);

vs=zeros(nui-2, param.samples_per_ui);
for i=1:param.samples_per_ui
    vs(:,i)=residual_response(param.samples_per_ui*(1:nui-2)+i);
end

if OP.DISPLAY_WINDOW,
    hwaitbar=waitbar(0);
end

% determine which pdf to use
if isequal(type, 'THRU')
    % one phase is interesting for thru
    phases = mod(t_s,param.samples_per_ui);
    if phases==0, phases = param.samples_per_ui; end
else
    phases=1:samp_UI;
end

mxV = zeros(size(phases));
for k=phases
    pdf_samples(k)=get_pdf_from_sampled_signal(vs(:,k), param.levels, delta_y); %#ok<AGROW>
    mxV(k)=sqrt(sum( pdf_samples(k).x.^2.*pdf_samples(k).y)); % standard deviation of PDF
    progress = k/length(phases);
    if OP.DISPLAY_WINDOW, waitbar(progress, hwaitbar, ['processing COM pdf ' chdata.base ] ); figure(hwaitbar); drawnow; end
end
[UNUSED_OUTPUT pxi]=max(mxV); %#ok<ASGLU>
pdf=pdf_samples(pxi);

if OP.DISPLAY_WINDOW,
    close(hwaitbar);
end

function [ pdf ] = get_pdf_from_sampled_signal( input_vector, L, BinSize ,FAST_NOISE_CONV)
% Create PDF from interference vector using successive delta-set convolutions.
%   input_vector = list of values of samples
%   return
%   pdf.x
%   pdf.y
%   pdf.vec
%   pdf.bin
if ~exist('FAST_NOISE_CONV','var')
    FAST_NOISE_CONV=0;
end
if max(input_vector) > BinSize
    input_vector=input_vector(abs(input_vector)>BinSize);
end
% for i = 1:length(input_vector)
%    if abs(input_vector(i)) < BinSize , input_vector(i)=0; end
%end

input_vector(abs(input_vector)<BinSize) = 0;
b=sign(input_vector);
[input_vector,index]=sort(abs(input_vector),'descend');
input_vector=input_vector.*b(index);
if FAST_NOISE_CONV
    sig_res=norm(input_vector(find(abs(input_vector)<.001,1)+1:end));
    res_pdf= normal_dist(sig_res,5,BinSize);
    input_vector=input_vector(1:find(abs(input_vector)<.001,1));
end
%% Equation 93A-39 %%
values = 2*(0:L-1)/(L-1)-1;
prob = ones(1,L)/L;

%% Initialize pdf to delta at 0
pdf=d_cpdf(BinSize, 0, 1);
empty_pdf=pdf;
for k = 1:length(input_vector)
    %     pdfn=d_cpdf(BinSize, abs(input_vector(k))*values, prob);
    pdfn=Init_PDF_Fast(empty_pdf, abs(input_vector(k))*values, prob);
    pdf=conv_fct(pdf, pdfn);
end
if FAST_NOISE_CONV
    %     pdf=conv_fct(pdf,res_pdf);
    pdf=conv_fct_TEST(pdf,res_pdf);
end

function [pdf,h_j_full,A_s_vec]=get_pdf_full(chdata, delta_y, t_s, param, OP,pdf_range)
t_s_orig=t_s;
%SBR=chdata.eq_pulse_response(:)'; % row vector SBR(1:t_s-14)=0;SBR(t_s+15:end)=0; for debug (AJG edit)
type=chdata.type;

pulse_orig=chdata.eq_pulse_response(:)';
%build arbitrary time axis with step size = 1/samples per ui
old_time=[0:length(pulse_orig)-1]/param.samples_per_ui;
%force t_s at time =0 (makes the other things below easy)
original_sample_time=old_time(t_s_orig);
old_time=old_time-original_sample_time;
%build new time axis that forces time=0 to be in the axis
%unless the new/old samples per UI are integer ratios, time 0 will not be
%there by default
samp_UI=param.samples_for_C2M;
new_timea=[0:-1/samp_UI:min(old_time)];
new_timeb=[0:1/samp_UI:max(old_time)];
new_time=[fliplr(new_timea) new_timeb(2:end)];
SBR=interp1(old_time,pulse_orig,new_time);
%new sample time is simply the point where new_time = 0
[tmp,t_s]=min(abs(new_time));

residual_response = SBR;

half_UI=get_center_of_UI(samp_UI);

if isequal(type, 'THRU')
    % for thru pulse response:
    % remove the cursor and the DFE postcursors (up to their limit), since
    % we only care about the residuals.
    
    %AJG021820
    if param.N_bg ==0
        ideal_cancelled_cursors = SBR(t_s+samp_UI*(1:param.ndfe));
    else
        ideal_cancelled_cursors = SBR(t_s+samp_UI*(1:param.N_bmax));
    end
    if param.dfe_delta ~= 0
        ideal_cancelled_cursors_q=floor(abs( ideal_cancelled_cursors/(residual_response(t_s).*param.dfe_delta) )).*residual_response(t_s)*param.dfe_delta.*sign(ideal_cancelled_cursors);
    else
        ideal_cancelled_cursors_q=ideal_cancelled_cursors;
    end
    
    if param.N_bg ==0
        bmax_vec=residual_response(t_s)*[param.bmax];
        bmin_vec=residual_response(t_s)*[param.bmin];
    else
        bmax_vec=residual_response(t_s)*[param.use_bmax];
        bmin_vec=residual_response(t_s)*[param.use_bmin];
    end
    effective_cancelled_cursors=dfe_clipper(ideal_cancelled_cursors_q,bmax_vec,bmin_vec);
    
    
    effective_cancellation_samples = kron(effective_cancelled_cursors, ones(1, samp_UI));
    dfetaps=effective_cancelled_cursors/SBR(t_s);
    
    % Apply a constant DFE coefficient 1/2 UI before and after each postcursor. Not
    % really needed for COM, but helps debugging. May be factored out in future revisions.
    
    %avoid dividing samp_UI by 2 in case it is not even
    start_cancel=t_s-half_UI+1+samp_UI;
    %AJG021820
    if param.N_bg ==0
        end_cancel=start_cancel+param.ndfe*samp_UI-1;
    else
        end_cancel=start_cancel+param.N_bmax*samp_UI-1;
    end
    residual_response(start_cancel:end_cancel) = ...
        residual_response(start_cancel:end_cancel) - effective_cancellation_samples;
    %else
    % for crosstalk pulse responses, nothing is cancelled, and all phases
    % are equally important.
    
    %remove entire cursor UI
    uiv_start=start_cancel-samp_UI;
    uiv_end=uiv_start+samp_UI-1;
    A_s_vec = param.R_LM*SBR(uiv_start:uiv_end)/(param.levels-1);
    residual_response(uiv_start:uiv_end)=0;
end

nui=round(length(residual_response)/samp_UI);


vs=transpose(reshape(residual_response(samp_UI+1:samp_UI*(nui-1)),samp_UI,nui-2));
%added vs_raw in order to calculate h_j.  vs_raw uses the pulse
%response without DFE included.  (Can't include DFE for jitter calc)
vs_raw=transpose(reshape(SBR(samp_UI+1:samp_UI*(nui-1)),samp_UI,nui-2));

% if OP.DISPLAY_WINDOW,
%     hwaitbar=waitbar(0);
% end

% determine which pdf to use
if isequal(type, 'THRU')
    % one phase is interesting for thru
    phases = mod(t_s,samp_UI);
    if phases==0, phases = samp_UI; end
else
    phases=1:samp_UI;
end

mxV = zeros(size(phases));

%phases reveals the raw position in the UI window of the cursor.
%shift_amount is the amount to shift so that it aligns with half_UI
shift_amount=half_UI-phases;
%vs_shift puts the cursor at the center
vs_shift=circshift(vs,[0 shift_amount]);
L=size(vs_raw,1);
%allow partial UI computation through pdf_range
%if pdf_range is empty, do full UI
if isempty(pdf_range)
    pdf_range=1:samp_UI;
else
    pdf_range=min(pdf_range):max(pdf_range);
end
h_j_full=zeros(L,samp_UI);
for k=pdf_range
    pdf(k)=get_pdf_from_sampled_signal(vs_shift(:,k), param.levels, delta_y); %#ok<AGROW>
    %mxV(k)=sqrt(sum( pdf_samples(k).x.^2.*pdf_samples(k).y)); % standard deviation of PDF
    %progress = k/length(phases);
    %if OP.DISPLAY_WINDOW, waitbar(progress, hwaitbar, ['processing COM pdf ' chdata.base ] ); figure(hwaitbar); drawnow; end
    
    %build the circshift of h_j_full into the loop to support a reduced
    %range of sampling points. circshift at the end only works if doing the
    %full range of sampling points.  And shifting before the loop will
    %yield the wrong answer at the edges of the UI
    hk=k-shift_amount;
    if hk<1
        hk=hk+samp_UI;
    elseif hk>samp_UI
        hk=hk-samp_UI;
    end
    if hk==1
        %when hk=1, the early UI is the last column
        h_j_full(1:L-1,k)=(vs_raw(2:end,hk+1)-vs_raw(1:end-1,samp_UI))/2*samp_UI;
    elseif hk==samp_UI
        %when hk=samp_UI, the late UI is the first column
        h_j_full(1:L-1,k)=(vs_raw(2:end,1)-vs_raw(1:end-1,hk-1))/2*samp_UI;
    else
        %for all other cases, do the normal late=+1, early = -1
        h_j_full(1:L,k)=(vs_raw(:,hk+1)-vs_raw(:,hk-1))/2*samp_UI;
    end
end
function [chdata, param] = get_s4p_files(param, OP, num_fext, num_next, file_list)
% filename parsing and acquisition
%------------------------------------------------------------------
%----------put files names into chdata structure ---------
% The thru file has the index of 1
% crosstalk file are indexed from 2
% nxi is incremented each time a file is read in  so that nxi will end
filepath=[]; % path name for file
nxi=0; % file index
% get the THRU file
if size(file_list,2) ~= 0
    file_list(1)=strrep(file_list(1),'\', filesep); % OS file convention conversion
    [filepath, basename, fileext]=fileparts(file_list{1});
    
else
    if OP.RX_CALIBRATION == 1 || OP.FIXTURE_CALIBRATION
        h = msgbox('enter test channel s-parameter file'); set(h,'Color',[1 .85 0]);
        movegui(h,'northeast')
    end
    if  OP.ERL == 2
        dir=fullfile(filepath, '*.s2p');
        [basename,filepath]=uigetfile(dir,'input RL measurement .s2p ');
        if filepath == 0
            error('No RL measurement file')
        end
    else
        dir=fullfile(filepath, '*.s4p');
        [basename,filepath]=uigetfile(dir,'input thru channel response .s4p ');
        if filepath == 0
            error('No Thru file')
        end
    end
    [UNUSED_OUTPUT, basename, fileext]=fileparts(basename); %#ok<ASGLU>
end
nxi=nxi+1;
chdata(nxi).filename = fullfile(filepath,  [basename fileext]);
chdata(nxi).ext = fileext;
[UNUSED_OUTPUT, dirname]=fileparts(filepath); %#ok<ASGLU>
%    chdata(nxi).base=[pth(max(strfind(pth,filesep))+1:end) '--' basename]; % add 1 directory back to basename
chdata(nxi).base=[OP.RUNTAG ' ' dirname  '--' basename]; % add 1 directory back to basename
% chdata(nxi).A=param.A_thru; % pam encoding amplitude reduction is do in reporting but is comprehending in crosstalk PDF
chdata(nxi).type='THRU';
chdata(nxi).ftr=param.fb*param.f_v;
param.base=chdata(nxi).base; %for print out function that don't pass chdata

% now get FEXT file names into chdata structure
kxi=nxi;
for nxi=kxi+1:num_fext+kxi
    lastfilepath=filepath;
    if size(file_list,2) ~= 0
        [filepath, basename, fileext]=fileparts(file_list{nxi});
    else
        if OP.RX_CALIBRATION == 1 || OP.FIXTURE_CALIBRATION
            h = msgbox('enter noise channel s-parameter file'); set(h,'Color',[1 .85 0]);
            movegui(h,'northeast')
        end
        if  param.tfx(1) == -1 && OP.ERL_ONLY && OP.ERL == 2;
            dir=fullfile(filepath, '*.s4p');
            [basename,filepath]=uigetfile(dir,'input noise channel response .s4p');
            if filepath==0
                error('Not enough NEXT files')
            end
        else
            dir=fullfile(filepath, '*.s4p');
            if OP.RX_CALIBRATION == 1 || OP.FIXTURE_CALIBRATION
                [basename,filepath]=uigetfile(dir,'input noise channel response .s4p');
            else
                [basename,filepath]=uigetfile(dir,['input fext channel response .s4p #', num2str(nxi-kxi)]);
            end
            if filepath==0
                error('Not enough NEXT files')
            end
        end
        [UNUSED_OUTPUT, basename, fileext]=fileparts(basename); %#ok<ASGLU>
    end
    if isempty( filepath), filepath=lastfilepath; end
    chdata(nxi).filename = fullfile(filepath,  [basename fileext]);
    chdata(nxi).ext = fileext;
    [UNUSED_OUTPUT, dirname]=fileparts(filepath); %#ok<ASGLU>
    chdata(nxi).base=[OP.RUNTAG ' ' dirname '--' basename ];
    %     chdata(nxi).A=param.a_fext;
    chdata(nxi).ftr=param.fb*param.f_f;
    chdata(nxi).type='FEXT';
end
% now get NEXT file names into chdata structure
kxi=num_fext+kxi;
for nxi=kxi+1:num_next+kxi
    lastfilepath=filepath;
    if size(file_list,2) ~= 0
        [filepath, basename, fileext]=fileparts(file_list{nxi});
    else
        dir=fullfile(filepath, '*.s4p');
        [basename,filepath]=uigetfile(dir,['input next channel response .s4p ', num2str(nxi-kxi)]);
        if filepath==0
            error('Not enough NEXT files')
        end
        [UNUSED_OUTPUT, basename, fileext]=fileparts(basename); %#ok<ASGLU>
    end
    if isempty( filepath), filepath=lastfilepath; end
    chdata(nxi).filename = fullfile(filepath,  [basename fileext]);
    chdata(nxi).ext = fileext;
    [UNUSED_OUTPUT, dirname]=fileparts(filepath); %#ok<ASGLU>
    chdata(nxi).base=[OP.RUNTAG ' ' dirname '--' basename ];
    %     chdata(nxi).A=param.A_next;
    chdata(nxi).ftr=param.fb*param.f_n;
    chdata(nxi).type='NEXT';
end

function [sigma_N] = get_sigma_eta_ACCM_noise(chdata,param,H_sy,H_r,H_ctf)
% 11-25-2020 correct integratation limits, should be infinity or highest specified frequency
% H_sy - PSD for Tx power delivery, not normally used and set to a vector 1's
% H_r - receiver filter, Butterworth
% H_ctf - total gain of CTLE and low freq filtering
% H_dc - the common mode channel gain
% param.eta_0 -input referred Rx noise
% param.AC_CM_RMS - AC CM source before Tx series source resistor.
% param.ACCM_MAX_Freq - Max frequency for ACCM source intergration
%% Equation 93A-35 - independent of FFE setting %%
sigma_N1 = sqrt(param.eta_0*sum( abs(H_sy(2:end) .* H_r(2:end) .* H_ctf(2:end) ).^2 .* diff(chdata(1).faxis)/1e9));% eta_0 is V^2/Ghz i.e. /1
if sum(param.AC_CM_RMS) ~= 0
    sigma_ACCM=0;
    f_int= chdata(1).faxis( chdata(1).faxis<=param.ACCM_MAX_Freq );
    for i=1:length(chdata)
        H_dc=abs(squeeze(chdata(i).sdc21));
        sigma_ACCM_acc= sqrt( 2*param.AC_CM_RMS_TX^2.*sum( abs(H_sy(2:length(f_int)) .* H_r(2:length(f_int)) .* H_ctf(2:length(f_int)).*H_dc(2:length(f_int)) ).^2 .* diff(f_int))/f_int(end) );
        sigma_ACCM=norm([sigma_ACCM_acc,sigma_ACCM]);
    end
    sigma_N=norm([sigma_N1,sigma_ACCM]);
else
    sigma_N=sigma_N1;
end
%%
function [ sigma_NE ] = get_sigma_noise( H_ctf, param, chdata, sigma_bn )
% for Rx calibratrion only
% the FEXT channel for calibration basically a DC connection unlike normal
% FEXT channels which are nearly open at DC channels
H_r = 1./polyval([1 2.613126 3.414214 2.613126 1], 1i*chdata(2).faxis/(param.f_r*param.fb));
idxfbby2=find( chdata(2).faxis(:) >= param.fb/2, 1);
if size(chdata,2) >= 2
    Hnoise_channel=chdata(2).sdd21;% rx package is already included tx is not
else
    Hnoise_channel=1;
end
H_np=Hnoise_channel.*H_ctf.*H_r;
%% Equation 93A-48 %%
sigma_NE = sigma_bn*sqrt(mean(abs(H_np(1:idxfbby2).^2)));

function [sigma_XT, sigma_FEXT, sigma_NEXT] = get_xtlk_noise( upsampled_txffe, type, param, chdata, C )
% Modified not to double count crosstalk: John Ewen 13/12/2018
% function sigma_XT = get_xtlk_noise( upsampled_txffe, type, param, chdata,ctle_indx,clow_indx, C,cursor_i)
index_f2=find(chdata(1).faxis(:)>param.fb,1,'first');
if isempty(index_f2), index_f2=length(chdata(1).faxis);end
f=chdata(1).faxis;
temp_angle=(param.samples_per_ui*param.sample_dt)*pi.*chdata(1).faxis;
if(f(1)==0)
    temp_angle(1)=1e-20;% we don't want to divide by zero
end
PWF_tx=ones(1,length(f));
if max(upsampled_txffe) > 0
    PWF_tx=zeros(1,length(f));
    [mcur,icur] = max(upsampled_txffe);
    for ii=1:length(upsampled_txffe)
        %         PWF_tx=upsampled_txffe(ii).*exp(-1j*2*pi*(ii-icur).*f/param.fb)+PWF_tx;
        % Adee Ran 2020-06-03 remove create large 2D matrix when 1D is all
        % that is needed
        term_ii = upsampled_txffe(ii).*exp(-1j*2*pi*(ii-icur).*f/param.fb);
        PWF_tx = PWF_tx + term_ii(:)';
        % /Adee
    end
end
PWF_rx=ones(1,length(f));
if exist('C','var')
    PWF_rx=zeros(1,length(f));
    for ii=-param.RxFFE_cmx:param.RxFFE_cpx
        PWF_rx=C(ii+param.RxFFE_cmx+1).*exp(-1j*2*pi*(ii+1).*f/param.fb)+PWF_rx;
    end
end
MDFEXT=0;MDNEXT=0;MDNEXT_ICN=0;MDFEXT_ICN=0;
for ii=2:size(chdata,2)
    SINC = sin(temp_angle)./temp_angle;
    PWF_data=SINC.^2;
    PWF=PWF_data.*PWF_rx; % power weight function
    PWFnext=abs(PWF);
    PWF=PWF_data.*PWF_rx.*PWF_tx; % power weight function
    PWFfext=abs(PWF);
    if isequal(chdata(ii).type, 'FEXT')
        MDFEXT=sqrt(abs(chdata(ii).sdd21ctf).^2+MDFEXT.^2); % power sum xtk
        MDFEXT_ICN=sqrt(2*chdata(ii).delta_f/param.f2*sum( chdata(ii).A^2*PWFfext(1:index_f2).*abs(MDFEXT(1:index_f2)).^2)); %eq 46
    elseif isequal(chdata(ii).type, 'NEXT')
        MDNEXT=sqrt(abs(chdata(ii).sdd21ctf).^2+MDNEXT.^2); % power sum xtk
        MDNEXT_ICN=sqrt(2*chdata(ii).delta_f/param.f2*sum( chdata(ii).A^2*PWFnext(1:index_f2).*abs(MDNEXT(1:index_f2)).^2)); %eq 47
    end
end
if nargout == 1 && isequal(type,'NEXT')
    sigma_XT = MDNEXT_ICN*sqrt((param.levels^2-1)/(3*(param.levels-1)^2));
elseif nargout == 1 && isequal(type,'FEXT')
    sigma_XT = MDFEXT_ICN*sqrt((param.levels^2-1)/(3*(param.levels-1)^2));
elseif nargout == 3
    sigma_XT=norm([ MDNEXT_ICN MDFEXT_ICN ])*sqrt((param.levels^2-1)/(3*(param.levels-1)^2));
    sigma_NEXT = MDNEXT_ICN*sqrt((param.levels^2-1)/(3*(param.levels-1)^2));
    sigma_FEXT = MDFEXT_ICN*sqrt((param.levels^2-1)/(3*(param.levels-1)^2));
end

function sigma_XT = get_xtlk_noise1( upsampled_txffe, type, param, chdata,C,cursor_i )
% calculate crosstalk sigma at worst phase, per equation 93A-33, for
% channels of chosen type (FEXT and NEXT are treated separately).

sigma_XT_sqd=0;
for k=1:param.num_s4p_files
    if isequal(chdata(k).type, type)
        sigma_i_sqd=0;
        effective_channel = filter(upsampled_txffe, 1, chdata(k).ctle_imp_response);
        N=round(length(effective_channel)/param.samples_per_ui)-3;
        sbr=filter(ones(param.samples_per_ui, 1), 1, effective_channel);
        if exist('C','var')
            [ sbr,~] = force( sbr ,param, 0 , cursor_i, C );
        end
        for m=1:param.samples_per_ui
            h=sbr(param.samples_per_ui*(1:N)+m);
            sigma_i_sqd=max(param.sigma_X^2*sum(h.^2),sigma_i_sqd);
        end
        sigma_XT_sqd = sigma_XT_sqd + sigma_i_sqd;
    end
end
sigma_XT=sqrt(sigma_XT_sqd);

function out=hrem(h,index,N_bf,bmaxg)

out=[ h(1:index-1) h(index:index+N_bf-1)- sign(h(index:index+N_bf-1)).*   (min( bmaxg, abs( h(index:index+N_bf-1) )))  h(index+N_bf:end) ];
% faster than single line function
% hrem =@(h,index,N_bf,bmaxg) ...
%     [ h(1:index-1) ...
%     h(index:index+N_bf-1)- sign(h(index:index+N_bf-1)).*   (min( bmaxg, abs( h(index:index+N_bf-1) ))) ...
%     h(index+N_bf:end) ]...
%     ;

%% floating DFE taps
function [Sout] = interp_Sparam(Sin,fin,fout, ...
    opt_interp_Sparam_mag, opt_interp_Sparam_phase,OP)
% Sout = interp_Sparam(Sin,fin,fout)
%
% Interpolate S-parameters Sin from frequency grid fin to frequency grid
% fout.

if ( fin(end)<fout(end) )
    %    warning('Channel high frequencies extrapolation might be inaccurate!');
end

H_mag = abs(Sin);
H_mag(H_mag<eps)=eps; % handle ill cases...
H_ph = unwrap(angle(Sin));
% For long delay channels, the result can turn anti-causal if frequency step is too coarse. Don't let the
% user ignore that.
if mean(diff(H_ph))>0
    if OP.DEBUG
        warning('Anti-causal response found. Finer frequency step is required for this channel');
    else
        error('Anti-causal response found. Finer frequency step is required for this channel');
    end
end

%opt_interp_Sparam_mag='linear_trend_to_DC';
switch opt_interp_Sparam_mag
    case {'linear_trend_to_DC','linear_trend_to_DC_log_trend_to_inf'}
        if -iscolumn(H_mag), H_mag=H_mag.';end
        if -iscolumn(fin), fin=fin.';end
        fin_x=fin;
        H_mag_x=H_mag(:);
        if fin(1)>0
            p=polyfit(fin(1:10), H_mag(1:10), 1);
            dc_trend_val=polyval(p, 0);
            fin_x=[0, fin_x];
            H_mag_x = [dc_trend_val; H_mag_x];
        end
        if fin(end)<fout(end)
            warn_state=warning('off', 'MATLAB:polyfit:RepeatedPointsOrRescale');
            mid_freq_ind=round(length(fin)/2);
            p=polyfit(fin(mid_freq_ind:end), H_mag(mid_freq_ind:end), 1);
            warning(warn_state);
            hf_trend_val=polyval(p, fout(end));
            if hf_trend_val>H_mag(end)
                hf_trend_val=H_mag(end);
                hf_logtrend_val = H_mag(end);
            elseif hf_trend_val<eps
                hf_trend_val=eps;
                hf_logtrend_val = realmin;
            end
            fin_x=[fin_x, fout(end)];
            H_mag_x = [H_mag_x; hf_trend_val];
        end
        H_mag_i = interp1(fin_x, H_mag_x, fout, 'linear', 'extrap');
        if strcmp(opt_interp_Sparam_mag,'linear_trend_to_DC_log_trend_to_inf')
            H_logmag_i = exp(interp1(fin_x, log([H_mag_x(1:end-1);hf_logtrend_val]), fout, 'linear', 'extrap'));
            indx = find(fout > fin(end),1,'first');
            H_mag_i(indx:end) = H_logmag_i(indx:end);
        end
    case 'trend_to_DC'
        % extrapolate to trend value at DC.
        if -iscolumn(H_mag), H_mag=H_mag.';end
        if -iscolumn(fin), fin=fin.';end
        fin_x=fin;
        H_mag_x=H_mag;
        if fin(1)>0
            p=polyfit(fin(1:10), log10(H_mag(1:10)), 1);
            dc_trend_val=10^polyval(p, 0);
            fin_x=[0, fin_x];
            H_mag_x = [dc_trend_val H_mag_x];
        end
        if fin(end)<fout(end)
            warn_state=warning('off', 'MATLAB:polyfit:RepeatedPointsOrRescale');
            mid_freq_ind=round(length(fin)/2);
            p=polyfit(fin(mid_freq_ind:end), log10(H_mag(mid_freq_ind:end)), 1);
            warning(warn_state);
            hf_trend_val=10^polyval(p, fout(end));
            if hf_trend_val>H_mag(end)
                hf_trend_val=H_mag(end);
            end
            fin_x=[fin_x, fout(end)];
            H_mag_x = [H_mag_x hf_trend_val];
        end
        H_mag_i = 10.^interp1(fin_x,log10(H_mag_x),fout,'linear', 'extrap');
    case 'extrap_to_DC_or_zero'
        % same as extrap_to_DC but detect AC-coupled channels and
        % extrapolate them to 0.
        if fin(1)>0 && 20*log10(H_mag(1))<-20
            % assume AC coupling, with 0 at DC
            H_mag_i = 10.^interp1([0, fin],[-100; log10(H_mag)],fout(fout<=fin(end)),'linear', 'extrap');
        else
            H_mag_i = 10.^interp1(fin, log10(H_mag), fout(fout<=fin(end)),'linear', 'extrap');
        end
        H_mag_i(fout>fin(end)) = H_mag(end);
    case 'extrap_to_DC'
        % first extrapolate down to DC, then use highest available frequency
        % for higher frequencies
        H_mag_i = 10.^interp1(fin,log10(H_mag),fout(fout<=fin(end)),'linear', 'extrap');
        H_mag_i(fout>fin(end)) = H_mag(end);
    case 'old'
        H_mag_i = interp1(fin,H_mag,fout,'linear','extrap');
    otherwise
        error('COM:Extrap:InvalidOption', 'opt_interp_Sparam_mag valid values are "old", "extrap_to_DC"');
end

H_ph_i = interp1(fin,H_ph,fout,'linear', 'extrap');

%opt_interp_Sparam_phase='trend_and_shift_to_DC';
%opt_interp_Sparam_phase='interp_cubic_to_dc_linear_to_inf';
switch opt_interp_Sparam_phase
    case 'old'
        H_ph_i = H_ph_i-H_ph_i(1);
    case 'zero_DC'
        H_ph_i(1) = 0;
    case 'interp_to_DC'
        if fin(1) ~= 0
            H_ph_i = interp1([0; fin(:)], [0; H_ph(:)], fout, 'linear', 'extrap');
        end
    case 'extrap_cubic_to_dc_linear_to_inf'
        if fin(1) ~= 0
            % estimate low frequency group delay
            group_delay = -diff(H_ph(:))./diff(fin(:));
            low_freq_gd = group_delay(1:50);
            %  calculate trend, throwing away outliers
            m = median(low_freq_gd); sigma = std(low_freq_gd);
            lf_trend = mean(low_freq_gd(abs(low_freq_gd-m)<sigma));
            % correct outliers in first 10 phase samples
            for k=10:-1:1
                H_ph(k) = H_ph(k+1) + lf_trend*(fin(k+1)-fin(k));
            end
            H_ph_cubic = interp1(fin, H_ph, fout, 'pchip', 'extrap');
            H_ph_linear = interp1(fin, H_ph, fout, 'linear', 'extrap');
            % modification - trend to inf
            if (1)
                high_freq_gd = group_delay(end-50:end);
                %  calculate trend, throwing away outliers
                m = median(high_freq_gd); sigma = std(high_freq_gd);
                hf_trend = -mean(high_freq_gd(abs(high_freq_gd-m)<sigma));
                hf_extrap_range = find(fout>fin(end));
                last_data_sample = hf_extrap_range(1)-1;
                H_ph_linear(hf_extrap_range) = H_ph_linear(last_data_sample) + (fout(hf_extrap_range)-fout(last_data_sample))*hf_trend;
                %                for k=hf_range
                %                    H_ph_linear(k) = H_ph_linear(k-1) + hf_trend*(fout(k)-fout(k-1));
                %                end
            end
            
            [UNUSED_OUTPUT, indx] = min(abs(H_ph_cubic-H_ph_linear)); %#ok<ASGLU>
            H_ph_i=H_ph_cubic;
            H_ph_i(indx:end) = H_ph_linear(indx:end);
            H_ph_i = H_ph_linear; % John Ewen 12/13/2019
        end
    case 'interp_and_shift_to_DC'
        if fin(1) ~= 0
            dc_phase_trend = H_ph(1)-(H_ph(2)-H_ph(1))/(fin(2)-fin(1))*fin(1);
            H_ph_i = interp1([0; fin(:)], [0; H_ph(:)-dc_phase_trend], fout, 'linear', 'extrap');
        end
    case 'trend_and_shift_to_DC'
        % estimate low frequency group delay
        group_delay = -diff(H_ph(:))./diff(fin(:));
        low_freq_gd = group_delay(1:50);
        %  calculate trend, throwing away outliers
        m = median(low_freq_gd); sigma = std(low_freq_gd);
        lf_trend = mean(low_freq_gd(abs(low_freq_gd-m)<sigma));
        fin_x=fin;
        H_ph_x=H_ph(:);
        if fin(1) ~= 0
            % correct outliers in first 10 phase samples
            for k=10:-1:1
                H_ph(k) = H_ph(k+1) + lf_trend*(fin(k+1)-fin(k));
            end
            
            % shift all phase data so that DC extrapolation to 0 follows trend
            dc_phase_trend = H_ph(1)+lf_trend*(fin(1)-0);
            fin_x=[0, fin_x];
            H_ph_x=[0; H_ph(:)-dc_phase_trend];
        end
        % Modification: extrapolate using trend. (interp1 with "extrap" extrapolates using just
        % the last two samples, so noise can create an inverted slope and
        % non-causal response).
        if fout(end)>fin(end)
            group_delay = -diff(H_ph_x(:))./diff(fin_x(:));
            % p=polyfit(fin_x', H_ph_x, 1);
            hf_phase_trend = H_ph_x(end)-median(group_delay)*(max(fout)-max(fin_x));
            % hf_phase_trend=polyval(p,max(fout));
            fin_x=[fin_x, fout(end)];
            H_ph_x=[H_ph_x; hf_phase_trend];
        end
        H_ph_i = interp1(fin_x, H_ph_x, fout, 'linear', 'extrap');
        
    otherwise
        error('COM:Extrap:InvalidOption', ...
            'debug_interp_Sparam valid values are "old", "zero_DC", "interp_to_DC", "interp_and_shift_to_DC", "trend_and_shift_to_DC", "interp_cubic_to_dc_linear_to_inf"');
end
H_i = H_mag_i.*exp(1j*H_ph_i);
Sout=H_i;

function[ s11out, s12out, s21out, s22out ] = makeTXpkg(faxis,param)
% create tx package for PTDR.
R_diepad = param.R_diepad;
C_diepad = param.C_diepad;
C_pkg_board = param.C_pkg_board;
% [ahealey] Unpack optional compensating L and "bump" C model parameters.
L_comp = [0, 0];
C_bump = [0, 0];
if isfield(param, 'L_comp')
    L_comp = param.L_comp;
end
if isfield(param, 'C_bump')
    C_bump = param.C_bump;
end
if isscalar(L_comp), L_comp = L_comp*[1, 1]; end
if isscalar(C_bump), C_bump = C_bump*[1, 1]; end
% [ahealey] End of modifications.
[~, mele]=size(param.z_p_next_cases);
switch mele
    case 1
        Zc1(1)=param.pkg_Z_c(1);% tx impedance
        Zc1(2)=param.pkg_Z_c(2);% rx
        Zp1(1)=param.Pkg_len_TX;% tx mm (length)
        Zp1(2)=param.Pkg_len_RX;% rx
        % [ahealey] Add series L shunt C to the on-die termination model.
        % [ s11out, s12out, s21out, s22out ]= make_pkg(faxis, Zp1(1), C_diepad(1), C_pkg_board(1),Zc1(1), param);
        [ s11out, s12out, s21out, s22out ]= make_pkg(faxis, Zp1(1), C_diepad(1), C_pkg_board(1),Zc1(1), param, L_comp(1), C_bump(1));
        % [ahealey] End of modification.
    case 4
        Zc1=[ param.pkg_Z_c(1,1) param.pkg_Z_c(2,1)] ;% Tx Rx impedance
        Zp1=[ param.Pkg_len_TX(1) param.Pkg_len_RX(1)] ;% Tx Rx length
        Zc2 =[ param.pkg_Z_c(1,2) param.pkg_Z_c(2,2)] ;
        Zp2 = [ param.Pkg_len_TX(2) param.Pkg_len_RX(2) ] ;
        Zc3 =[ param.pkg_Z_c(1,3) param.pkg_Z_c(2,3)] ;
        Zp3 = [ param.Pkg_len_TX(3) param.Pkg_len_RX(3) ] ;
        Zc4 = [ param.pkg_Z_c(1,4) param.pkg_Z_c(2,4)] ;
        Zp4 = [ param.Pkg_len_TX(4) param.Pkg_len_RX(4) ] ;
        Cv = [ param.C_v(1) param.C_v(2)  ];
        %
        % [ahealey] Add series L shunt C to the on-die termination model.
        % [ s11outa, s12outa, s21outa, s22outa ]= make_pkg(faxis, Zp1(1),  C_diepad(1),0 ,Zc1(1), param);
        [ s11outa, s12outa, s21outa, s22outa ]= make_pkg(faxis, Zp1(1),  C_diepad(1),0 ,Zc1(1), param, L_comp(1), C_bump(1));
        % [ahealey] End of modification.
        %
        [ s11outb, s12outb, s21outb, s22outb ]= make_pkg(faxis, Zp2(1) , 0, 0 ,Zc2(1), param); % add pth
        [ s11outc, s12outc, s21outc, s22outc ]  = combines4p(  s11outa, s12outa, s21outa, s22outa, s11outb, s12outb, s21outb, s22outb    );
        %
        %
        [ s11outd, s12outd, s21outd, s22outd ]= make_pkg(faxis, Zp3(1) , 0, Cv(1) ,Zc3(1), param); % add pth
        [ s11oute, s12oute, s21oute, s22oute ]  = combines4p(  s11outc, s12outc, s21outc, s22outc, s11outd, s12outd, s21outd, s22outd    );
        %
        [ s11outf, s12outf, s21outf, s22outf ]= make_pkg(faxis, Zp4(1) , 0, C_pkg_board(1), Zc4(1), param); % add pth
        [ s11out, s12out, s21out, s22out ]  = combines4p(  s11oute, s12oute, s21oute, s22oute, s11outf, s12outf, s21outf, s22outf    );
    otherwise
        error('package symtax error')
end


%% functions for estimating HEO or EW
function [ s11out, s12out, s21out, s22out ] = make_pkg(f, pkg_len, cpad, cball, pkg_z, param, varargin)
f(f<eps)=eps;
tau = param.pkg_tau; gamma0_a1_a2=param.pkg_gamma0_a1_a2; Lenscale=pkg_len; zref=param.Z0;
%% Equation 93A-8
s11pad= -1i*2*pi.*f*cpad*zref./(2+1i*2*pi.*f*cpad*zref);
s21pad= 2./(2+1i*2*pi.*f*cpad*zref);

% [ahealey] Add compensating L and shunt C (bump) when requested.
s12pad = s21pad;
s22pad = s11pad;
if nargin > 6
    lcomp = varargin{1};
    s11comp = (1i*2*pi*f*lcomp/zref)./(2+1i*2*pi*f*lcomp/zref);
    s21comp = 2./(2+1i*2*pi*f*lcomp/zref);
    [s11pad, s12pad, s21pad, s22pad] = combines4p( ...
        s11pad, s12pad, s21pad, s22pad, ...
        s11comp, s21comp, s21comp, s11comp);
end
if nargin > 7
    cbump = varargin{2};
    s11bump = -1i*2*pi.*f*cbump*zref./(2+1i*2*pi.*f*cbump*zref);
    s21bump = 2./(2+1i*2*pi.*f*cbump*zref);
    [s11pad, s12pad, s21pad, s22pad] = combines4p( ...
        s11pad, s12pad, s21pad, s22pad, ...
        s11bump, s21bump, s21bump, s11bump);
end
% [ahealey] End of modifications.

[ S11, S12, S21, S22 ] = synth_tline(f, pkg_z, zref, gamma0_a1_a2, tau, Lenscale); %#ok<NASGU,ASGLU>
% [ahealey] Symmetry cannot be assumed with more complex termination models.
% [ s11out1, s12out1, s21out1, s22out1 ]= ...
%     combines4p(  s11pad, s21pad, s21pad, s11pad, S11, S21, S21, S11 ); % first part of equation 93A-15
[s11out1, s12out1, s21out1, s22out1] = combines4p( ...
    s11pad, s12pad, s21pad, s22pad, ...
    S11, S21, S21, S11);
% [ahealey] End of modifications.

%% Equation 93A-8
s11ball= -1i*2*pi.*f*cball*zref./(2+1i*2*pi.*f*cball*zref);
s21ball= 2./(2+1i*2*pi.*f*cball*zref);
[ s11out, s12out, s21out, s22out ]= ...
    combines4p( s11out1, s12out1, s21out1, s22out1, s11ball, s21ball, s21ball, s11ball );% second part of equation 93A-15

function missingParameter (parameterName)
error( 'error:badParameterInformation', ...
    'The data for mandatory parameter %s is missing or incorrect' , parameterName);

function pdf = normal_dist(sigma,nsigma,binsize)
pdf.BinSize=binsize;
pdf.Min=-round(nsigma*sigma/binsize);
pdf.x=(pdf.Min:-pdf.Min)*binsize;
pdf.y=exp(-pdf.x.^2/(2*sigma^2+eps));
pdf.y=pdf.y/sum(pdf.y);

function result=optimize_fom(OP, param, chdata, sigma_bn,do_C2M)
%% input
% chdata(1).uneq_imp_response is the impulse response input expected to be normalized to At, peak drive voltage
% baud_rate - baud rate in seconds
% param.samples_per_ui = samples per UI of IR
% param.max_ctle - maximum ac to dc gain in dB
% param.tx_ffe(1) - maximum pre cursor (positive value)
% param.tx_ffe(2) - maximum post cursor (positive value)
% param.tx_ffe_step - sweep step size for tx pre and post taps
% param.ndfe - number of reference dfe taps
% do_C2M.  set to 0 for standard optimize_fom.  set to 1 for optimize_fom_for_C2M
% output
% result.eq.txle - [ precusor curosr postcursor]: pre and post are negative
% result.eq.ctle - index of CTLE parameters in table
% result.IR - impulse response
% result.avail_signal - maximum signal after equalization
% result.avail_sig_index - index in result.IR of max signal
% result.best_FOM - best raw ISI


min_number_of_UI_in_response=40;
baud_rate=1/param.ui;
% H_r = 1./polyval([1 2.613126 3.414214 2.613126 1], 1i*chdata(1).faxis/(param.f_r*param.fb));
f=chdata(1).faxis;
param.fb_BW_cutoff=param.f_r;
factor_3db=0.473037;
param.fb_BT_cutoff=factor_3db*param.f_r;
if OP.Bessel_Thomson
    a = bessel( param.BTorder );
    acoef=fliplr( a );
    H_bt =a(1)./ polyval(acoef, (1i*f./(param.fb_BT_cutoff*param.fb)));
else
    H_bt=ones(1,length(f));
end
if OP.Butterworth
    H_bw = 1./polyval([1 2.613126 3.414214 2.613126 1], 1i*f./(param.fb_BW_cutoff*param.fb));
else
    H_bw=ones(1,length(f));
end
H_r=H_bw.*H_bt;

% system noise H_sy PSD
if OP.USE_ETA0_PSD
    fspike=1e9;
    H_sy=sinc(sqrt(2)*(chdata(1).faxis-fspike)/(fspike)).^2;
else
    H_sy=ones(1,length(chdata(1).faxis));
end
cm4_values = param.tx_ffe_cm4_values;
cm3_values = param.tx_ffe_cm3_values;
cm2_values = param.tx_ffe_cm2_values;
cm1_values = param.tx_ffe_cm1_values;
cp1_values = param.tx_ffe_cp1_values;
cp1_values = param.tx_ffe_cp1_values;
gdc_values = param.ctle_gdc_values;
Gffe_values = param.cursor_gain;
switch param.CTLE_type
    case 'CL93'
    case 'CL120d'
        g_DC_HP_values =param.g_DC_HP_values;
    case 'CL120e'
        f_HP_Z=param.f_HP_Z;
        f_HP_P=param.f_HP_P;
        
end
best_ctle = [];
best_FOM = -inf;
best_txffe = [];
delta_sbr = [];
best_bmax=param.bmax;
%AJG021820
best_bmin=param.bmin;
h_J=[];
pxi=0;
if OP.DISPLAY_WINDOW
    hwaitbar=waitbar(0);
else
    fprintf('FOM search');
end
FOM=0;
if ~OP.RxFFE
    Gffe_values=0;
end
param.ndfe_passed=param.ndfe;
old_loops=0;
new_loops=0;

%GDC Qual construction
gqual= param.gqual;
g2qual=param.g2qual;
if ~strcmp(param.CTLE_type,'CL120d')
    qual=ones(1,length(gdc_values));
else
    if isempty(gqual) && isempty(g2qual)
        qual=ones(length(g_DC_HP_values),length(gdc_values));
    else
        qual=zeros(length(g_DC_HP_values),length(gdc_values));
        
        %prepare gqual and g2qual
        [g2qual,si]=sort(g2qual,'descend');
        gqual=gqual(si,:);
        tmp=g2qual;
        g2qual=zeros(length(tmp),2);
        for kk=1:length(tmp)
            if kk==1
                g2qual(kk,:)=[tmp(kk)+eps tmp(kk)];
            else
                g2qual(kk,:)=[tmp(kk-1) tmp(kk)];
            end
            gqual(kk,:)=sort(gqual(kk,:),'descend');
        end
        
        %Qual Construction
        for jj=1:length(g_DC_HP_values)
            for ii=1:length(gdc_values)
                for kk=1:size(gqual,1)
                    if g_DC_HP_values(jj) >= g2qual(kk,2) && g_DC_HP_values(jj) < g2qual(kk,1)
                        if gdc_values(ii) >= gqual(kk,2) && gdc_values(ii) < gqual(kk,1)
                            qual(jj,ii)=1;
                            break;
                        end
                    end
                end
            end
        end
    end
end

progress_interval=0.025;
if do_C2M
    loop_count=[1 2];
    T_O=floor((param.T_O/1000)*param.samples_per_ui);
    T_O=max(0,T_O);
else
    loop_count=1;
    T_O=0;
end

for i=loop_count
    
    for Gffe_index=1:length(Gffe_values)
        param.current_ffegain=Gffe_values(Gffe_index);
        for ctle_index=1:length(gdc_values)
            g_dc = gdc_values(ctle_index);
            kacdc = 10^(g_dc/20);
            CTLE_fp1 = param.CTLE_fp1(ctle_index);
            CTLE_fp2 = param.CTLE_fp2(ctle_index);
            CTLE_fz = param.CTLE_fz(ctle_index);
            switch param.CTLE_type
                case 'CL93'
                    lf_indx=1;
                case 'CL120d'
                    lf_indx=length(g_DC_HP_values);
                case 'CL120e'
                    lf_indx=1;
                    HP_Z = param.f_HP_Z(ctle_index);
                    HP_P = param.f_HP_P(ctle_index);
            end
            for  g_LP_index=1:lf_indx
                
                %GDC Qual Check
                if qual(g_LP_index,ctle_index)==0
                    pxi=pxi+length(cm2_values)* length(cm1_values)*length(cp1_values)*length(cm3_values)*length(cm4_values);
                    continue;
                end
                
                ctle_gain = (kacdc + 1i*chdata(1).faxis/CTLE_fz) ./ ...
                    ((1+1i*chdata(1).faxis/CTLE_fp1).*(1+1i*chdata(1).faxis/CTLE_fp2));
                switch param.CTLE_type
                    case 'CL93'
                        H_low=1;
                        kacde_DC_low=1;
                    case 'CL120d'
                        g_DC_low = g_DC_HP_values(g_LP_index);
                        f_HP=param.f_HP(g_LP_index);
                        kacde_DC_low = 10^(g_DC_low/20);
                        H_low=(kacde_DC_low +  1i*chdata(1).faxis/f_HP)./(1 + 1i*chdata(1).faxis/f_HP);
                    case 'CL120e' % z1 has been adusted on read in
                        H_low=(1 +  1i*chdata(1).faxis/HP_Z)./(1 + 1i*chdata(1).faxis/HP_P);
                end
                H_ctf=H_low.*ctle_gain;
                if OP.TDMODE
                    uneq_field='uneq_pulse_response';
                    ctle_field='ctle_pulse_response';
                else
                    uneq_field='uneq_imp_response';
                    ctle_field='ctle_imp_response';
                end
                if OP.INCLUDE_CTLE==1
                    for k=1:param.num_s4p_files
                        ir_peak = max(abs(chdata(k).(uneq_field)));
                        ir_last  = find(abs(chdata(k).(uneq_field))>ir_peak*OP.impulse_response_truncation_threshold, 1, 'last');
                        chdata(k).(uneq_field) = chdata(k).(uneq_field)(1:ir_last);
                        chdata(k).(ctle_field) = TD_CTLE(chdata(k).(uneq_field), baud_rate ...
                            , CTLE_fz, CTLE_fp1, CTLE_fp2, g_dc, param.samples_per_ui);
                        switch param.CTLE_type
                            case 'CL93'
                            case 'CL120d'
                                chdata(k).(ctle_field) = TD_CTLE(chdata(k).(ctle_field), baud_rate, f_HP, f_HP,100e100 , g_DC_low , param.samples_per_ui);
                            case 'CL120e' % z1 has been adusted on read in
                                chdata(k).(ctle_field) = TD_CTLE(chdata(k).(ctle_field), baud_rate, HP_Z,HP_P,100e100 , 0 , param.samples_per_ui);
                        end
                    end
                    
                else
                    for k=1:param.num_s4p_files
                        chdata(k).(ctle_field) = chdata(k).(uneq_field);
                    end
                end
                for k=1:param.num_s4p_files
                    chdata(k).sdd21ctf=chdata(k).sdd21.*H_ctf; % sdd21 is a VTF, includes H_t, H_f, and package
                end
                %% Equation 93A-22 %%
                %         figure(1000)
                %         semilogx(chdata(1).faxis/1e9,db(H_ctf))
                %         hold on
                if OP.RX_CALIBRATION
                    ctle_gain2 = (kacdc + 1i*chdata(2).faxis/CTLE_fz) ./ ...
                        ((1+1i*chdata(2).faxis/CTLE_fp1).*(1+1i*chdata(2).faxis/CTLE_fp2));
                    switch param.CTLE_type
                        case 'CL93'
                            H_low2=1;
                        case 'CL120d'
                            g_DC_low = g_DC_HP_values(g_LP_index);
                            f_HP=param.f_HP(g_LP_index);
                            kacde_DC_low = 10^(g_DC_low/20);
                            H_low2=(kacde_DC_low +  1i*chdata(1).faxis/f_HP)./(1 + 1i*chdata(1).faxis/f_HP);
                        case 'CL120e' % z1 has been adusted on read in
                            H_low2=(1 +  1i*chdata(1).faxis/HP_Z)./(1 + 1i*chdata(1).faxis/HP_P);
                    end
                    H_ctf2=H_low2.*ctle_gain2;
                end
                % RIM 11-30-2020 moved to a subfunction
                [sigma_N] = get_sigma_eta_ACCM_noise(chdata,param,H_sy,H_r,H_ctf);
                if OP.RX_CALIBRATION
                    sigma_ne = get_sigma_noise( H_ctf2, param, chdata, sigma_bn); %% Equation 93A-48 %%
                    sigma_NEXT=sqrt(param.eta_0*sum( abs(H_sy(2:end).^2 .* H_r(2:end).*2 .* H_ctf(2:end).^2 ) .* diff(chdata(1).faxis)/chdata(1).faxis(end)));
                else
                    %% Equations 93A-33 and 93A-34  for NEXT - independent of TXFFE setting %%
                    sigma_NEXT =  get_xtlk_noise( [0 1 0], 'NEXT', param, chdata );
                    sigma_ne=0;
                end
                Chold=[];
                if param.GDC_MIN ~= 0 && min(gdc_values(ctle_index)) + min(g_DC_HP_values(g_LP_index)) < param.GDC_MIN
                    pxi=pxi+length(cm2_values)* length(cm1_values)*length(cp1_values)*length(cm3_values)*length(cm4_values);
                    continue;
                end
                for k_cm4=1:length(cm4_values)
                    cm4=cm4_values(k_cm4);
                    for k_cm3=1:length(cm3_values)
                        cm3=cm3_values(k_cm3);
                        for k_cm2=1:length(cm2_values)
                            cm2=cm2_values(k_cm2);
                            for k_cm1=1:length(cm1_values)
                                cm1=cm1_values(k_cm1);
                                for k_cp1=1:length(cp1_values)
                                    
                                    
                                    cp1=cp1_values(k_cp1);
                                    pxi=pxi+1;
                                    runs=(length(cm2_values)* length(cm1_values)*length(cp1_values)*length(gdc_values)*length(cm3_values)*length(cm4_values)*lf_indx*length(Gffe_values) );
                                    progress = pxi/runs;
                                    if OP.DISPLAY_WINDOW
                                        if ~mod(pxi,floor(runs*progress_interval))
                                            waitbar(progress, hwaitbar, 'Linear equalization tuning'); figure(hwaitbar); drawnow;
                                        end
                                    else
                                        if ~mod(pxi,floor(runs*progress_interval)), fprintf('%i%% ', round(progress*100) );end
                                    end
                                    cur=5;
                                    if length(cm4_values) == 1  && cm4_values == 0; cm4c =[];cur=cur-1; else cm4c=cm4; end
                                    if length(cm3_values) == 1  && cm3_values == 0; cm3c =[];cur=cur-1;else cm3c=cm3;end
                                    if length(cm2_values) == 1  && cm2_values == 0; cm2c =[];cur=cur-1;else cm2c=cm2;end
                                    if length(cm1_values) == 1  && cm1_values == 0; cm1c =[];cur=cur-1;else cm1c=cm1;end
                                    if length(cp1_values) == 1  && cp1_values == 0; cp1c =[];else cp1c=cp1;end
                                    txffe = [cm4c, cm3c, cm2c, cm1c, 1-abs(cm3)-abs(cm4)-abs(cm2)-abs(cm1)-abs(cp1), cp1c];
                                    % Skip combinations with small values of c(0), not guaranteed to be supported by all transmitters.
                                    if txffe(cur)<param.tx_ffe_c0_min
                                        continue;
                                    end
                                    old_loops=old_loops+1;
                                    % speedup "local search" heuristic - Adee Ran 03-17-2020
                                    % skip configurations more than 2 steps away from current "best" point on any grid direction
                                    if param.LOCAL_SEARCH>0 && ~isinf(best_FOM) && ...
                                            ((k_cm1>1 && length(cp1_values)>1 && abs(k_cp1-find(cp1_values==best_txffe(cur+1)))>param.LOCAL_SEARCH) ...
                                            || (k_cm2>1 && length(cm1_values)>1 && abs(k_cm1-find(cm1_values==best_txffe(cur-1)))>param.LOCAL_SEARCH) ...
                                            || (k_cm3>1 && length(cm2_values)>1 && abs(k_cm2-find(cm2_values==best_txffe(cur-2)))>param.LOCAL_SEARCH) ...
                                            || (k_cm4>1 && length(cm3_values)>1 && abs(k_cm3-find(cm3_values==best_txffe(cur-3)))>param.LOCAL_SEARCH) ...
                                            || (g_LP_index>1 && length(cm4_values)>1 && abs(k_cm4-find(cm4_values==best_txffe(cur-4)))>param.LOCAL_SEARCH) ...
                                            || (ctle_index>1 && abs(g_LP_index-best_G_high_pass)>param.LOCAL_SEARCH))
                                        
                                        continue;
                                    end
                                    new_loops=new_loops+1;
                                    if OP.TDMODE
                                        [ sbr ] = FFE( txffe , cur-1, param.samples_per_ui, chdata(1).ctle_pulse_response );
                                    else
                                        [ effective_channel ] = FFE( txffe , cur-1, param.samples_per_ui, chdata(1).ctle_imp_response );
                                        %                             upsampled_txffe = zeros(1, param.samples_per_ui*5+1); % start with zeros everywhere
                                        %                             upsampled_txffe(1+(0:5)*param.samples_per_ui) = txffe; % "plant" the coefficients in the desired locations
                                        %                             effective_channel = filter(upsampled_txffe, 1, chdata(1).ctle_imp_response);
                                        %sbr=filter(ones(param.samples_per_ui, 1), 1, effective_channel);
                                        %"conv2" is faster than filter. Just need to chop off extra points at the end
                                        sbr=conv2(effective_channel,ones(param.samples_per_ui, 1));
                                        sbr=sbr(1:end-param.samples_per_ui+1);
                                    end
                                    sbr1=sbr;
                                    [UNUSED_OUTPUT, sbr_peak_i]=max(abs(sbr)); %#ok<ASGLU>
                                    triple_transit_time = round(sbr_peak_i*2/param.samples_per_ui)+20;
                                    if min_number_of_UI_in_response < triple_transit_time
                                        min_number_of_UI_in_response = triple_transit_time;
                                    end
                                    
                                    %% initial guess at cursor location (t_s)  - based on approximate zero crossing
                                    %[UNUSED_OUTPUT, sbr_peak_i]=max(abs(sbr)); %#ok<ASGLU>
                                    %zxi = find(diff(sign(sbr-.01*max(sbr)))>=1);
                                    %this statement is slow.  speed up by searching (1:sbr_peak_i)
                                    zxi = find(diff(sign(sbr(1:sbr_peak_i)-.01*max(sbr)))>=1);
                                    %zxi = zxi(zxi<sbr_peak_i);
                                    zxi = zxi(sbr_peak_i - zxi < 4*param.samples_per_ui);
                                    if isempty(zxi)
                                        continue;
                                    elseif length(zxi)>1
                                        zxi=zxi(end);
                                    end
                                    if param.ndfe==0
                                        max_dfe1=0;
                                    else
                                        max_dfe1=param.bmax(1);
                                    end
                                    %% adjust cursor_i to Solve equation 93A-25 %%
                                    % Muller-Mueller criterion with DFE
                                    mm_range = zxi+(0:2*param.samples_per_ui);
                                    switch OP.CDR
                                        case 'Mod-MM'
                                            mm_metric = ...
                                                abs(sbr(mm_range+param.samples_per_ui)-max_dfe1*sbr(mm_range));
                                        otherwise % MM
                                            mm_metric = ...
                                                abs(sbr(mm_range-param.samples_per_ui) - max(sbr(mm_range+param.samples_per_ui)-max_dfe1*sbr(mm_range), 0));
                                    end
                                    [UNUSED_OUTPUT, mm_cursor_offset] = min(mm_metric); %#ok<ASGLU>
                                    cursor_i = zxi+mm_cursor_offset-1;
                                    
                                    cursor = sbr(cursor_i);
                                    
                                    if OP.RxFFE
                                        % [ sbr, C]=force(sbr,param,OP,cursor_i);
                                        [ sbr, C]=force(sbr,param,OP,zxi+param.samples_per_ui);
                                        if isrow(sbr), sbr=sbr';end
                                        Ccur_i=param.RxFFE_cmx+1;
                                        if C(Ccur_i) < param.ffe_main_cursor_min, continue; end
                                        if param.ffe_post_tap_len ~=0
                                            if abs(C(Ccur_i +1)) > param.ffe_post_tap1_max, continue; end
                                            if (param.ffe_post_tap_len > 1)
                                                if sum(abs(C((Ccur_i +2):end))  > param.ffe_tapn_max) , continue; end
                                            end
                                        end
                                        if param.ffe_pre_tap_len ~=0
                                            if abs(C(Ccur_i -1)) > param.ffe_pre_tap1_max, continue; end
                                            if (param.ffe_pre_tap_len > 1)
                                                %                                             if sum(abs(C((Ccur_i +2):end)) > param.ffe_tapn_max) , continue; end
                                                if sum(abs(C(1:(Ccur_i - 2))) > param.ffe_tapn_max) , continue; end % 11.22.2018 Yasou Hadaka
                                            end
                                        end
                                        
                                        %% second guess at cursor location (t_s)  - based on approximate zero crossing
                                        %This entire block is now inside the "if OP.RxFFE" to avoid unnecessary re-calculation
                                        [UNUSED_OUTPUT, sbr_peak_i]=max(abs(sbr)); %#ok<ASGLU>
                                        %zxi = find(diff(sign(sbr-.01*max(sbr)))>=1);
                                        zxi = find(diff(sign(sbr(1:sbr_peak_i)-.01*max(sbr)))>=1);
                                        %zxi = zxi(zxi<sbr_peak_i);
                                        zxi = zxi(sbr_peak_i - zxi < 4*param.samples_per_ui);
                                        if isempty(zxi)
                                            continue;
                                        elseif length(zxi)>1
                                            zxi=zxi(end);
                                        end
                                        if param.ndfe==0
                                            max_dfe1=0;
                                        else
                                            max_dfe1=param.bmax(1);
                                        end
                                        %% adjust cursor_i to Solve equation 93A-25 %%
                                        % Muller-Mueller criterion with DFE
                                        mm_range = zxi+(0:2*param.samples_per_ui);
                                        switch OP.CDR
                                            case 'Mod-MM'
                                                mm_metric = ...
                                                    abs(sbr(mm_range+param.samples_per_ui)-max_dfe1*sbr(mm_range));
                                            otherwise % MM
                                                mm_metric = ...
                                                    abs(sbr(mm_range-param.samples_per_ui) - max(sbr(mm_range+param.samples_per_ui)-max_dfe1*sbr(mm_range), 0));
                                        end
                                        [UNUSED_OUTPUT, mm_cursor_offset] = min(mm_metric); %#ok<ASGLU>
                                        cursor_i = zxi+mm_cursor_offset-1;
                                        
                                        cursor = sbr(cursor_i);
                                    end
                                    A_p=sbr(sbr_peak_i);
                                    %% 93A.1.6 step c defines A_s %%
                                    A_s = param.R_LM*cursor/(param.levels-1);
                                    if isempty(delta_sbr)
                                        delta_sbr = sbr;
                                    end
                                    sbr=sbr(:);
                                    %% Equation 93A-27 "otherwise" case %% param.N_bmax is param.ndfe if groups are not used
                                    
                                    param.ndfe=param.N_bmax;
                                    far_cursors = sbr(cursor_i-T_O+param.samples_per_ui*(param.ndfe+1):param.samples_per_ui:end);
                                    t=((cursor_i+param.samples_per_ui*(param.ndfe+1):param.samples_per_ui:length(sbr))-(cursor_i+param.samples_per_ui*(param.ndfe+1)))*...
                                        param.ui/param.samples_per_ui;
                                    precursors = sbr(cursor_i-param.samples_per_ui:-param.samples_per_ui:1);
                                    precursors = precursors(end:-1:1);
                                    
                                    %                                 % Error message if the sbr is not long enough for the specified range of Nb
                                    %                                 if length(sbr) < cursor_i+param.samples_per_ui*(param.ndfe+1)
                                    %                                     close(hwaitbar);
                                    %                                     error('Pulse Response contains %d samples after the cursor. Specified Nb requires %d samples after the cursor.' ...
                                    %                                         , length(sbr)-cursor_i, param.samples_per_ui*(param.ndfe+1));
                                    %                                 end
                                    
                                    %% Equation 93A-27, when 1<=n<=N_b
                                    dfecursors=sbr(cursor_i+param.samples_per_ui*(1):param.samples_per_ui:cursor_i+param.samples_per_ui*(param.ndfe));
                                    if param.dfe_delta ~= 0
                                        dfecursors_q=floor(abs(dfecursors/sbr(cursor_i))./param.dfe_delta).*param.dfe_delta.*sign(dfecursors)*sbr(cursor_i);
                                        
                                    else
                                        dfecursors_q=dfecursors;
                                    end
                                    if param.N_bg ~=0
                                        %% floating taps
                                        postcurors= sbr(cursor_i+param.samples_per_ui:param.samples_per_ui:end);
                                        
                                        [floating_tap_locations, floating_tap_coef, hisi, bmax]= floatingDFE( postcurors ,param.ndfe_passed,param.N_bf,param.N_bg,param.N_bmax, param.bmaxg, sbr(cursor_i), param.dfe_delta  );
                                        newbmax= [ param.bmax bmax(param.ndfe_passed+1:param.N_bmax)].';
                                        param.use_bmax=newbmax;
                                        %AJG021820
                                        param.use_bmin=[param.bmin bmax(param.ndfe_passed+1:param.N_bmax)*-1].';
                                    else
                                        param.use_bmax=param.bmax;
                                        %AJG021820
                                        param.use_bmin=param.bmin;
                                    end
                                    
                                    %AJG021820
                                    actual_dfecursors=dfe_clipper(dfecursors_q,sbr(cursor_i)*param.use_bmax(:),sbr(cursor_i)*param.use_bmin(:));
                                    if do_C2M
                                        dfecursors_windowed=sbr(cursor_i-T_O+param.samples_per_ui*(1):param.samples_per_ui:cursor_i+param.samples_per_ui*(param.ndfe)-T_O);
                                        % readjust SBR
                                        if 0
                                            %PR_DFE_center not currently used, so this is in "if 0" statement
                                            PR_DFE_center=sbr;
                                            for n=1:param.ndfe
                                                %                                                 for ix=-param.samples_per_ui/2: param.samples_per_ui/2
                                                %                                                     i_sample=ix+n*param.samples_per_ui+cursor_i;
                                                %                                                     dper=sbr(i_sample)- actual_dfecursors(n);
                                                %                                                     PR_DFE_center(i_sample)=dper;
                                                %                                                 end
                                                i_sample=(-param.samples_per_ui/2: param.samples_per_ui/2)+n*param.samples_per_ui+cursor_i;
                                                PR_DFE_center(i_sample)=sbr(i_sample)-actual_dfecursors(n);
                                            end
                                        end
                                        excess_dfe_cursors=dfecursors_windowed-actual_dfecursors;
                                    else
                                        excess_dfe_cursors=dfecursors-actual_dfecursors;
                                    end
                                    dfetaps=actual_dfecursors/sbr(cursor_i);
                                    
                                    if length(dfetaps) >= param.N_tail_start && param.N_tail_start ~=0
                                        tail_RSS=norm(dfetaps(param.N_tail_start:end));
                                        if  tail_RSS ~= 0
                                            if tail_RSS >= param.B_float_RSS_MAX
                                                param.use_bmax(param.N_tail_start:end)= ...
                                                    min(tail_RSS, param.B_float_RSS_MAX) * sign(dfetaps(param.N_tail_start:end)).*dfetaps(param.N_tail_start:end) /tail_RSS;
                                                %AJG021820
                                                param.use_bmin(param.N_tail_start:end)= ...
                                                    min(tail_RSS, param.B_float_RSS_MAX) * -1 * sign(dfetaps(param.N_tail_start:end)).*dfetaps(param.N_tail_start:end) /tail_RSS;
                                            end
                                        end
                                        
                                        %AJG021820
                                        actual_dfecursors=dfe_clipper(dfecursors_q,sbr(cursor_i)*param.use_bmax(:),sbr(cursor_i)*param.use_bmin(:));
                                        excess_dfe_cursors=dfecursors-actual_dfecursors;
                                        dfetaps=actual_dfecursors/sbr(cursor_i);
                                        
                                    else
                                        tail_RSS=0;
                                    end
                                    %% Eq. 93A-28 %%
                                    sampling_offset = mod(cursor_i, param.samples_per_ui);
                                    %ensure we can take early sample
                                    if sampling_offset<=1
                                        sampling_offset=sampling_offset+param.samples_per_ui;
                                    end
                                    if (OP.LIMIT_JITTER_CONTRIB_TO_DFE_SPAN)
                                        cursors_early_sample = sbr(cursor_i-1+param.samples_per_ui*(-1:param.ndfe));
                                        cursors_late_sample = sbr(cursor_i+1+param.samples_per_ui*(-1:param.ndfe));
                                    else
                                        cursors_early_sample = sbr(sampling_offset-1:param.samples_per_ui:end);
                                        cursors_late_sample = sbr(sampling_offset+1:param.samples_per_ui:end);
                                    end
                                    % ensure lengths are equal
                                    cursors_early_sample = cursors_early_sample(1:length(cursors_late_sample));
                                    h_J = (cursors_late_sample-cursors_early_sample)/2*param.samples_per_ui;
                                    
                                    %% Equation 93A-30 %%
                                    % since A_s = param.R_LM*cursor/(param.levels-1), cursor=(param.levels-1)*A_s/param.R_LM
                                    sigma_TX = (param.levels-1)*A_s/param.R_LM*10^(-param.SNR_TX/20);
                                    
                                    %% Equation 93A-31 %%
                                    sigma_ISI = param.sigma_X*norm([precursors; excess_dfe_cursors; far_cursors]);
                                    ISI_N=param.sigma_X*norm( far_cursors);
                                    %% break if FOM has no chance of beating old e
                                    OP.exe_mode=1;
                                    switch OP.EXE_MODE
                                        case 0
                                        case 1
                                            if (20*log10(A_s/sigma_ISI) < best_FOM)
                                                continue
                                            end
                                        case 2
                                            if (20*log10(A_s/sigma_ISI) < best_FOM)
                                                break
                                            end
                                    end
                                    %% Equation 93A-32 %%
                                    sigma_J = norm([param.A_DD param.sigma_RJ])*param.sigma_X*norm(h_J);
                                    
                                    %% Equations 93A-33 and 93A-34 for FEXT (depends on TXFFE setting) %%
                                    if OP.RX_CALIBRATION
                                        sigma_XT=0;
                                    else
                                        if ~OP.RxFFE
                                            % sigma_FEXT =  get_xtlk_noise( 0, 'FEXT', param ,chdata);
                                            % sigma_XT = norm([sigma_NEXT sigma_FEXT]);
                                            [sigma_XT,~,~] =  get_xtlk_noise( txffe, 'FEXT', param ,chdata); %with three outputs, the sigma_XT includes both FEXT and NEXT zhilei huang 01/11/2019
                                            %% Equation 93A-36 denominator (actually its sqrt)
                                        else % John Ewen: 13/12/20018
                                            [sigma_XT,~,~] =  get_xtlk_noise( txffe, 'FEXT', param ,chdata, C); %with three outputs, the sigma_XT includes both FEXT and NEXT zhilei huang 01/11/2019
                                            %                                         sigma_FEXT_ffe=  get_xtlk_noise( 0, 'FEXT', param, chdata, C );
                                            %                                         if ~OP.RxFFE
                                            %                                             sigma_NEXT_ffe =  get_xtlk_noise(0, 'NEXT', param, chdata);
                                            %                                         else
                                            %                                             sigma_NEXT_ffe =  get_xtlk_noise(0, 'NEXT', param, chdata , C);
                                            %                                         end
                                            %                                         sigma_XT = norm([sigma_NEXT_ffe sigma_FEXT_ffe])*sqrt((param.levels^2-1)/(3*(param.levels-1)^2));
                                        end
                                    end
                                    if OP.RxFFE % modify sigma_N with rx noise from the rx ffe
                                        index_f2=find(chdata(1).faxis(:)>param.fb,1,'first');
                                        if isempty(index_f2), index_f2=length(chdata(1).faxis);end
                                        f=chdata(1).faxis;
                                        H_Rx_FFE=zeros(1,length(f));
                                        for ii=-param.RxFFE_cmx:param.RxFFE_cpx
                                            H_Rx_FFE=C(ii+param.RxFFE_cmx+1).*exp(-1j*2*pi*(ii+1).*f/param.fb)+H_Rx_FFE;
                                        end
                                        sigma_N =sqrt(param.eta_0*sum(H_sy(2:end) .* abs(H_r(2:end) .* H_ctf(2:end) .*H_Rx_FFE(2:end)).^2 .* diff(chdata(1).faxis)/chdata(1).faxis(end)));
                                    end
                                    %% Equation 93A-36 (note log argument is voltage rather than power ratio)
                                    total_noise_rms  = norm([sigma_ISI sigma_J sigma_XT sigma_N sigma_TX sigma_ne]);
                                    if do_C2M
                                        if param.Noise_Crest_Factor == 0
                                            ber_q = sqrt(2)*erfcinv(2*param.specBER);
                                        else
                                            ber_q=param.Noise_Crest_Factor;
                                        end
                                        if OP.force_pdf_bin_size
                                            delta_y = OP.BinSize;
                                        else
                                            delta_y = min(A_s/1000, OP.BinSize);
                                        end
                                        ne_noise_pdf = normal_dist(0, ber_q, delta_y);
                                        cci_pdf = normal_dist(0, ber_q, delta_y);
                                        chdata(1).eq_pulse_response=sbr;
                                        tmp_result.t_s= cursor_i;
                                        tmp_result.A_s=A_s;
                                        EH_1st= 2*(A_s-erfcinv(param.specBER*2)*2/sqrt(2)*total_noise_rms);
                                        if EH_1st <=  param.Min_VEO_Test/1000 -.001
                                            %                                         sprintf( '%g.1 As ..  %g.1 EH\n',A_s*1000,EH_1st*1000)
                                            continue
                                        else
                                            %                                          sprintf( ' OK before %g As ..  %g EH\n',A_s*1000,EH_1st*1000)
                                        end
                                        Struct_Noise.sigma_N=sigma_N;
                                        Struct_Noise.sigma_TX=sigma_TX;
                                        Struct_Noise.cci_pdf=cci_pdf;
                                        Struct_Noise.ber_q=ber_q;
                                        Struct_Noise.ne_noise_pdf=ne_noise_pdf;
                                        [Left_EW,Right_EW,eye_contour,EH_T_C2M,EH_B_C2M]=COM_eye_width(chdata,delta_y,tmp_result,param,OP,Struct_Noise,1);
                                        EH=EH_T_C2M-EH_B_C2M;
                                        N_i=(A_s*2-EH)/2;
                                        if EH <= param.Min_VEO_Test/1000
                                            %                                         sprintf( 'After  As=%.1f ..  EH=%.1f  EH_1st=%.1f  \n',A_s*1000,EH*1000, EH_1st*1000)
                                            continue
                                        else
                                            %                                         sprintf( '<strong> After  As=%.1f ..  EH=%.1f  EH_1st=%.1f </strong> \n',A_s*1000,EH*1000, EH_1st*1000)
                                        end
                                        FOM =db(A_s/N_i);
                                    else
                                        FOM = 20*log10(A_s/total_noise_rms);
                                        %                                             if strfind(param.CTLE_type,'CL120e')
                                        %                                                 FOM = A_s*C(param.RxFFE_cmx+1)-total_noise_rms_ffe;
                                    end
                                    
                                    if (FOM > best_FOM)
                                        best_current_ffegain=param.current_ffegain;
                                        best_txffe = txffe;
                                        best_sbr = sbr;
                                        best_ctle = ctle_index;
                                        best_G_high_pass =g_LP_index;
                                        best_FOM = FOM;
                                        best_cursor_i = cursor_i;
                                        if ~OP.TDMODE
                                            best_IR=effective_channel;
                                        end
                                        best_sigma_N = sigma_N;
                                        best_h_J = h_J;
                                        best_A_s=A_s;
                                        best_A_p=A_p;
                                        best_ISI=ISI_N;
                                        best_bmax=param.use_bmax;
                                        %AJG021820
                                        best_bmin=param.use_bmin;
                                        best_tail_RSS=tail_RSS;
                                        best_dfetaps=dfetaps;
                                        if param.N_bg ~=0
                                            best_floating_tap_locations=floating_tap_locations;
                                            best_floating_tap_coef=floating_tap_coef;
                                        end
                                        if OP.RxFFE, best_RxFFE=C;end
                                        %                                 display(pxi);
                                    end
                                    
                                end
                            end
                        end
                    end
                end
            end
        end
    end
    if do_C2M
        if  best_FOM == -inf
            param.Min_VEO_Test=0;
        else
            break
        end
    end
end
if 0
    fprintf('old loops = %d\n',old_loops);
    fprintf('new loops = %d\n',new_loops);
    display(sprintf('\n :loops = %g',pxi))
end

if ~exist('best_cursor_i', 'var')% take last setting
    display('equalization failed')
    best_bmax=param.bmax;
    %AJG021820
    best_bmin=param.bmin;
    best_tail_RSS=0;
    best_current_ffegain=0;
    best_txffe = txffe;
    best_sbr = sbr;
    best_ctle = ctle_index;
    if OP.RxFFE, best_RxFFE=C;end
    best_G_high_pass =g_LP_index;
    best_FOM = FOM;
    best_cursor_i = cursor_i;
    if ~OP.TDMODE
        best_IR=effective_channel;
    end
    best_sigma_N = sigma_N;
    best_h_J = h_J;
    best_A_p=max(sbr);
    best_ISI=1;
    best_dfetaps= sbr( cursor_i+param.samples_per_ui*(1):param.samples_per_ui:cursor_i+param.samples_per_ui*(param.ndfe))/sbr(cursor_i) ;
    best_A_s= sbr( cursor_i);
    if param.N_bg ~=0
        best_floating_tap_locations=[];
        best_floating_tap_coef=[];
    end
    if do_C2M
        return
    end
    %     return
end

best_cursor = best_sbr(best_cursor_i);
% report during debug
PRin=filter(ones(param.samples_per_ui, 1),1, chdata(1).uneq_imp_response);
f=1e8:1e8:100e9;
param.fb_BW_cutoff=param.f_r;
factor_3db=0.473037;
param.fb_BT_cutoff=factor_3db*param.f_r;
if OP.Bessel_Thomson
    a = bessel( param.BTorder );
    acoef=fliplr( a );
    H_bt =a(1)./ polyval(acoef, (1i*f./(param.fb_BT_cutoff*param.fb)));
else
    H_bt=ones(1,length(f));
end
if OP.Butterworth
    H_bw = 1./polyval([1 2.613126 3.414214 2.613126 1], 1i*f./(param.fb_BW_cutoff*param.fb));
else
    H_bw=ones(1,length(f));
end
H_r=H_bw.*H_bt;

ctle_gain1 = (10^(gdc_values(best_ctle)/20) + 1i*f/param.CTLE_fz(best_ctle)) ./ ...
    ((1+1i*f/param.CTLE_fp1(best_ctle)).*(1+1i*f/param.CTLE_fp2(best_ctle)));

switch param.CTLE_type
    case 'CL93'
        H_low=1;
    case 'CL120d'
        H_low=(10^(param.g_DC_HP_values(best_G_high_pass)/20) +  1i*f/param.f_HP(best_G_high_pass))./(1 + 1i*f/param.f_HP(best_G_high_pass));
    case 'CL120e'
        H_low=(1 +  1i*f/param.f_HP_Z(best_ctle))./ (1 + 1i*f/param.f_HP_P(best_ctle));
end
ctle_gain=H_low.*ctle_gain1.*H_r;



lsbr=length(sbr);
t=0:param.ui/param.samples_per_ui:(lsbr-1)*param.ui/param.samples_per_ui;

sampled_best_sbr_precursors_t = (best_cursor_i/param.samples_per_ui:-1:1/param.samples_per_ui)*param.ui;
sampled_best_sbr_precursors_t = sampled_best_sbr_precursors_t(end:-1:2); % exclude cursor
sampled_best_sbr_precursors = best_sbr(round(sampled_best_sbr_precursors_t/param.ui*param.samples_per_ui));
sampled_best_sbr_postcursors_t = (best_cursor_i:param.samples_per_ui:lsbr)/param.samples_per_ui*param.ui;
sampled_best_sbr_postcursors_t = sampled_best_sbr_postcursors_t(2:end); % exclude cursor
sampled_best_sbr_postcursors = best_sbr(round(sampled_best_sbr_postcursors_t/param.ui*param.samples_per_ui));
sampled_best_sbr_dfecursors_t = (best_cursor_i/param.samples_per_ui+(1:param.ndfe_passed))*param.ui;
if param.N_bg ~=0
    sampled_best_sbr_fdfecursors_t = (best_cursor_i/param.samples_per_ui+(best_floating_tap_locations))*param.ui;
end
% apply max tap value constraint
dfe_cursors =    sampled_best_sbr_postcursors(1:param.ndfe);
dfe_SBRcursors = sampled_best_sbr_postcursors(1:param.ndfe);
if  isrow(best_bmax) == 1, best_bmax=best_bmax.';end

%AJG021820
if  isrow(best_bmin) == 1, best_bmin=best_bmin.';end
DFE_taps_mV=dfe_clipper(dfe_cursors,best_cursor*best_bmax(1:param.ndfe),best_cursor*best_bmin(1:param.ndfe));
if param.N_bg ~=0
    FDFE_taps_mV=DFE_taps_mV(best_floating_tap_locations);
end

sampled_best_sbr_postcursors(1:param.ndfe) = dfe_SBRcursors-DFE_taps_mV;
Symbol_Adj = (param.levels-1);% 3A.1.6
if OP.DEBUG ~=0
    if OP.DISPLAY_WINDOW && ~OP.RX_CALIBRATION
        % display pulse responses in one axis per test case.
        figure_name = sprintf('Equalization effect: %s', param.base);
        fig=findobj('Name', figure_name);
        if isempty(fig), fig=figure('Name', figure_name); end
        figure(fig);set(gcf,'Tag','COM');
        movegui(fig,'north')
        %figure(fig.Number);
        hax = subplot(length(OP.pkg_len_select), 1, param.package_testcase_i);
        plot(t,best_sbr/Symbol_Adj,'disp', 'Half Symbol Equalized end-to-end PR');
        hold on
        PRplt(1:param.samples_per_ui+5)=PRin(1); % line up with the ffe introduced delay
        PRplt(param.samples_per_ui+6:length(t))=PRin(1:length(t)-param.samples_per_ui-5);
        plot(t,PRplt/Symbol_Adj,'r','disp', 'Half Symbol Unequalized end-to-end PR');
        stem(t(best_cursor_i),best_sbr(best_cursor_i)/Symbol_Adj,'g','disp','Cursor (sample point)');
        title(sprintf('Case %d', param.package_testcase_i));
        ylabel('volts')
        xlabel('seconds')
        
        plot(sampled_best_sbr_precursors_t-param.ui/param.samples_per_ui, sampled_best_sbr_precursors'/Symbol_Adj, 'kx', 'disp','Pre cursors');
        plot(sampled_best_sbr_postcursors_t-param.ui/param.samples_per_ui, sampled_best_sbr_postcursors'/Symbol_Adj, 'ko', 'disp','Post cursors');
        if param.ndfe_passed ~=0
            stem(sampled_best_sbr_dfecursors_t-param.ui/param.samples_per_ui,DFE_taps_mV(1:param.ndfe_passed)/Symbol_Adj','m', 'LineWidth',2,'disp','DFE-canceled cursors');
        end
        if param.N_bg ~=0
            stem(sampled_best_sbr_fdfecursors_t-param.ui/param.samples_per_ui,FDFE_taps_mV/Symbol_Adj, 'MarkerFaceColor','red','MarkerEdgeColor','m','LineWidth',1,'disp','FDFE-canceled cursors');
        end
        grid on
        legend show
        zoom xon
        set(hax, 'tag', 'EQE');
        %
        figure(110);set(gcf,'Tag','COM');
        set(gcf, 'Name', 'CTLE selection');
        movegui(gcf, 'southeast');
        semilogx(f,20*log10(abs(ctle_gain)), 'disp', sprintf('Case %d', param.package_testcase_i));
        hold on
        fbaud_tick=find(f >= baud_rate, 1);
        fnq_tick=find(f >= baud_rate/2, 1);
        stem(f(fnq_tick),20*log10(abs(ctle_gain(fnq_tick))),'g', 'handlevisibility', 'off');
        stem(f(fbaud_tick),20*log10(abs(ctle_gain(fbaud_tick))),'g', 'handlevisibility', 'off');
        recolor_plots(gca);
        title('CTF/w Rx Filter Response')
        ylabel('dB')
        xlabel('Hz')
    end
    display(['FOM:                ' ,num2str(best_FOM, 2),' dB']);
    display(['TXFFE coefficients: ' ,mat2str(best_txffe) ] );
    display(['SNR ISI:                ' ,num2str(20*log10(best_A_p/best_ISI), 2),' dB']);
    display(['CTLE DC gain:       ' ,num2str(gdc_values(best_ctle)), ' dB']);
    display(['CTF peaking gain:  ' ,num2str(20*log10(max(abs(ctle_gain))), 2), ' dB']);
    display(['Symbol Available signal:   ' ,num2str(best_cursor/Symbol_Adj)]);
end
if OP.DISPLAY_WINDOW
    close(hwaitbar);
else
    fprintf('\n');
end

% % eq_data
result.cur=cur;
result.txffe = best_txffe;
result.ctle = best_ctle;
result.best_G_high_pass=best_G_high_pass;
result.DFE_taps = best_dfetaps; %relative
result.DFE_taps_i = best_cursor_i+(1:param.ndfe)*param.samples_per_ui;
if param.N_bg ~=0
    result.floating_tap_locations=best_floating_tap_locations;
    result.floating_tap_coef=best_floating_tap_coef;
end
result.A_s = best_A_s;
result.t_s = best_cursor_i;
result.sigma_N = best_sigma_N;
result.h_J = best_h_J;
result.FOM = best_FOM;
if ~OP.TDMODE
    result.IR = best_IR;
end
result.t=t;
result.sbr=best_sbr;
if OP.RxFFE, result.RxFFE=best_RxFFE;end
% changed RIM 4/17/2019 use sum(IR) for Vf of originl IR at N_b UI
result.A_p = max(chdata(1).uneq_pulse_response);
its=find(chdata(1).uneq_pulse_response>=max(chdata(1).uneq_pulse_response),1,'first');
try
    result.A_f = sum(chdata(1).uneq_imp_response(1:its+param.N_v*param.samples_per_ui) );
    result.A_f = sum(chdata(1).uneq_pulse_response(1:its+param.N_v*param.samples_per_ui) )/param.samples_per_ui;
catch
    result.A_f = sum(chdata(1).uneq_imp_response(1:end) );
end
result.Pmax_by_Vf=result.A_p/result.A_f;
result.ISI =best_ISI;
result.SNR_ISI=20*log10(best_A_p/best_ISI);
result.best_current_ffegain=best_current_ffegain;
result.best_bmax=best_bmax;
%AJG021820
result.best_bmin=best_bmin;
result.tail_RSS=best_tail_RSS;


function sgm = pdf2sgm(pdf)
avg = sum(pdf.x .* pdf.y);
sgm = sqrt(sum((pdf.x - avg).^2 .* pdf.y));
% end yasuo patch


%% adding tx packgage
function cdf=pdf_to_cdf(pdf)

%Transform PDF to CDF
%The CDF is natively calculated from negative-to-positive voltage.
%This only gives BER calculation for bottom eye.  Need to also
%calculate a CDF of reversed PDF to get top eye.  The final CDF is the
%min of top and bottom CDF values.
%If only interested in one side, a simple cumsum on y is all that is needed.

cdf.yB=cumsum(pdf.y);
cdf.yT=fliplr(cumsum(fliplr(pdf.y)));
cdf.y=min([cdf.yB(:) cdf.yT(:)],[],2);
cdf.x=pdf.x;
function plot_bathtub_curves(hax, max_signal, sci_pdf, cci_pdf, isi_and_xtalk_pdf, noise_pdf,jitt_pdf, combined_interference_and_noise_pdf, bin_size)
cursors = d_cpdf(bin_size,max_signal*[-1 1], [1 1]/2);
signal_and_isi_pdf = conv_fct(cursors, sci_pdf);
signal_and_xtalk_pdf = conv_fct(cursors, cci_pdf);
signal_and_channel_noise_pdf = conv_fct(cursors, isi_and_xtalk_pdf);
signal_and_system_noise_pdf = conv_fct(cursors, noise_pdf);
signal_and_system_jitt_pdf = conv_fct(cursors, jitt_pdf);
signal_and_total_noise_pdf = conv_fct(cursors, combined_interference_and_noise_pdf);
%% Added by Bill Kirkland, June 14, 2017
cursors_l = cursors; cursors_l.y(cursors_l.x>0) = 0;
cursors_r = cursors; cursors_r.y(cursors_r.x<0) = 0;
signal_and_total_noise_pdf_l = conv_fct(cursors_l, combined_interference_and_noise_pdf);
signal_and_total_noise_pdf_r = conv_fct(cursors_r, combined_interference_and_noise_pdf);

semilogy(signal_and_isi_pdf.x, abs(cumsum(signal_and_isi_pdf.y)-0.5) ,'r','Disp','ISI', 'parent', hax)
hold on
semilogy(signal_and_xtalk_pdf.x, abs(cumsum(signal_and_xtalk_pdf.y)-0.5) ,'b','Disp','Xtalk', 'parent', hax)
semilogy(signal_and_channel_noise_pdf.x, abs(cumsum(signal_and_channel_noise_pdf.y)-0.5) ,'c','Disp','ISI+Xtalk', 'parent', hax)
semilogy(signal_and_system_noise_pdf.x, abs(cumsum(signal_and_system_noise_pdf.y)-0.5) ,'m','Disp','Jitter, SNR_TX,RL_M, eta_0 noise', 'parent', hax)
semilogy(signal_and_system_jitt_pdf.x, abs(cumsum(signal_and_system_jitt_pdf.y)-0.5) ,'g','Disp','Jitter noise', 'parent', hax)

%% Added by Bill Kirkland, June 14, 2017
% modification allows bathtub curves to cross over and hence one can
% directly read the noise component.
%semilogy(signal_and_total_noise_pdf.x, abs(cumsum(signal_and_total_noise_pdf.y)-0.5) ,'k','Disp','total noise PDF', 'parent', hax)
vbt_l = abs(0.5-cumsum(signal_and_total_noise_pdf_l.y));
vbt_r = fliplr(0.5-(cumsum(fliplr(signal_and_total_noise_pdf_r.y))));
semilogy(signal_and_total_noise_pdf_l.x, vbt_l ,'k','Disp','total noise PDF left', 'parent', hax)
semilogy(signal_and_total_noise_pdf_r.x, vbt_r ,'k','Disp','total noise PDF right', 'parent', hax)

hc=semilogy(max_signal*[-1 -1 1 1], [0.5 1e-20 1e-20 0.5], '--ok');
set(get(get(hc,'Annotation'),'LegendInformation'), 'IconDisplayStyle','off');

ylabel(hax, 'Probability')
xlabel(hax, 'volts')
legend(hax, 'show')


function plot_pie_com(hax, max_signal, sci_pdf, cci_pdf, isi_and_xtalk_pdf, noise_pdf, combined_interference_and_noise_pdf, bin_size,param)
BER=param.specBER;
delta_dB=param.delta_IL;

iex.combined_interference_and_noise_pdf=find(abs(cumsum(combined_interference_and_noise_pdf.y))>= BER, 1, 'first');
iex.noise_pdf=find(abs(cumsum(noise_pdf.y))>= BER, 1, 'first');
iex.cci_pdf=find(abs(cumsum(cci_pdf.y))>= BER, 1, 'first');
iex.sci_pdf=find(abs(cumsum(sci_pdf.y))>= BER, 1, 'first');


maxn(1)=abs(sci_pdf.x(iex.sci_pdf));
maxn(2)=abs(noise_pdf.x(iex.noise_pdf));
maxn(3)=abs(cci_pdf.x(iex.cci_pdf));
maxn_tot=abs(combined_interference_and_noise_pdf.x(iex.combined_interference_and_noise_pdf));

COM=20*log10(max_signal/maxn_tot);
COM_per_noise(1:3)=COM*(maxn(1:3).^2/sum(maxn.^2));
COM_ISI=sprintf(  '%.2g%%',100*COM_per_noise(1)/sum(COM_per_noise));
COM_SYS=sprintf(  '%.2g%%',100*COM_per_noise(2)/sum(COM_per_noise));
COM_xtalk=sprintf('%.2g%%',100*COM_per_noise(3)/sum(COM_per_noise));

pfctr=exp(-0.09054*delta_dB);% less loss
mfctr=exp(0.09054*delta_dB); % more loss

%less loss
plus_maxn(1)=abs(sci_pdf.x(iex.sci_pdf))*pfctr;
plus_maxn(2)=abs(noise_pdf.x(iex.noise_pdf));
plus_maxn(3)=abs(cci_pdf.x(iex.cci_pdf))*pfctr;
% plus_maxn_tot=(maxn(1)*pfctr+maxn(2)+maxn(3)*pfctr)*maxn_tot/sum(plus_maxn);
plus_maxn_tot=norm(plus_maxn);

minus_maxn(1)=abs(sci_pdf.x(iex.sci_pdf))*mfctr;
minus_maxn(2)=abs(noise_pdf.x(iex.noise_pdf));
minus_maxn(3)=abs(cci_pdf.x(iex.cci_pdf))*mfctr;
% minus_maxn_tot=(maxn(1)*mfctr+maxn(2)+maxn(3)*mfctr)*maxn_tot/sum(minus_maxn);
minus_maxn_tot=norm(minus_maxn);

% more loss
COMp=20*log10(max_signal*pfctr/maxn_tot);
COMp_per_noise(1:3)=COMp*(plus_maxn(1:3).^2/plus_maxn_tot^2);
COMp_ISI=sprintf(  '%.2gdB',COMp_per_noise(1));
COMp_SYS=sprintf(  '%.2gdB',COMp_per_noise(2));
COMp_xtalk=sprintf('%.2gdB',COMp_per_noise(3));
% less loss
COMm=20*log10(max_signal*mfctr/maxn_tot);
COMm_per_noise(1:3)=COMm*(minus_maxn(1:3).^2/minus_maxn_tot^2);
COMm_ISI=sprintf(  '%.2gdB',COMm_per_noise(1));
COMm_SYS=sprintf(  '%.2gdB',COMm_per_noise(2));
COMm_xtalk=sprintf('%.2gdB',COMm_per_noise(3));

xax=[[delta_dB delta_dB delta_dB];[ 0 0 0 ]; [ -delta_dB -delta_dB -delta_dB]];


if(COM<0)
    return
end

labels= { ['ISI ' COM_ISI] ['System noise/jitter ' COM_SYS] [' Crosstalk ' COM_xtalk]};

% pie(COM_per_noise,labels)
% legend(labels,{'ISI COM','System noise/jitter COM','Crosstalk COM'});
% legend('show','Location','bestoutside')
nullbar= [ 0 0 0 ];
bar(xax,[nullbar ;COM_per_noise; nullbar],'stacked')
% bar(xax,[COMp_per_noise;COM_per_noise;COMm_per_noise],'stacked')
hold on
bar(xax,[[COMp 0 0 ];nullbar;nullbar],'stacked','w','barwidth',.5)
bar(xax,[nullbar ;nullbar; [0 0 COMm]],'stacked','w','barwidth',.5)

plot([-delta_dB,0, delta_dB], [param.pass_threshold ,param.pass_threshold ,param.pass_threshold ],'r')
% ax=gca;
% ax.XTickLabels = {'decreasing loss','-','increasing loss'};
set(gca,'XTickLabel', {'decreasing loss','-','increasing loss'})
grid on
legend(labels,'Location','north')
xlabel(['if loss is changed by ' sprintf('%.2gdB',delta_dB) ])
ylabel('COM (dB)')
hold off




function [chdata, param] = process_sxp(param, OP, chdata, SDDch);
num_files=length(chdata);
if ~OP.DISPLAY_WINDOW, fprintf('reading file '); end
for i=1:num_files
    if param.package_testcase_i==1  && i==1
        if OP.TDR && i==1
            if ~OP.SHOW_BRD
                S.Parameters(1,1,:)=chdata(i).sdd11_orig ;% SDDch(:,1,1);chdata(i).sdd11_raw
                if ~param.FLAG.S2P
                    S.Parameters(1,2,:)=chdata(i).sdd12_orig ;%SDDch(:,1,2);chdata(i).sdd11_raw
                    S.Parameters(2,1,:)=chdata(i).sdd21_orig ;%SDDch(:,2,1);
                    S.Parameters(2,2,:)=chdata(i).sdd22_orig ;%SDDch(:,2,2);
                    S.Impedance=100;
                    S.NumPorts=2; % rim 2/26/2019 correct from S.NumPorts=4;
                else
                    S.Impedance=100;
                    S.NumPorts=1;
                end
                S.Frequencies=chdata(i).faxis;
            else
                S.Parameters(1,1,:)=chdata(i).sdd11_raw ;% SDDch(:,1,1);chdata(i).sdd11_raw
                if ~param.FLAG.S2P
                    S.Parameters(1,2,:)=chdata(i).sdd12_raw ;%SDDch(:,1,2);chdata(i).sdd11_raw
                    S.Parameters(2,1,:)=chdata(i).sdd21_raw ;%SDDch(:,2,1);
                    S.Parameters(2,2,:)=chdata(i).sdd22_raw ;%SDDch(:,2,2);
                    S.Impedance=100;
                    S.NumPorts=2; % rim 2/26/2019 correct from S.NumPorts=4;
                else
                    S.Impedance=100;
                    S.NumPorts=1;
                end
                S.Frequencies=chdata(i).faxis;
            end
            if OP.TDR_W_TXPKG
                if OP.ERL == 2
                    error('Cannot add pacakge to s2p files. ERL==2 not supportted if TDR_W_TXPKG = 1')
                end
                R_diepad = param.R_diepad;
                C_diepad = param.C_diepad;
                C_pkg_board = param.C_pkg_board;
                pkg_Z_Tx=param.pkg_Z_c(1);
                pkg_Z_Rx=param.pkg_Z_c(2);
                % RX package length is assumed to be the same for all
                % channel types.  swap Cp and Cd for Tx. TDR_W_TXPKG only
                % for  Rx pkg
                %                     [ s11in, s12in, s21in, s22in ]= make_pkg(S.Frequencies, param.Pkg_len_RX,  C_pkg_board(2) ,C_diepad(2), pkg_Z_Rx, param);
                [ s11in, s12in, s21in, s22in ] = makeTXpkg(S.Frequencies,param);
                [ S.Parameters(1,1,:), S.Parameters(1,2,:), S.Parameters(2,1,:), S.Parameters(2,2,:)]  = ...
                    combines4p(  s11in, s12in, s21in, s22in, ...
                    S.Parameters(1,1,:), S.Parameters(1,2,:), S.Parameters(2,1,:), S.Parameters(2,2,:) );
                %                     S=sparameters(S.Parameters,S.Frequencies,100);
                S=SL(S,S.Frequencies,R_diepad(1)*2);
                chdata(i).TX_RL=S.Parameters(2,2,:);
                S.Parameters(1,1,:)=chdata(i).sdd11_orig; % when looking at Tx don't include package
            end
            
            % need to combine S wiht is page and channel
            if param.FLAG.S2P
                port_sel=1;
            else
                port_sel=[1 2];
            end
            OP.impulse_response_truncation_threshold=1e-5; %Only for TDR not returned out of "process_sxp" function
            for ipsl=1:length(port_sel) % do for both port if s4p
                for izt=1:length(param.Z_t) % do for all tdr impedances
                    param.RL_sel=port_sel(ipsl); % this used in get_TDR
                    %                     OP.interp_sparam_phase='interp_to_DC'; % better for return loss
                    %                     OP.interp_sparam_mag='trend_to_DC';
                    OP.interp_sparam_mag='linear_trend_to_DC';
                    % OP.interp_sparam_mag='extrap_to_DC_or_zero';
                    OP.interp_sparam_phase='extrap_cubic_to_dc_linear_to_inf';
                    TDR_results(izt,ipsl) = get_TDR(S, OP, param,param.Z_t(izt),ipsl);
                    %                     if param.Z_t(izt) ~=50
                    %                         fprintf('<strong> Warning: </strong> Zt not 50 ohms ERL and TDR result should be scrutinized\n');
                    %                     end
                    if ipsl ==1
                        chdata(i).TDR11(izt).ZSR=[TDR_results(izt,1).tdr];
                        chdata(i).TDR11(izt).t=TDR_results(izt,1).t;
                        chdata(i).TDR11(izt).avgZport=[TDR_results(izt,1).avgZport];
                        if OP.PTDR, chdata(i).PDTR11(izt).ptdr=TDR_results(izt,1).ptdr_RL;end
                    else
                        chdata(i).TDR22(izt).ZSR=[TDR_results(izt,2).tdr];
                        chdata(i).TDR22(izt).t=TDR_results(izt,2).t;
                        chdata(i).TDR22(izt).avgZport=[TDR_results(izt,2).avgZport];
                        if OP.PTDR, chdata(i).PDTR22(izt).ptdr=TDR_results(izt,2).ptdr_RL;end
                    end
                    if OP.PTDR && i==1
                        if ipsl ==1
                            chdata(i).TDR11(izt).ERL=[TDR_results(izt,1).ERL];
                            chdata(i).TDR11(izt).ERLRMS=[TDR_results(izt,1).ERLRMS];
                        else
                            if ~param.FLAG.S2P
                                chdata(i).TDR22(izt).ERL=[TDR_results(izt,2).ERL];
                                chdata(i).TDR22(izt).ERLRMS=[TDR_results(izt,2).ERLRMS];
                            else
                                chdata(i).TDR22(izt).ERL=[];
                                chdata(i).TDR22(izt).ERLRMS=[];
                            end
                        end
                    else
                        chdata(i).TDR11(izt).ERL=[];
                        chdata(i).TDR22(izt).ERL=[];
                        chdata(i).TDR11(izt).ERLRMS=[];
                        chdata(i).TDR22(izt).ERLRMS=[];
                    end
                end
            end
        end
        
    end
    if OP.DISPLAY_WINDOW && OP.DEBUG && OP.TDR
        h=figure(180);set(gcf,'Tag','COM');
        if param.package_testcase_i==1 && i == 1
            if i==1
                htabgroup = uitabgroup(h);
                htab1 = uitab(htabgroup, 'Title', 'TDR TX');
                htab3 = uitab(htabgroup, 'Title', 'PTDR TX');
                hax1 = axes('Parent', htab1);
                hax3 = axes('Parent', htab3);
                if ~param.FLAG.S2P
                    htab2 = uitab(htabgroup, 'Title', 'TDR RX');
                    htab4 = uitab(htabgroup, 'Title', 'PTDR RX');
                    hax2 = axes('Parent', htab2);
                    hax4 = axes('Parent', htab4);
                end
            end
            set(h,'CurrentAxes',hax1)
            hold on
            plot(chdata(i).TDR11(izt).t(:),chdata(i).TDR11(izt).ZSR,'disp',[ chdata(i).base ' Tx port']);
            hold off
            legend (hax1, 'off');grid on;zoom xon;
            set(legend (hax1, 'show'), 'interp', 'none');
            
            if ~param.FLAG.S2P
                set(h,'CurrentAxes',hax2)
                hold on
                plot(chdata(i).TDR22(izt).t(:),chdata(i).TDR22(izt).ZSR,'disp',[ chdata(i).base ' Rx port']);
                hold off
                legend (hax2, 'off');grid on;zoom xon;
                set(legend (hax2, 'show'), 'interp', 'none');
            end
            
            set(h,'CurrentAxes',hax3)
            hold on
            if OP.PTDR
                for izt=1:length(param.Z_t)
                    msg=[ chdata(i).base ' Tx port Zt='  num2str(param.Z_t(izt),3) '  ERL=', num2str(chdata(i).TDR11(izt).ERL, 3) 'db'];
                    plot(chdata(i).TDR11(izt).t/param.ui,chdata(i).PDTR11(izt).ptdr,'disp',msg);
                    msg=['PTDR Zt=' num2str(param.Z_t(izt)/param.ui,3) ' worst sampled noise cursors ' 'Tx port Zt='  num2str(param.Z_t(izt),3) ];
                    stem(TDR_results(izt,1).WC_ptdr_samples_t/param.ui,TDR_results(izt,1).WC_ptdr_samples,'disp',msg);
                end
            end
            hold off
            legend (hax3, 'off');grid on;zoom xon;
            set(legend (hax3, 'show'), 'interp', 'none');
            if ~param.FLAG.S2P
                set(h,'CurrentAxes',hax4)
                hold on
                if OP.PTDR
                    for izt=1:length(param.Z_t)
                        msg=[ chdata(i).base ' Tx port Zt='  num2str(param.Z_t(izt),3) '  ERL=', num2str(chdata(i).TDR22(izt).ERL, 3) 'db'];
                        plot(chdata(i).TDR22(izt).t/param.ui,chdata(i).PDTR22(izt).ptdr,'disp',msg);
                        msg=['PTDR Zt=' num2str(param.Z_t(izt),3)/param.ui ' worst sampled noise cursors ' 'Rx port Zt='  num2str(param.Z_t(izt),3) ];
                        stem(TDR_results(izt,2).WC_ptdr_samples_t/param.ui,TDR_results(izt,2).WC_ptdr_samples,'disp',msg);
                    end
                end
                hold off
                legend (hax4, 'off');grid on;zoom xon;
                set(legend (hax4, 'show'), 'interp', 'none');
            end
        end
    end
    if param.FLAG.S2P, return; end
end
%

function S =r_parrelell2(zref,f,rpad)
S.Parameters(1,1,:) =  -zref/(rpad*(zref/rpad + 2)).*ones(1,length(f));
S.Parameters(2,2,:) =  -zref/(rpad*(zref/rpad + 2)).*ones(1,length(f));
S.Parameters(2,1,:) =   2/(zref/rpad + 2).*ones(1,length(f));
S.Parameters(1,2,:) =  2/(zref/rpad + 2).*ones(1,length(f));
Sm=sparameters(S.Parameters,f,zref);





function [chdata, param] = read_PR_files(param, OP, chdata)
%% Read in pulse response
% extract s-parameter data from files and apply tx and rx filters as well as package filters
num_files=length(chdata);
M=param.samples_per_ui;
if ~OP.DISPLAY_WINDOW, fprintf('reading file '); end
for i=1:num_files
    if OP.DISPLAY_WINDOW; hwaitbar=waitbar(0);end
    progress = i/num_files;
    if OP.DISPLAY_WINDOW
        [~,a]=fileparts(chdata(i).filename);
        waitbar(progress, hwaitbar, ['Processing ' a]); figure(hwaitbar); drawnow;
    else
        fprintf('%i ',i);
    end
    switch chdata(i).ext
        case '.csv'
            vt=load(chdata(i).filename);
            % many PR's need to have precursors added: we'll a 2 here RIM 5-24-2021
            chdata(i).uneq_pulse_response=[ zeros(3*param.samples_per_ui, 1) ; vt(:,2) ];
            dt=vt(2,1)-vt(1,1);
            chdata(i).t= [ (0:3*param.samples_per_ui-1).'*dt ; vt(:,1)+3*param.samples_per_ui*dt];
            
            
            step_shifting_vector=kron(ones(1,floor(length( chdata(i).uneq_pulse_response)/M)) ,[ 1  zeros(1,M-1) ]) ;
            step_response=filter( chdata(i).uneq_pulse_response,1, step_shifting_vector);
            Vf=step_response(end);
            chdata(i).uneq_imp_response=step_response-circshift(step_response,1); % too noisey to be usefull
            chdata(i).uneq_imp_response(1)=chdata(i).uneq_imp_response(2);
            
    end
end
function [param,OP]= read_ParamConfigFile(paramFile,OP)
%warning('off','MATLAB:xlsread:Mode'); % suppress warning messages for reading the settings file from XLS
try
    [filepath,name,ext] = fileparts(paramFile);
    if ~isempty(filepath)
        filepath=[filepath '\'];
    end
    matcongfile=[filepath name '.mat'];
    switch upper(ext)
        case upper('.mat')
            load(matcongfile)
        otherwise
            [status,sheets] = xlsfinfo(paramFile); %#ok<ASGLU>
            if isempty(strfind(cell2mat(sheets), 'COM_Settings'))
                error('COM_Settings tab not in file')
            end
            [na1, na2, parameter] = xlsread(paramFile,'COM_Settings','',''); %#ok<ASGLU> % Import data from the settings file (imports the entire sheet)
    end
    
catch ME %#ok<NASGU>
    warning('off','MATLAB:xlsread:Mode'); % suppress warning messages for reading the settings file from XLS
    [filepath,name,ext] = fileparts(paramFile);
    if ~isempty(filepath)
        filepath=[filepath '\'];
    end
    matcongfile=[filepath '\' name '.mat'];
    switch upper(ext)
        case upper('.mat')
            load(matcongfile)
        otherwise
            [status,sheets] = xlsfinfo(paramFile); %#ok<ASGLU>
            if isempty(strfind(cell2mat(sheets), 'COM_Settings'))
                error('COM_Settings tab not in file')
            end
            [na1, na2, parameter] = xlsread(paramFile,'','','basic'); %#ok<ASGLU> % Import data from the settings file (imports the entire sheet)
    end
end
% just need to define so we can pass
param.c=[.4e-12 .4e-12];
param.alen=[ 20 30 550  ];
param.az=[100 120 100];

% make control for package/channel reflection control
param.kappa1= xls_parameter(parameter, 'kappa1', true,1);
param.kappa2= xls_parameter(parameter, 'kappa2', true,1);


% make compatible with presentation of kappa

% Default values are given for parameters when they are common to all clauses in 802.3bj and 803.2bm.

param.fb = xls_parameter(parameter, 'f_b')*1e9;
param.f2 = xls_parameter(parameter, 'f_2', true, param.fb/1e9 )*1e9;
param.max_start_freq = xls_parameter(parameter, 'f_min')*1e9;
param.f1 = xls_parameter(parameter, 'f_1', true, param.max_start_freq/1e9)*1e9; % for icn and ild computations
param.max_freq_step = xls_parameter(parameter, 'Delta_f')*1e9;
param.tx_ffe_c0_min = xls_parameter(parameter, 'c(0)', false);
param.tx_ffe_cm4_values = xls_parameter(parameter, 'c(-4)', true,0); % eval if string
param.tx_ffe_cm3_values = xls_parameter(parameter, 'c(-3)', true,0); % eval if string
param.tx_ffe_cm2_values = xls_parameter(parameter, 'c(-2)', true,0); % eval if string
param.tx_ffe_cm1_values = xls_parameter(parameter, 'c(-1)', true,0); % eval if string
param.tx_ffe_cp1_values = xls_parameter(parameter, 'c(1)', true,0); % eval if string
param.ndfe = xls_parameter(parameter, 'N_b');
param.N_v = xls_parameter(parameter, 'N_v',true,param.ndfe); % add for finding steady state voltage
param.N_bx = xls_parameter(parameter, 'N_bx', true, param.ndfe); % eval if stri
% support for floating taps
param.N_bg=xls_parameter(parameter, 'N_bg', true,0); % use as a switch, 0 means no float, 3 groups allowe
param.N_bf=xls_parameter(parameter, 'N_bf', true,6); % number of taps in group
param.N_bmax=xls_parameter(parameter, 'N_bmax', true, param.ndfe); % number of UI's to apply groups
param.N_bmax=xls_parameter(parameter, 'N_f', true, param.ndfe); % number of UI's to apply groups let N_f override
if param.N_bg == 0, param.N_bmax=param.ndfe; end
param.bmaxg=xls_parameter(parameter, 'bmaxg', true,0.2); % max dfe range for group
% support for tail tap power limitations
param.B_float_RSS_MAX=xls_parameter(parameter, 'B_float_RSS_MAX', true,0);
param.N_tail_start=xls_parameter(parameter, 'N_tail_start', true,0);
%
param.dfe_delta = xls_parameter(parameter, 'N_b_step', true,0); % eval if string
param.ffe_pre_tap_len=xls_parameter(parameter, 'ffe_pre_tap_len', true,0); % eval if string
param.RxFFE_cmx=param.ffe_pre_tap_len;
param.ffe_post_tap_len=xls_parameter(parameter, 'ffe_post_tap_len', true,0); % eval if string
param.RxFFE_cpx=param.ffe_post_tap_len;
param.ffe_tap_step_size=xls_parameter(parameter, 'ffe_tap_step_size', true,0); % eval if string
param.RxFFE_stepz=param.ffe_tap_step_size;
param.ffe_main_cursor_min=xls_parameter(parameter, 'ffe_main_cursor_min', true,0.7); % eval if string
param.ffe_pre_tap1_max=xls_parameter(parameter, 'ffe_pre_tap1_max', true,.7); % eval if string
param.ffe_post_tap1_max=xls_parameter(parameter, 'ffe_post_tap1_max', true,.7); % eval if string
param.ffe_tapn_max=xls_parameter(parameter, 'ffe_tapn_max', true,.7); % eval if string
param.ffe_backoff=xls_parameter(parameter, 'ffe_backoff', true,4); % eval if string
if param.RxFFE_cmx ~= 0 || param.RxFFE_cpx ~=0
    OP.RxFFE= true;
else
    OP.RxFFE=false;
end

param.g_DC_HP_values = xls_parameter(parameter, 'g_DC_HP', true,[]); % eval if string
param.f_HP = 1e9*xls_parameter(parameter, 'f_HP_PZ', true, []); % fp1 is in GHz
param.f_HP_Z = 1e9*xls_parameter(parameter, 'f_HP_Z', true, []); % fp1 is in GHz
param.f_HP_P = 1e9*xls_parameter(parameter, 'f_HP_P', true, []); % fp1 is in GHz


param.Min_VEO= xls_parameter(parameter, 'EH_min', true,0); % for testing MTF and C2M
param.Max_VEO= xls_parameter(parameter, 'EH_max', true,inf); % for testing MTF and C2M
param.Min_VEO_Test= xls_parameter(parameter, 'EH_min_test', true,0); % for testing MTF and C2M
param.Min_VEO_Test= xls_parameter(parameter, 'Min_VEO_Test', true,param.Min_VEO_Test); % for testing MTF and C2M

param.CTLE_type= xls_parameter(parameter, 'CTLE_type', false,'CL93'); % eval if string
if ~isempty(param.g_DC_HP_values) ; param.CTLE_type='CL120d';end % overrides CL93 if g_DC_HP_values is a spreadsheet entry
if ~isempty(param.f_HP_Z) ; param.CTLE_type='CL120e';end % overrides CL93 and CL120d if f_HP_Z is a spreadsheet entry
% always read in main ctle values. They would be interpreted different baseed
% on the clause they apply because of different CTF equations
param.ctle_gdc_values = xls_parameter(parameter, 'g_DC', true); % eval if string
param.CTLE_fp1 = 1e9*xls_parameter(parameter, 'f_p1', true, param.fb/4); % fp1 is in GHz
param.CTLE_fp2 = 1e9*xls_parameter(parameter, 'f_p2', true, param.fb); % fp2 is in GHz
param.CTLE_fz = 1e9*xls_parameter(parameter, 'f_z', true, param.fb/4); % fz is in GHz
% the contex of the poles an zeros are determined by the clause
switch param.CTLE_type
    case 'CL93'
        param.ctle_gdc_values = xls_parameter(parameter, 'g_DC', true); % eval if string
        param.CTLE_fp1 = 1e9*xls_parameter(parameter, 'f_p1', true, param.fb/4); % fp1 is in GHz
        param.CTLE_fp2 = 1e9*xls_parameter(parameter, 'f_p2', true, param.fb); % fp2 is in GHz
        param.CTLE_fz = 1e9*xls_parameter(parameter, 'f_z', true, param.fb/4); % fz is in GHz
    case 'CL120d'
        param.g_DC_HP_values = xls_parameter(parameter, 'g_DC_HP', true,[]); % eval if string
        param.f_HP = 1e9*xls_parameter(parameter, 'f_HP_PZ', true, []); % fp1 is in GHz
    case 'CL120e'
        % re adjust to get TD_CTLE to work with C:120e equation without
        % changing TD_CTLE code
        param.CTLE_fz =param.CTLE_fz ./ 10.^(param.ctle_gdc_values/20);
end
param.GDC_MIN = xls_parameter(parameter, 'GDC_MIN',true, 0); % max ACDC gain, if 0 ignore
param.cursor_gain=xls_parameter(parameter, 'crusor_gain', true,0); % only FFE and not supported
param.a_thru = xls_parameter(parameter, 'A_v', true);
param.a_fext = xls_parameter(parameter, 'A_fe', true);
param.a_next = xls_parameter(parameter, 'A_ne', true);
param.a_icn_fext = xls_parameter(parameter, 'A_ft', true, param.a_fext);
param.a_icn_next = xls_parameter(parameter, 'A_nt', true, param.a_next );
param.levels = xls_parameter(parameter, 'L');
param.specBER = xls_parameter(parameter, 'DER_0');
param.pass_threshold = xls_parameter(parameter, 'COM Pass threshold',false,0);
param.ERL_pass_threshold = xls_parameter(parameter, 'ERL Pass threshold',false,0);
param.VEC_pass_threshold = xls_parameter(parameter, 'VEC Pass threshold',false,0);

param.sigma_RJ = xls_parameter(parameter, 'sigma_RJ');
param.A_DD = xls_parameter(parameter, 'A_DD');
param.eta_0 = xls_parameter(parameter, 'eta_0');

param.SNDR = xls_parameter(parameter, 'SNR_TX', true);
param.R_LM = xls_parameter(parameter, 'R_LM');

param.samples_per_ui = xls_parameter(parameter, 'M', 32);
% This will keep bmax length 0 if Nb=0

%AJG021820
param.bmax(1:param.ndfe)     = xls_parameter(parameter, 'b_max(1)');
if isempty(param.bmax)
    param.bmin=param.bmax;
else
    param.bmin(1:param.ndfe)     =xls_parameter(parameter, 'b_min(1)', true,-param.bmax(1) );
end
if param.ndfe >= 2
    param.bmax(2:param.ndfe) = xls_parameter(parameter, 'b_max(2..N_b)', true, .2);
    param.bmin(2:param.ndfe) = xls_parameter(parameter, 'b_min(2..N_b)', true, -1*param.bmax(2:param.ndfe) );
end

param.gqual=xls_parameter(parameter, 'G_Qual', true,[]);
param.g2qual=xls_parameter(parameter, 'G2_Qual', true,[]);
%verify gqual and gqual2 input
if ~isempty(param.gqual) || ~isempty(param.g2qual)
    if size(param.gqual,1)~=length(param.g2qual)
        error('gqual and g2qual size mismatch');
    end
    if size(param.gqual,2)~=2
        error('gqual must be Nx2 matrix');
    end
end


% eval if string for all three - can use different for TX and RX
param.C_pkg_board = xls_parameter(parameter, 'C_p', true)*1e-9; % C_p in nF
param.C_diepad = xls_parameter(parameter, 'C_d', true)*1e-9; % C_d in nF
% [ahealey] Read values for optional compensating L and "bump" C
param.L_comp = xls_parameter(parameter, 'L_s', true, 0)*1e-9; % L_s in nH
param.C_bump = xls_parameter(parameter, 'C_b', true, 0)*1e-9; % C_b in nF
% [ahealey] End of modifications.
param.C_v = xls_parameter(parameter, 'C_v', true,0)*1e-9; % C_v in nF (via cap)
param.R_diepad = xls_parameter(parameter, 'R_d', true);
param.Z_t = xls_parameter(parameter, 'Z_t', true,50);
param.TR_TDR = xls_parameter(parameter, 'TR_TDR', true , 8e-3);


param.Z0 = xls_parameter(parameter, 'R_0', 50);
param.z_p_tx_cases = xls_parameter(parameter, 'z_p (TX)', true).'; % eval if string
[ncases, mele]=size(param.z_p_tx_cases);
if mele ==2
    param.flex=2;
elseif mele==4
    param.flex=4;
elseif mele==1
    param.flex=1;
else
    error(springf('config file syntax error'))
end

% board parameters
param.C_0 = xls_parameter(parameter, 'C_0', true,0)*1e-9; % C_0 in nF (cap at tp0)
param.C_1 = xls_parameter(parameter, 'C_1', true,0)*1e-9; % C_1 in nF (cap at tp1)
%
param.z_p_next_cases = xls_parameter(parameter, 'z_p (NEXT)', true).'; % eval if string
[ncases1, mele1]=size(param.z_p_next_cases);
if ncases ~= ncases1 || mele ~= mele1
    error('All TX, NEXT, FEXT, Rx cases must agree');
else
end
param.z_p_fext_cases = xls_parameter(parameter, 'z_p (FEXT)', true).'; % eval if string
[ncases1, mele1]=size(param.z_p_fext_cases);
if ncases ~= ncases1 ||  mele ~= mele1
    error('All TX, NEXT, FEXT, Rx cases must agree');
else
end
param.z_p_rx_cases = xls_parameter(parameter, 'z_p (RX)', true).'; % eval if string
[ncases1, mele1]=size(param.z_p_rx_cases);
if ncases ~= ncases1 ||  mele ~= mele1
    error('All TX, NEXT, FEXT, Rx cases must agree');
else
end

% Table 93A-3 parameters
param.pkg_gamma0_a1_a2 = xls_parameter(parameter, 'package_tl_gamma0_a1_a2', true, [0 1.734e-3 1.455e-4]);
param.pkg_tau = xls_parameter(parameter, 'package_tl_tau', true, 6.141e-3);
param.pkg_Z_c = xls_parameter(parameter, 'package_Z_c', true, 78.2).';
[ ncases1, mele1]=size(param.pkg_Z_c);
if   mele ~= mele1
    error('tx rx pairs must have thesame number element entries as TX, NEXT, FEXT, Rx');
else
end
if mele1==2 % fuill in a array if only a 2 element flex package is specified
    for ii=1:ncases
        param.z_p_fext_casesx(ii,:)=  [param.z_p_fext_cases(ii,:)' ;[ 0 ; 0 ]]';
        param.z_p_next_casesx(ii,:)=  [param.z_p_next_cases(ii,:)' ;[ 0 ; 0 ]]';
        param.z_p_tx_casesx(ii,:)=  [param.z_p_tx_cases(ii,:)' ;[ 0 ; 0 ]]';
        param.z_p_rx_casesx(ii,:)=  [param.z_p_rx_cases(ii,:)' ;[ 0 ; 0 ]]';
    end
    param.z_p_fext_cases =  param.z_p_fext_casesx;
    param.z_p_next_cases=  param.z_p_next_casesx;
    param.z_p_tx_cases=  param.z_p_tx_casesx;
    param.z_p_rx_cases=  param.z_p_rx_casesx;
    param.pkg_Z_c=[param.pkg_Z_c' ;[ 100 100 ; 100 100 ]]';
end
% Table 92-12 parameters
param.brd_gamma0_a1_a2 = xls_parameter(parameter, 'board_tl_gamma0_a1_a2', true, [0 4.114e-4 2.547e-4]); % eval string, default
param.brd_tau = xls_parameter(parameter, 'board_tl_tau', false, 6.191e-3);
param.brd_Z_c = xls_parameter(parameter, 'board_Z_c', true, 109.8);
param.z_bp_tx = xls_parameter(parameter, 'z_bp (TX)', false, 151);
param.z_bp_next = xls_parameter(parameter, 'z_bp (NEXT)', false, 72);
param.z_bp_fext = xls_parameter(parameter, 'z_bp (FEXT)', false, 72);
param.z_bp_rx = xls_parameter(parameter, 'z_bp (RX)', false, 151);

% Unofficial parameters
param.snpPortsOrder = xls_parameter(parameter, 'Port Order', true, [1 3 2 4]);
param.delta_IL=xls_parameter(parameter, 'delta_IL', false, 1);
% Deprecated parameters - affect only frequency domain analysis.
param.f_v = xls_parameter(parameter, 'f_v', false, 4);
param.f_f = xls_parameter(parameter, 'f_f', false, 4);
param.f_n = xls_parameter(parameter, 'f_n', false, 4);
param.f_r = xls_parameter(parameter, 'f_r', false, 4); % used for receive filter in COM
param.fb_BT_cutoff= xls_parameter(parameter, 'TDR_f_BT_3db', false, 0.4730); %adjust to 1/2.114
param.BTorder =  xls_parameter(parameter, 'BTorder', false, 4);
param.beta_x= xls_parameter(parameter, 'beta_x', false, 0);% for pkg adj 10e9
param.rho_x= xls_parameter(parameter, 'rho_x', false, .618);
param.tfx= xls_parameter(parameter, 'fixture delay time', true, -1);
param.Grr_limit=xls_parameter(parameter, 'Grr_limit', false, 1);
param.Grr=xls_parameter(parameter, 'Grr', false, param.Grr_limit);
param.Gx=xls_parameter(parameter, 'Gx', false, 0);
switch param.Gx
    case 0
        param.Grr=param.Grr; % just use older Grr ir gx not specified
    case 1
        param.Grr=2; % use newer Grr
end

param.LOCAL_SEARCH=xls_parameter(parameter,'Local Search',true,0); % default to to for now ,if 0 search is full grid
% Operational control variables
%OP.include_pcb = xls_parameter(parameter, 'Include PCB (table 92-13)', false, 0);
param.Tukey_Window=xls_parameter(parameter,'Tukey_Window',true,0); % default to to for now ,if 0 search is full grid
param.Noise_Crest_Factor= xls_parameter(parameter, 'Noise_Crest_Factor', true, 0);
param.AC_CM_RMS = xls_parameter(parameter, 'AC_CM_RMS', true, 0);
param.ACCM_MAX_Freq=xls_parameter(parameter, 'ACCM_MAX_Freq', true, param.fb);
param.T_O = xls_parameter(parameter, 'T_O', true, 0 ); % units are mUI
param.T_O = xls_parameter(parameter, 'T_h', true, param.T_O  ); % if T_h specified use for T_O else use T_O

param.samples_for_C2M =xls_parameter(parameter, 'samples_for_C2M', true, 100 ); % need finer sample for c2m vec

OP.Histogram_Window_Weight=xls_parameter(parameter, 'Histogram_Window_Weight', false, 'rectangle' );
param.QL=xls_parameter(parameter, 'QL', true, 2.5 );


OP.include_pcb = xls_parameter(parameter, 'Include PCB', false); % wrong table reference! use no default
OP.exit_if_deployed = xls_parameter(parameter, 'exit if deployed', false,0); % wrong table reference! use no default
OP.INCLUDE_CTLE = xls_parameter(parameter, 'INCLUDE_CTLE', false, 1);
OP.EXE_MODE= xls_parameter(parameter, 'EXE_MODE', false, 1);% 12/21 0:legacy 1:fast 2:superfast
OP.INCLUDE_FILTER = xls_parameter(parameter, 'INCLUDE_TX_RX_FILTER', false, 1);
OP.force_pdf_bin_size = xls_parameter(parameter, 'Force PDF bin size', false, 0);
OP.BinSize = xls_parameter(parameter, 'PDF bin size', false, 1e-5);
OP.DEBUG = xls_parameter(parameter, 'DIAGNOSTICS', false, false);
OP.DISPLAY_WINDOW = xls_parameter(parameter, 'DISPLAY_WINDOW', false, true);
OP.CSV_REPORT = xls_parameter(parameter, 'CSV_REPORT', false, true);
OP.SAVE_TD=xls_parameter(parameter, 'SAVE_TD', false, false);
OP.SAVE_FIGURES=xls_parameter(parameter, 'SAVE_FIGURES', false, false);
OP.SAVE_FIGURE_to_CSV=xls_parameter(parameter, 'SAVE_FIGURE_to_CSV', false, false);
OP.GET_FD = xls_parameter(parameter, 'Display frequency domain', false, OP.GET_FD);
OP.INC_PACKAGE = xls_parameter(parameter, 'INC_PACKAGE', false, true);
if ~OP.INC_PACKAGE
    fprintf('<strong> Warning!!! INC_PACKAGE=0 not fully supported, instead, set Zp,Cd, and Cp parameters to zero and Zp select to 1 </strong>\n');
end

OP.EW = xls_parameter(parameter, 'EW', false, false); % RIM 3-18-2021 change defaults
OP.IDEAL_TX_TERM = xls_parameter(parameter, 'IDEAL_TX_TERM', false, false);
if OP.IDEAL_TX_TERM
    fprintf('<strong> Warning!!! IDEAL_TX_TERM not supported, instead, set Zp,Cd, and Cp parameters to zero and Zp select to 1 </strong>\n');
end
OP.IDEAL_RX_TERM = xls_parameter(parameter, 'IDEAL_RX_TERM', false, false);
if OP.IDEAL_RX_TERM
    fprintf('<strong> Warning!!! IDEAL_RX_TERM not supported, instead, set Zp,Cd, and Cp parameters to zero and Zp select to 1 </strong>\n');
end

OP.TDMODE = xls_parameter(parameter, 'TDMODE',false, OP.TDMODE);

OP.FT_COOP = xls_parameter(parameter, 'FT_COOP',false, false);
OP.RESULT_DIR = regexprep(xls_parameter(parameter, 'RESULT_DIR'), '\\', filesep);
OP.RESULT_DIR=strrep(OP.RESULT_DIR,'{date}',date);
OP.BREAD_CRUMBS = xls_parameter(parameter, 'BREAD_CRUMBS',false, false);
OP.COM_CONTRIBUTION_CURVES = xls_parameter(parameter, 'COM_CONTRIBUTION',false,0);
OP.ENFORCE_CAUSALITY = xls_parameter(parameter, 'Enforce Causality', false, 0);
OP.EC_REL_TOL = xls_parameter(parameter, 'Enforce Causality REL_TOL', false, 1e-2);
OP.EC_DIFF_TOL = xls_parameter(parameter, 'Enforce Causality DIFF_TOL', false, 1e-3);
OP.EC_PULSE_TOL = xls_parameter(parameter, 'Enforce Causality pulse start tolerance', false, 0.01);
OP.pkg_len_select = xls_parameter(parameter, 'z_p select', true, 1);  % eval if string
OP.RX_CALIBRATION = xls_parameter(parameter, 'RX_CALIBRATION', false, false);
OP.sigma_bn_STEP = xls_parameter(parameter, 'Sigma BBN step', false, 5e-3);
OP.BBN_Q_factor = xls_parameter(parameter, 'BBN Q factor', false, 5);
OP.force_BBN_Q_factor = xls_parameter(parameter, 'Force BBN Q factor', false, false);
OP.transmitter_transition_time = xls_parameter(parameter, 'T_r', true , 8e-3);
OP.RL_norm_test=xls_parameter(parameter, 'ERL_FOM', false, 1);

OP.T_r_meas_point = xls_parameter(parameter, 'T_r_meas_point', false, 0);
OP.T_r_filter_type= xls_parameter(parameter, 'T_r_filter_type', false, 0);
OP.FORCE_TR = xls_parameter(parameter, 'FORCE_TR', false, false);
% Control with OP.T_r_filter_type and OP.T_r_meas_point for backward
% compatibility
if OP.FORCE_TR
    OP.T_r_meas_point=0;
    OP.T_r_filter_type=1;
end
OP.TDR = xls_parameter(parameter, 'TDR', false, false);
OP.TDR_duration= xls_parameter(parameter, 'TDR_duration', false, 5);
OP.N = xls_parameter(parameter, 'N', false, 0);
OP.WC_PORTZ = xls_parameter(parameter, 'WC_PORTZ', false, false);
OP.T_k= xls_parameter(parameter, 'T_k', false, .6)*1e-9;
OP.ERL_ONLY = xls_parameter(parameter, 'ERL_ONLY', false,0); % compute ERL only
OP.ERL=xls_parameter(parameter, 'ERL', false, false);
if OP.ERL
    OP.PTDR=1;
else
    OP.PTDR=0;
end % ERL needs to do a TDR
OP.SHOW_BRD= xls_parameter(parameter, 'SHOW_BRD', false,0);% indclude added board in TDR and ERL

if OP.WC_PORTZ , OP.TDR=1;end % WC_PORTZ needs to do a TDR
OP.TDR_W_TXPKG = xls_parameter(parameter, 'TDR_W_TXPKG', false,0);
OP.Bessel_Thomson=xls_parameter(parameter, 'Bessel_Thomson', false, false);
OP.TDR_Butterworth=xls_parameter(parameter, 'TDR_Butterworth', false, true);
OP.Butterworth=xls_parameter(parameter, 'Butterworth', false, 1);
OP.inc_reflect_board=xls_parameter(parameter, 'inc_reflect_board', false,0);


OP.LIMIT_JITTER_CONTRIB_TO_DFE_SPAN = xls_parameter(parameter, 'LIMIT_JITTER_CONTRIB_TO_DFE_SPAN', false, false);
%OP.impulse_response_truncation_threshold = xls_parameter(parameter, 'Impulse response truncatio threshold', false, 1e-3);
OP.impulse_response_truncation_threshold = xls_parameter(parameter, 'Impulse response truncation threshold', false, 1e-3);
OP.interp_sparam_mag = xls_parameter(parameter, 'S-parameter magnitude extrapolation policy', false, 'linear_trend_to_DC');
OP.interp_sparam_phase = xls_parameter(parameter, 'S-parameter phase extrapolation policy', false, 'extrap_cubic_to_dc_linear_to_inf');
OP.PMD_type= xls_parameter(parameter, 'PMD_type', false,'C2C'); %
OP.PHY= xls_parameter(parameter, 'PHY', false, OP.PMD_type); %
if strcmpi(OP.PHY,'C2M')
    OP.EW=true;
end
if param.Min_VEO ~=0 && strcmpi(OP.PHY,'C2C')
    OP.PHY='C2Mcom';
end
OP.TDECQ=xls_parameter(parameter, 'TDECQ', false, 0); %
switch lower(OP.TDECQ)
    case {false 'none' 'vma'}
    otherwise
        error('%s not recognized for Histogram_Window_Weigh this feature is limited',OP.TDECQ)
end
OP.RUNTAG = xls_parameter(parameter, 'RUNTAG', false, '');
if isnan(OP.RUNTAG), OP.RUNTAG='';end
if isnumeric(OP.RUNTAG), OP.RUNTAG=num2str(OP.RUNTAG);end
OP.CDR=xls_parameter(parameter, 'CDR', false, 'MM');% 12/21 from Yuchun Lu to accomdate 'Mod-MM'

% Parameters for error burst probability calculation. Not officially used.
OP.use_simple_EP_model = xls_parameter(parameter, 'Use simple error propagation model', false, false);
OP.nburst = xls_parameter(parameter, 'Max burst length calculated', false, 0);
OP.COM_EP_margin = xls_parameter(parameter, 'Error propagation COM margin', false, 0);
OP.USE_ETA0_PSD = xls_parameter(parameter, 'USE_ETA0_PSD', false, 0); % used eta_0 PSD equaiton for sigma_n
OP.SAVE_CONFIG2MAT = xls_parameter(parameter, 'SAVE_CONFIG2MAT', false, 1); % used to save parameter in mat file
if OP.SAVE_CONFIG2MAT
    save(matcongfile ,'parameter');
end
OP.PLOT_CM = xls_parameter(parameter, 'PLOT_CM', false, 0); % print CM plots
% need to make sure TD mode does not invoke FD operations
if OP.TDMODE % need to set GET_FD false of TDMODE
    OP.GET_FD=false;
    param.FLAG.S2P=0;
    OP.ERL_ONLY=0;
    OP.ERL=0;
    OP.PTDR=0;
    OP.TDR=0;
    OP.RX_CALIBRATION=0;
end

%%
function [data, SDD, SDC] = read_p2_s2params(infile, plot_ini_s_params, plot_dif_s_params, ports,OP)
%% FUNCTION :: read_sp4_sparams
%
% Description
%   Read the fid of single-ended 4-port complex S-parameters
%   in Touchstone format 'file' and convert to the internal
%   format using the port transform 'ports'
%
%   Created by Mike Y. He
%   April 22, 2005
%
%   Reused some code from
%   Anthony Sanders, Alex Deas, Bob Davidov (24 January 2005)
%   for touchstone 4-port S-matrix import.
%
%   Modified (2012-July-27) by Ken Young to match current indexing scheme and
%   optimized for quicker parameter matching and parsing. also, separated out
%   the plotting algorithms into their own sub-function routines
%
% Input Variables (required)
%   file                -- The s4p file to be read and converted
%   plot_ini_sparams    -- Plot the initial s-parameter information. For debugging purposes
%   plot_diff_sparams   -- Plot the differential s-parameter information. For debugging purposes
%   swap                -- Re-order the port layout
%
% Output/Return Variables
%   sch         -- the data matrix contain the network parameter data points
%   schFreq     -- the frequency vector of the network parameter data points
%   sdc         -- the differential in/common-mode out s-parameter data matrix
%   sdd         -- the differential in/differential out s-parameter data matrix
%
% Local Variables (scoped locally to current function only)
%   fid                 -- the opened file being processed
%   fileDataLine        -- the current line in the fid
%   optionLine          -- the options line in the s4p file
%   ports               -- the order of the ports being processed
%   idxNetParams        -- the position (line) of the network parameter data points in the s4p file
%   netParamDataPoints  -- the network parameter data points of the current line of the fid
%   schFreq             -- the frequency points in the s4p file stored as a separate vector
%   sch                 -- the network parameter data points from the s4p file converted in matrix form
%   S                   --
%   T                   --
%   W                   --
%   D                   --
%   sdd                 -- the differential in/differential out form of network parameter data points
%   sdc                 -- the differential in/common-mode out form of network parameter data points
%   plot_ini_sparams    -- Plot the initial s-parameter information. For debugging purposes
%   plot_diff_sparams   -- Plot the differential s-parameter information. For debugging purposes
%

% open the fid
fid = fopen(infile, 'r');

data_file = textscan(fid, '%s', 'delimiter', sprintf('\n'));
data_file = data_file{1,1}; % removes a dimension layer of cell array depth

fclose(fid);

% backwards compatibility settings. can be removed in updated code.
if ~exist('OP', 'var'); OP.DISPLAY_WINDOW = true; end
if isempty(ports); ports = [1 2]; end % default order normally used.
ports = [1 2];


if OP.DISPLAY_WINDOW
    set(0,'defaulttextinterpreter','none') % prevents subscripting character in displayed messages
    hMsgBox = msgbox(infile, 'Reading S-Parameter File'); % display a progress bar for reading the s-parameter file(s)
end

% parse the file array until the s-parameters options line is found
idxLine = 1;
while true
    if regexp(data_file{idxLine}, '^#', 'once')
        % tests the option fileDataLine cell array to see if it contains...
        if ~isempty(regexpi(data_file{idxLine}, '\sGHz')); freq_scale = 1e9; end
        if ~isempty(regexpi(data_file{idxLine}, '\sMHz')); freq_scale = 1e6; end
        if ~isempty(regexpi(data_file{idxLine}, '\sKHz')); freq_scale = 1e3; end
        if ~isempty(regexpi(data_file{idxLine}, '\sHz'));  freq_scale = 1e0; end
        if ~isempty(regexpi(data_file{idxLine}, 'RI')) && ~isempty(regexpi(data_file{idxLine}, 'R'))
            format = 'complex';
        end
        if ~isempty(regexpi(data_file{idxLine}, 'MA')) && ~isempty(regexpi(data_file{idxLine}, 'R'))
            format = 'linpolar';
        end
        if ~isempty(regexpi(data_file{idxLine}, 'DB')) && ~isempty(regexpi(data_file{idxLine}, 'R'))
            format = 'dBpolar';
        end
        
        % if the format or freq_scale are not set then exit with error
        if isempty(freq_scale)
            error( 'read_s4p:MissingOptionsParameter', ['\n\t THE FREQUENCY SCALE PARAMETER INFORMATION IS MISSING' ...
                'FROM THE OPTIONS LINE IN THE S2P FILE']);
        elseif isempty(format)
            error( 'read_s4p:MissingOptionsParameter', ['\n\t THE FORMAT PARAMETER INFORMATION IS MISSING' ...
                'FROM THE OPTIONS LINE IN THE S2P FILE']);
        end
        
        idxLine = idxLine+1; %step one more line before exiting loop
        break
    end
    idxLine = idxLine+1;
end

col_1=1; col_2=2; % This is to maintain current indexing compatibility as well as readability
schFreqAxis = zeros(1,int8((length(data_file)-idxLine)/2)); % preallocated array length for memory and time purposes
idxNetParams = 1; % each 2-port network parameter data point set will receive an index

freqCounter = 0; % variable that will be used to capture number of freq points.

while true
    if isempty(data_file{idxLine}) || ~isempty(regexp(data_file{idxLine}, '^!', 'once'))
        idxLine = idxLine+1;
    else
        for row=1:2
            %             dataPointsLine = eval (['[' data_file{idxLine} ']']);
            dataPointsLine= sscanf(data_file{idxLine}, '%f'); % Close V2015 memory leak - Adam Gregory
            if mod(length(dataPointsLine),4)==1
                schFreqAxis(idxNetParams) = dataPointsLine(1) * freq_scale;
                dataPointsLine(1) = []; % removes the frequency value from the current data fileDataLine cell array
                % so that only the network parameter data is in the array
            end
            if mod(length(dataPointsLine),4)~=0  % sanity check to make sure the network parameter data point line structure is correct
                msg = sprintf(['\n'...
                    '\n\t ******************************************************' ...
                    '\n\t ** There is an error in the s2p network parameter   **' ...
                    '\n\t ** data point information structure at line %G.'        ...
                    '\n\t **              ENDING PROGRAM                      **' ...
                    '\n\t ******************************************************' ...
                    '\n\n\n\t Press any key to exit (an error will be displayed). \n\n']);
                disp(msg)
                pause on;
                pause;
                pause off;
                error( 'COM:reads_2p:SparamFileDataError', 'INCORRECT S4P NETWORK PARAMETER DATA POINT STRUCTURE');
            else
                map = [[ 1 2 3 4];[ 7 8 5 6]];
                switch format
                    case 'complex'
                        sch(idxNetParams, ports(row), ports(col_1)) = dataPointsLine(map(row,1)) + 1i*dataPointsLine(map(row,2)); %#ok<AGROW>
                        sch(idxNetParams, ports(row), ports(col_2)) = dataPointsLine(map(row,3)) + 1i*dataPointsLine(map(row,4)); %#ok<AGROW>
                    case 'linpolar'
                        sch(idxNetParams, ports(row), ports(col_1)) = dataPointsLine(map(row,1)) * exp(1i*dataPointsLine(map(row,2))*(pi/180)); %#ok<AGROW>
                        sch(idxNetParams, ports(row), ports(col_2)) = dataPointsLine(map(row,3)) * exp(1i*dataPointsLine(map(row,4))*(pi/180)); %#ok<AGROW>
                    case 'dBpolar'
                        sch(idxNetParams, ports(row), ports(col_1)) = 10^(dataPointsLine(map(row,1))/20) * exp(1i*dataPointsLine(map(row,2))*(pi/180)); %#ok<AGROW>
                        sch(idxNetParams, ports(row), ports(col_2)) = 10^(dataPointsLine(map(row,3))/20) * exp(1i*dataPointsLine(map(row,4))*(pi/180)); %#ok<AGROW>
                    otherwise
                end
            end
            
        end
        freqCounter = freqCounter+1; % capturing number of freq points for preallocation.
        idxLine = idxLine+1;
        if idxLine>length(data_file)
            break;
        end
        
        idxNetParams = idxNetParams + 1;
    end
    if idxLine > length(data_file)
        break
    end
end
%freqCounter = freqCounter+1; % capturing number of freq points for preallocation.
D=NaN(size(sch));
% calculate differential s parameter matrix from single ended
for i=1:size(sch,1)
    S(:,:) = sch(i,:,:);
    T = [1 1  ; 1 -1  ];
    W = T * (S / T);
    D(i,:,:) = W(:,:);
end

% D matrix should be
% Scc11 Scd11 Scc12 Scd21
% Sdc11 Sdd11 Sdc12 Sdd12
% Scc21 Scd21 Scc22 Scd22
% Sdc21 Sdd21 Sdc22 Sdd22

% proper values
SDD(:,1,1) = D(:,1,2);
SSC(:,1,1)=  D(:,2,1);



% backwards compatibility output variables
data.m = sch;
schFreqAxis=schFreqAxis(1:freqCounter); % truncating preallocated array to number of freq points.
data.freq = schFreqAxis;
colors = 'rgbk';

if (plot_ini_s_params == 1)
    figure('name', 'Single-ended s-parameters');set(gcf,'Tag','COM');
    for mj=1:4
        %         subplot(2,2,mj);
        for mi=1:4
            plot(data.freq, 20*log10(abs(data.m(:,mj,mi)+1.0e-15)), ...
                colors(mi), 'linewidth', 2, 'disp', sprintf('S%d%d', mj, mi));
            hold on
        end
        xlabel('Frequency (Hz)');
        ylabel('Magnitude (dB)');
        legend show
        grid on
        title(sprintf('Output port %d', mj));
    end
end
plot_dif_s_params =0;
if (plot_dif_s_params == 1)
    figure('name', 'Mixed-mode s-parameters');set(gcf,'Tag','COM');
    %     subplot(2,1,1);
    for mj=1:1
        for mi=1:1
            plot(data.freq, 20*log10(abs(squeeze(SDD(:,mj,mi)))), ...
                colors((mj-1)*2+mi), 'linewidth',2, 'disp', sprintf('SDD%d%d', mj, mi));
            hold on
        end
    end
    xlabel('Frequency (Hz)');
    ylabel('Magnitude (dB)');
    legend show
    grid on
    title(infile);
    
    %     subplot(2,1,2);
    %     for mj=1:2
    %         for mi=1:2
    %             plot(data.freq, 20*log10(abs(SDC(:,mj,mi))+1.0e-15), ...
    %                 colors((mj-1)*2+mi), 'linewidth',2, 'disp', sprintf('SDC%d%d', mj, mi));
    %             hold on
    %         end
    %     end
    %     xlabel('Frequency (Hz)');
    %     ylabel('Magnitude (dB)');
    %     legend show
    %     grid on
end

if OP.DISPLAY_WINDOW, close(hMsgBox); end
% end read_sp2_sparam

function [data, SDD, SDC, SCC ] = read_p4_s4params(infile, plot_ini_s_params, plot_dif_s_params, ports,OP)
%% FUNCTION :: read_sp4_sparams
%
% Description
%   Read the fid of single-ended 4-port complex S-parameters
%   in Touchstone format 'file' and convert to the internal
%   format using the port transform 'ports'
%
%   Created by Mike Y. He
%   April 22, 2005
%
%   Reused some code from
%   Anthony Sanders, Alex Deas, Bob Davidov (24 January 2005)
%   for touchstone 4-port S-matrix import.
%
%   Modified (2012-July-27) by Ken Young to match current indexing scheme and
%   optimized for quicker parameter matching and parsing. also, separated out
%   the plotting algorithms into their own sub-function routines
%
% Input Variables (required)
%   file                -- The s4p file to be read and converted
%   plot_ini_sparams    -- Plot the initial s-parameter information. For debugging purposes
%   plot_diff_sparams   -- Plot the differential s-parameter information. For debugging purposes
%   swap                -- Re-order the port layout
%
% Output/Return Variables
%   sch         -- the data matrix contain the network parameter data points
%   schFreq     -- the frequency vector of the network parameter data points
%   sdc         -- the differential in/common-mode out s-parameter data matrix
%   sdd         -- the differential in/differential out s-parameter data matrix
%
% Local Variables (scoped locally to current function only)
%   fid                 -- the opened file being processed
%   fileDataLine        -- the current line in the fid
%   optionLine          -- the options line in the s4p file
%   ports               -- the order of the ports being processed
%   idxNetParams        -- the position (line) of the network parameter data points in the s4p file
%   netParamDataPoints  -- the network parameter data points of the current line of the fid
%   schFreq             -- the frequency points in the s4p file stored as a separate vector
%   sch                 -- the network parameter data points from the s4p file converted in matrix form
%   S                   --
%   T                   --
%   W                   --
%   D                   --
%   sdd                 -- the differential in/differential out form of network parameter data points
%   sdc                 -- the differential in/common-mode out form of network parameter data points
%   plot_ini_sparams    -- Plot the initial s-parameter information. For debugging purposes
%   plot_diff_sparams   -- Plot the differential s-parameter information. For debugging purposes
%

% open the fid
fid = fopen(infile, 'r');

data_file = textscan(fid, '%s', 'delimiter', sprintf('\n'));
data_file = data_file{1,1}; % removes a dimension layer of cell array depth

fclose(fid);

% backwards compatibility settings. can be removed in updated code.
if ~exist('OP', 'var'); OP.DISPLAY_WINDOW = true; end
if isempty(ports); ports = [1 3 2 4]; end % default order normally used.

% adjust ports to maintain the meaning [in1, in2 , out1, out2] when one
% pair is reversed.
ports_adj=ports; for k=1:4, ports_adj(k)=find(ports==k); end; ports=ports_adj;

if OP.DISPLAY_WINDOW
    hMsgBox = msgbox(infile, 'Reading S-Parameter File'); % display a progress bar for reading the s-parameter file(s)
end

% parse the file array until the s-parameters options line is found
idxLine = 1;
while true
    if regexp(data_file{idxLine}, '^#', 'once')
        % tests the option fileDataLine cell array to see if it contains...
        if ~isempty(regexpi(data_file{idxLine}, '\sGHz')); freq_scale = 1e9; end
        if ~isempty(regexpi(data_file{idxLine}, '\sMHz')); freq_scale = 1e6; end
        if ~isempty(regexpi(data_file{idxLine}, '\sKHz')); freq_scale = 1e3; end
        if ~isempty(regexpi(data_file{idxLine}, '\sHz'));  freq_scale = 1e0; end
        if ~isempty(regexpi(data_file{idxLine}, 'RI')) && ~isempty(regexpi(data_file{idxLine}, 'R'))
            format = 'complex';
        end
        if ~isempty(regexpi(data_file{idxLine}, 'MA')) && ~isempty(regexpi(data_file{idxLine}, 'R'))
            format = 'linpolar';
        end
        if ~isempty(regexpi(data_file{idxLine}, 'DB')) && ~isempty(regexpi(data_file{idxLine}, 'R'))
            format = 'dBpolar';
        end
        
        % if the format or freq_scale are not set then exit with error
        if isempty(freq_scale)
            error( 'read_s4p:MissingOptionsParameter', ['\n\t THE FREQUENCY SCALE PARAMETER INFORMATION IS MISSING' ...
                'FROM THE OPTIONS LINE IN THE S4P FILE']);
        elseif isempty(format)
            error( 'read_s4p:MissingOptionsParameter', ['\n\t THE FORMAT PARAMETER INFORMATION IS MISSING' ...
                'FROM THE OPTIONS LINE IN THE S4P FILE']);
        end
        
        idxLine = idxLine+1; %step one more line before exiting loop
        break
    end
    idxLine = idxLine+1;
end

col_1=1; col_2=2; col_3=3; col_4=4; % This is to maintain current indexing compatibility as well as readability
schFreqAxis = zeros(1,int8((length(data_file)-idxLine)/4)); % preallocated array length for memory and time purposes
idxNetParams = 1; % each 4-port network parameter data point set will receive an index

freqCounter = 0; % variable that will be used to capture number of freq points.

while true
    if isempty(data_file{idxLine}) || ~isempty(regexp(data_file{idxLine}, '^!', 'once'))
        idxLine = idxLine+1;
    else
        for row=1:4
            %             dataPointsLine = eval (['[' data_file{idxLine} ']']);
            dataPointsLine= sscanf(data_file{idxLine}, '%f'); % Close V2015 memory leak - Adam Gregory
            if mod(length(dataPointsLine),8)==1
                schFreqAxis(idxNetParams) = dataPointsLine(1) * freq_scale;
                dataPointsLine(1) = []; % removes the frequency value from the current data fileDataLine cell array
                % so that only the network parameter data is in the array
            end
            if mod(length(dataPointsLine),8)~=0  % sanity check to make sure the network parameter data point line structure is correct
                msg = sprintf(['\n'...
                    '\n\t ******************************************************' ...
                    '\n\t ** There is an error in the s4p network parameter   **' ...
                    '\n\t ** data point information structure at line %G.'        ...
                    '\n\t **              ENDING PROGRAM                      **' ...
                    '\n\t ******************************************************' ...
                    '\n\n\n\t Press any key to exit (an error will be displayed). \n\n']);
                disp(msg)
                pause on;
                pause;
                pause off;
                error( 'COM:reads_4p:SparamFileDataError', 'INCORRECT S4P NETWORK PARAMETER DATA POINT STRUCTURE');
            else
                switch format
                    case 'complex'
                        sch(idxNetParams, ports(row), ports(col_1)) = dataPointsLine(1) + 1i*dataPointsLine(2); %#ok<AGROW>
                        sch(idxNetParams, ports(row), ports(col_2)) = dataPointsLine(3) + 1i*dataPointsLine(4); %#ok<AGROW>
                        sch(idxNetParams, ports(row), ports(col_3)) = dataPointsLine(5) + 1i*dataPointsLine(6); %#ok<AGROW>
                        sch(idxNetParams, ports(row), ports(col_4)) = dataPointsLine(7) + 1i*dataPointsLine(8); %#ok<AGROW>
                    case 'linpolar'
                        sch(idxNetParams, ports(row), ports(col_1)) = dataPointsLine(1) * exp(1i*dataPointsLine(2)*(pi/180)); %#ok<AGROW>
                        sch(idxNetParams, ports(row), ports(col_2)) = dataPointsLine(3) * exp(1i*dataPointsLine(4)*(pi/180)); %#ok<AGROW>
                        sch(idxNetParams, ports(row), ports(col_3)) = dataPointsLine(5) * exp(1i*dataPointsLine(6)*(pi/180)); %#ok<AGROW>
                        sch(idxNetParams, ports(row), ports(col_4)) = dataPointsLine(7) * exp(1i*dataPointsLine(8)*(pi/180)); %#ok<AGROW>
                    case 'dBpolar'
                        sch(idxNetParams, ports(row), ports(col_1)) = 10^(dataPointsLine(1)/20) * exp(1i*dataPointsLine(2)*(pi/180)); %#ok<AGROW>
                        sch(idxNetParams, ports(row), ports(col_2)) = 10^(dataPointsLine(3)/20) * exp(1i*dataPointsLine(4)*(pi/180)); %#ok<AGROW>
                        sch(idxNetParams, ports(row), ports(col_3)) = 10^(dataPointsLine(5)/20) * exp(1i*dataPointsLine(6)*(pi/180)); %#ok<AGROW>
                        sch(idxNetParams, ports(row), ports(col_4)) = 10^(dataPointsLine(7)/20) * exp(1i*dataPointsLine(8)*(pi/180)); %#ok<AGROW>
                    otherwise
                end
            end
            idxLine = idxLine+1;
            if idxLine>length(data_file)
                break;
            end
        end
        
        freqCounter = freqCounter+1; % capturing number of freq points for preallocation.
        idxNetParams = idxNetParams + 1;
    end
    if idxLine > length(data_file)
        break
    end
end

D=NaN(size(sch));
% calculate differential s parameter matrix from single ended
for i=1:size(sch,1)
    S(:,:) = sch(i,:,:);
    T = [1 1 0 0 ; 1 -1 0 0 ; 0 0 1 1 ; 0 0 1 -1];
    W = T * (S / T);
    D(i,:,:) = W(:,:);
end

% D matrix should be
% Scc11 Scd11 Scc12 Scd21
% Sdc11 Sdd11 Sdc12 Sdd12
% Scc21 Scd21 Scc22 Scd22
% Sdc21 Sdd21 Sdc22 Sdd22

% proper values
SDD(:,1,1) = D(:,2,2);
SDD(:,2,2) = D(:,4,4);
SDD(:,1,2) = D(:,2,4);
SDD(:,2,1) = D(:,4,2);

SDC(:,1,1) = D(:,2,1);
SDC(:,2,2) = D(:,4,3);
SDC(:,1,2) = D(:,2,3);
SDC(:,2,1) = D(:,4,1);

SCC(:,1,1) = D(:,1,1);
SCC(:,2,2) = D(:,3,3);
SCC(:,1,2) = D(:,1,3);
SCC(:,2,1) = D(:,3,1);

% backwards compatibility output variables
data.m = sch;
schFreqAxis=schFreqAxis(1:freqCounter); % truncating preallocated array to number of freq points.
data.freq = schFreqAxis;
colors = 'rgbk';

if (plot_ini_s_params == 1)
    figure('name', 'Single-ended s-parameters');set(gcf,'Tag','COM');
    for mj=1:4
        subplot(2,2,mj);
        for mi=1:4
            plot(data.freq, 20*log10(abs(data.m(:,mj,mi)+1.0e-15)), ...
                colors(mi), 'linewidth', 2, 'disp', sprintf('S%d%d', mj, mi));
            hold on
        end
        xlabel('Frequency (Hz)');
        ylabel('Magnitude (dB)');
        legend show
        grid on
        title(sprintf('Output port %d', mj));
    end
end
plot_dif_s_params =0;
if (plot_dif_s_params == 1)
    figure('name', 'Mixed-mode s-parameters');set(gcf,'Tag','COM');
    %     subplot(2,1,1);
    for mj=1:2
        for mi=1:2
            plot(data.freq, 20*log10(abs(SDD(:,mj,mi))), ...
                colors((mj-1)*2+mi), 'linewidth',2, 'disp', sprintf('SDD%d%d', mj, mi));
            hold on
        end
    end
    xlabel('Frequency (Hz)');
    ylabel('Magnitude (dB)');
    legend show
    grid on
    title(infile);
    %
    %     subplot(2,1,2);
    %     for mj=1:2
    %         for mi=1:2
    %             plot(data.freq, 20*log10(abs(SDC(:,mj,mi))+1.0e-15), ...
    %                 colors((mj-1)*2+mi), 'linewidth',2, 'disp', sprintf('SDC%d%d', mj, mi));
    %             hold on
    %         end
    %     end
    %     xlabel('Frequency (Hz)');
    %     ylabel('Magnitude (dB)');
    %     legend show
    %     grid on
end

if OP.DISPLAY_WINDOW, close(hMsgBox); end
% end read_sp4_sparam
function [chdata, param, SDDch,SDDp2p] = read_s4p_files(param, OP, chdata)
%% extract s-parameter and convert to differential mode
% extract s-parameter data from files and apply tx and rx filters as well as package filters
num_files=length(chdata);
if ~OP.DISPLAY_WINDOW, fprintf('reading file '); end
for i=1:num_files
    if OP.DISPLAY_WINDOW; hwaitbar=waitbar(0);end
    progress = i/num_files;
    if OP.DISPLAY_WINDOW
        [~,a]=fileparts(chdata(i).filename);
        waitbar(progress, hwaitbar, ['Processing ' a]); figure(hwaitbar); drawnow;
    else
        fprintf('%i ',i);
    end
    
    % Skip reading file if it was already read (multiple test cases)
    if (~isfield(chdata(i), 'faxis')) || isempty(chdata(i).faxis)
        param.FLAG.S2P=0;
        switch chdata(i).ext
            case '.s2p' % for differential return loss
                [Sch,SDDch] = read_p2_s2params(chdata(i).filename,  0, 0, param.snpPortsOrder, OP);
                chdata(i).fmaxi = length(Sch.freq);
                chdata(i).faxis = Sch.freq;
                chdata(i).sdd11_raw = transpose(SDDch(1:chdata(i).fmaxi,1,1));
                param.FLAG.S2P=1;
                SDDp2p(i)=NaN;
            case '.s4p'
                if ~isfield(param,'snpPortsOrder') || isempty(param.snpPortsOrder)
                    param.snpPortsOrder = [1 3 2 4]; % default order normally used.
                elseif length(param.snpPortsOrder) ~= 4
                    error( 'warning:sNpFilePortMismatch', ...
                        '\n\t The number of ports defined (%G) does not match the sNp file type (%s)', ...
                        length(param.snpPortsOrder), ...
                        chdata(i).ext ...
                        );
                end
                % read function returns differnetial mode parameters
                if param.package_testcase_i==1 % added to speed up cases e.g. don't read file in twice
                    [Sch, SDDch, SDCch] = read_p4_s4params(chdata(i).filename,  0, 0, param.snpPortsOrder, OP);
                    param.holdsdata(i).Sch= Sch;
                    param.holdsdata(i).SDDch=  SDDch;
                    param.holdsdata(i).SDCch= SDCch;
                else
                    Sch=param.holdsdata(i).Sch;
                    SDDch=param.holdsdata(i).SDDch;
                    SDCch=param.holdsdata(i).SDCch;
                end
                chdata(i).fmaxi = length(Sch.freq);
                
                if Sch.freq(chdata(i).fmaxi) < param.fb
                    warning('COM:read_s4p:MaxFreqTooLow', ...
                        'In %s: the maximum frequency provided, %g, is less than the signaling rate: %g', ...
                        chdata(i).filename, Sch.freq(end), param.fb);
                end
                if Sch.freq(1) > param.max_start_freq
                    warning('COM:read_s4p:StartFreqTooHigh', ...
                        'In %s: minimum frequency, %.2g GHz, is larger than the recommended %.2g GHz', ...
                        chdata(i).filename, Sch.freq(1)/1e9, param.max_start_freq/1e9);
                end
                freqstep=diff(Sch.freq);
                % ignore frequency differences up to 1 Hz - possible numerical artifacts
                if max(freqstep)-min(freqstep) > 1
                    warning('COM:read_s4p:NonUniformFreqSpacing', 'In %s: non-uniform frequency steps: min=%.3g GHz, max=%.3g GHz', ...
                        chdata(i).filename, min(freqstep)/1e9, max(freqstep)/1e9);
                end
                if max(freqstep) - param.max_freq_step > 1
                    warning('COM:read_s4p:FreqStepTooHigh', 'In %s: frequency step, %.2g GHz, is larger than the recommended %.2g GHz', ...
                        chdata(i).filename, max(freqstep)/1e9, param.max_freq_step/1e9);
                end
                
                chdata(i).faxis = Sch.freq;
                chdata(i).sdd12_raw = transpose(SDDch(1:chdata(i).fmaxi,1,2));
                chdata(i).sdd21_raw = transpose(SDDch(1:chdata(i).fmaxi,2,1));
                chdata(i).sdd22_raw = transpose(SDDch(1:chdata(i).fmaxi,2,2));
                chdata(i).sdd11_raw = transpose(SDDch(1:chdata(i).fmaxi,1,1));
                % mode conversion
                chdata(i).sdc12_raw = transpose(SDCch(1:chdata(i).fmaxi,1,2));
                chdata(i).sdc21_raw = transpose(SDCch(1:chdata(i).fmaxi,2,1));
                chdata(i).sdc22_raw = transpose(SDCch(1:chdata(i).fmaxi,2,2));
                chdata(i).sdc11_raw = transpose(SDCch(1:chdata(i).fmaxi,1,1));
            otherwise
        end
        %
        % differential response is sdd21 <------------------------------
        % differential return losses are sdd11 and sdd22
        % save original
        switch chdata(i).ext
            case '.s2p' % for differential return loss
                chdata(i).sdd11_orig=chdata(i).sdd11_raw;
                chdata(i).sdd11=chdata(i).sdd11_raw;
            otherwise
                chdata(i).sdd11_orig=chdata(i).sdd11_raw;
                chdata(i).sdd22_orig=chdata(i).sdd22_raw;
                chdata(i).sdd12_orig=chdata(i).sdd12_raw;
                chdata(i).sdd21_orig=chdata(i).sdd21_raw;
                if OP.include_pcb
                    % add boards to sdd
                    [chdata(i).sdd11_raw, chdata(i).sdd12_raw, chdata(i).sdd21_raw, chdata(i).sdd22_raw] = add_brd(chdata(i), param, OP);
                    
                end
                chdata(i).sdd11=chdata(i).sdd11_raw;
                chdata(i).sdd22=chdata(i).sdd22_raw;
        end
    else
        SDDch(:,1,2)=chdata(i).sdd12_raw;
        SDDch(:,2,1)=chdata(i).sdd21_raw;
        SDDch(:,1,1)=chdata(i).sdd11_raw;
        SDDch(:,2,2)=chdata(i).sdd22_raw;
    end
    if ~param.FLAG.S2P
        if  OP.INC_PACKAGE ~= 0 || (OP.RX_CALIBRATION == 1  && i==1)
            if (OP.RX_CALIBRATION == 1  && i==2)
                chdata(i).sdd21=chdata(i).sdd21_raw;
            else
                if param.kappa1 == 1 && param.kappa2 == 1
                    [chdata(i).sdd21p,param,SDDp2p(i)]= s21_pkg(chdata(i).sdd21_raw, chdata(i).sdd11_raw, chdata(i).sdd22_raw, chdata(i).faxis, param, OP, chdata(i).type, i, chdata(i));
                    chdata(i).sdd21=chdata(i).sdd21p;
                    if 1 % for AC CM noise inclusion
                        [chdata(i).sdc21p,param,SDCp2p(i)]= sdc21_pkg(chdata(i).sdc21_raw, chdata(i).sdc11_raw, chdata(i).sdc22_raw, chdata(i).faxis, param, OP, chdata(i).type, i, chdata(i));
                        chdata(i).sdc21=chdata(i).sdc21p;
                    end
                else % remove return loss at port to determine reflectionless isi
                    chadata_RL_wo_TXRL=chdata(i).sdd11_raw;
                    chadata_RL_wo_RXRL=chdata(i).sdd22_raw;
                    if param.kappa1 == 0
                        chadata_RL_wo_TXRL= zeros(1, length(chdata(i).sdd22_raw));
                    else
                        chadata_RL_wo_TXRL=chdata(i).sdd11_raw*param.kappa1;
                    end
                    if param.kappa2 == 0
                        chadata_RL_wo_RXRL= zeros(1, length(chdata(i).sdd22_raw));
                    else
                        chadata_RL_wo_RXRL=chdata(i).sdd22_raw*param.kappa2;
                    end
                    [chdata(i).sdd21p,param,SDDp2p(i)]= s21_pkg(chdata(i).sdd21_raw, chadata_RL_wo_TXRL, chadata_RL_wo_RXRL, chdata(i).faxis, param, OP, chdata(i).type, i, chdata(i));
                    chdata(i).sdd21=chdata(i).sdd21p;
                    if 1 % for AC CM noise inclusion
                        [chdata(i).sdc21p,param,SDCp2p(i)]= sdc21_pkg(chdata(i).sdc21_raw, chdata(i).sdc11_raw, chdata(i).sdc22_raw, chdata(i).faxis, param, OP, chdata(i).type, i, chdata(i));
                        chdata(i).sdc21=chdata(i).sdc21p;
                    end
                end
            end
        else
            chdata(i).sdd21=chdata(i).sdd21_raw;
        end
        chdata(i).sdd21f=chdata(i).sdd21_raw; % used for FD analysis i.e. not filtered
    end
end
if ~OP.DISPLAY_WINDOW, fprintf('\n'); end

function result = readdataSnPx(filename, nport)
%function [freq, cs] = readdataSnPx(filename, nport)
% [freq, cs] = readdataSnP(filename, nport, format, nheader)
%
% Read Touchstone file with frequencies in units of Hertz
%
% Input:
% ======
% filename: Name of the Touchstone/SnP file
% nport: Number of ports
% format: 'RI' for real/imag, 'MA' for mag/angle (check option line in the
%         Touchstone file)
% nheader: Number of header lines (comment lines plus option line in the
%          Touchstone file)
%
% Output:
% =======
% freq: Vector of frequencies [Hz]
% cs: 3-D array of complex-valued S parameters where cs(i,j,k) is S(i,j)
%     at frequency freq(k)
%
% Note: If frequency unit is not Hertz (but GHz, MHz etc.) simply scale
% frequencies appropriately after reading the data.
%
% Ref.: Touchstone(R) File Format Specification, Rev.1.1,
% EIA/IBIS Open Forum, 2002.
%
% Written by Henning Braunisch, September 2004.
% Updated by Steven Krooswyk, April 2006.


fid = fopen(filename, 'r');


% Skip header lines
str = ' ';
n = 0;
while ~strcmp(str(1),'#')
    str = fgetl(fid);
    if isempty(str)
        str=' ' ;
        if n > 1000
            display('error: could not find config line (#)')
            break
        end
    end
    n = n + 1;
end

% parse configuration line
A=sscanf(str,'%1s %2s %1s %2s %1s %2s',[1,inf]);
p = find(A=='S');           %position of 'S'
units = lower(A(2:p-1));    %units before 'S'
format = A(p+1:p+2);        %format after 'S'

% skip any more header lines
%while ~str

nk = 0; % frequency counter
while 1
    
    [temp, count] = fscanf(fid, '%f', 1);
    if count == 0
        temp2 = fscanf(fid, '%s', 1);
        if ~isempty(temp2), fgetl(fid); continue, end;
        break
    end
    nk = nk+1; freq(1,nk) = temp;  %#ok<AGROW>
    for ni = 1:nport
        for nj = 1:nport
            switch lower(format)
                case 'ma'
                    mag = fscanf(fid, '%f', 1); ang = fscanf(fid, '%f', 1);
                    cs(ni,nj,nk) = mag * exp(1i*ang*pi/180); %#ok<AGROW>
                case 'ri'
                    re = fscanf(fid, '%f', 1); im = fscanf(fid, '%f', 1);
                    cs(ni,nj,nk) = complex(re, im); %#ok<AGROW>
                case 'db'
                    db = fscanf(fid, '%f', 1); ang = fscanf(fid, '%f', 1);
                    M = 10^(db/20);
                    %re = M*cos(ang);
                    %im = M*sin(ang);
                    re =  M*cos(ang * pi / 180);
                    im =  M*sin(ang * pi / 180);
                    cs(ni,nj,nk) = complex(re, im); %#ok<AGROW>
                otherwise
                    error('readdataSnP: Unknown data format');
            end
        end
    end
end

fclose(fid);

% If 2-port then swap S_12 and S_21 per Touchstone spec
if nport == 2
    temp = cs(2,1,:);
    cs(2,1,:) = cs(1,2,:);
    cs(1,2,:) = temp;
end

% Update freq units to Hz
switch lower(units)
    case 'hz'
        
    case 'khz'
        freq=freq.*1e3;
    case 'mhz'
        freq=freq.*1e6;
    case 'ghz'
        freq=freq.*1e9;
end

% passivity check
result.freq = freq;
result.cs    = cs;

function recolor_plots(ax)

if ~verLessThan('matlab', '8.4.0')
    return
end
colors='brgcmk';
ch=flipud(get(ax, 'children'));

for k=1:length(ch)
    set(ch(k), 'Color', colors(mod(k-1, length(colors))+1));
    set(ch(k), 'LineWidth', 2*floor((k-1)/length(colors))+1);
end
legend (ax, 'off');
warning('off', 'MATLAB:legend:PlotEmpty');
set(legend (ax, 'show'), 'interp', 'none');

function result = reduce(var1)
% --- Reduce 1x1xn array to 1xn (aka squeeze)
out = zeros(1,length(var1));
out(1,:) = var1(1,1,:);
result=out;

function [sdd21p,param,SCH] = s21_pkg(sdd21, sdd11, sdd22, faxis, param, OP, channel_type, channel_number,chdatalocal)
% concatenates package reflections with s21,s11,and s22 with spec return loss (gammas)
% faxis is the frequency array
% sdd21, sdd11, sdd22 are the corresponding array of differential parameters
sdd12=sdd21;
Z0=param.Z0;

% The following three parameters have possibly different valuesF for TX and
% RX (so can be 2-element vectors).
R_diepad = param.R_diepad;
C_diepad = param.C_diepad;
C_pkg_board = param.C_pkg_board;
% [ahealey] Unpack optional compensating L and "bump" C model parameters.
L_comp = [0, 0];
C_bump = [0, 0];
if isfield(param, 'L_comp')
    L_comp = param.L_comp;
end
if isfield(param, 'C_bump')
    C_bump = param.C_bump;
end
if isscalar(L_comp), L_comp = L_comp*[1, 1]; end
if isscalar(C_bump), C_bump = C_bump*[1, 1]; end
% [ahealey] End of modifications.
param.Tx_rd_sel=1;
param.Rx_rd_sel=2;
% generate TX package according to channel type.
[ncases, mele]=size(param.z_p_next_cases);
switch upper(channel_type)
    case 'THRU'
        switch mele
            case 1
                Zc1(1)=param.pkg_Z_c(1);% tx impedance
                Zc1(2)=param.pkg_Z_c(2);% rx
                Zp1(1)=param.Pkg_len_TX;% tx mm (length)
                Zp1(2)=param.Pkg_len_RX;% rx
            case 4
                Zc1=[ param.pkg_Z_c(1,1) param.pkg_Z_c(2,1)] ;% Tx Rx impedance
                Zp1=[ param.Pkg_len_TX(1) param.Pkg_len_RX(1)] ;% Tx Rx length
                Zc2 =[ param.pkg_Z_c(1,2) param.pkg_Z_c(2,2)] ;
                Zp2 = [ param.Pkg_len_TX(2) param.Pkg_len_RX(2) ] ;
                Zc3 =[ param.pkg_Z_c(1,3) param.pkg_Z_c(2,3)] ;
                Zp3 = [ param.Pkg_len_TX(3) param.Pkg_len_RX(3) ] ;
                Zc4 = [ param.pkg_Z_c(1,4) param.pkg_Z_c(2,4)] ;
                Zp4 = [ param.Pkg_len_TX(4) param.Pkg_len_RX(4) ] ;
                Cv = [ param.C_v(1) param.C_v(2)  ];
            otherwise
                error('package symtax error')
        end
    case 'NEXT'
        switch mele
            case 1
                Zc1(1)=param.pkg_Z_c(1);% tx impedance
                Zc1(2)=param.pkg_Z_c(2);% rx
                Zp1(1)=param.Pkg_len_NEXT;% tx mm (length)
                Zp1(2)=param.Pkg_len_RX;% rx
            case 4
                Zc1=[ param.pkg_Z_c(1,1) param.pkg_Z_c(2,1)] ;% Tx Rx impedance
                Zp1=[ param.Pkg_len_NEXT(1) param.Pkg_len_RX(1)] ;% Tx Rx length
                Zc2 =[ param.pkg_Z_c(1,2) param.pkg_Z_c(2,2)] ;
                Zp2 = [ param.Pkg_len_NEXT(2) param.Pkg_len_RX(2) ] ;
                Zc3 =[ param.pkg_Z_c(1,3) param.pkg_Z_c(2,3)] ;
                Zp3 = [ param.Pkg_len_NEXT(3) param.Pkg_len_RX(3) ] ;
                Zc4 = [ param.pkg_Z_c(1,4) param.pkg_Z_c(2,4)] ;
                Zp4 = [ param.Pkg_len_NEXT(4) param.Pkg_len_RX(4) ] ;
                Cv = [ param.C_v(1) param.C_v(2)  ];
            otherwise
                error('package symtax error')
        end
    case 'FEXT'
        switch mele
            case 1
                Zc1(1)=param.pkg_Z_c(1);% tx impedance
                Zc1(2)=param.pkg_Z_c(2);% rx
                Zp1(1)=param.Pkg_len_FEXT;% tx mm (length)
                Zp1(2)=param.Pkg_len_RX;% rx
            case 4
                Zc1=[ param.pkg_Z_c(1,1) param.pkg_Z_c(2,1)] ;% Tx Rx impedance
                Zp1=[ param.Pkg_len_FEXT(1) param.Pkg_len_RX(1)] ;% Tx Rx length
                Zc2 =[ param.pkg_Z_c(1,2) param.pkg_Z_c(2,2)] ;
                Zp2 = [ param.Pkg_len_FEXT(2) param.Pkg_len_RX(2) ] ;
                Zc3 =[ param.pkg_Z_c(1,3) param.pkg_Z_c(2,3)] ;
                Zp3 = [ param.Pkg_len_FEXT(3) param.Pkg_len_RX(3) ] ;
                Zc4 = [ param.pkg_Z_c(1,4) param.pkg_Z_c(2,4)] ;
                Zp4 = [ param.Pkg_len_FEXT(4) param.Pkg_len_RX(4) ] ;
                Cv = [ param.C_v(1) param.C_v(2)  ];
            otherwise
                error('package symtax error')
        end
end
% tx package
switch mele
    case 1
        % [ahealey] Add series L shunt C to the on-die termination model.
        % [ s11out, s12out, s21out, s22out ]= make_pkg(faxis, Zp1(1), C_diepad(1), C_pkg_board(1),Zc1(1), param);
        [ s11out, s12out, s21out, s22out ]= make_pkg(faxis, Zp1(1), C_diepad(1), C_pkg_board(1),Zc1(1), param, L_comp(1), C_bump(1));
    case 4
        % [ s11outa, s12outa, s21outa, s22outa ]= make_pkg(faxis, Zp1(1),  C_diepad(1),0 ,Zc1(1), param);
        [ s11outa, s12outa, s21outa, s22outa ]= make_pkg(faxis, Zp1(1),  C_diepad(1),0 ,Zc1(1), param, L_comp(1), C_bump(1));
        % [ahealey] End of modifications.
        %
        [ s11outb, s12outb, s21outb, s22outb ]= make_pkg(faxis, Zp2(1) , 0, 0 ,Zc2(1), param); % add pth
        [ s11outc, s12outc, s21outc, s22outc ]  = combines4p(  s11outa, s12outa, s21outa, s22outa, s11outb, s12outb, s21outb, s22outb    );
        %
        %
        [ s11outd, s12outd, s21outd, s22outd ]= make_pkg(faxis, Zp3(1) , 0, Cv(1) ,Zc3(1), param); % add pth
        [ s11oute, s12oute, s21oute, s22oute ]  = combines4p(  s11outc, s12outc, s21outc, s22outc, s11outd, s12outd, s21outd, s22outd    );
        %
        [ s11outf, s12outf, s21outf, s22outf ]= make_pkg(faxis, Zp4(1) , 0, C_pkg_board(1), Zc4(1), param); % add pth
        [ s11out, s12out, s21out, s22out ]  = combines4p(  s11oute, s12oute, s21oute, s22oute, s11outf, s12outf, s21outf, s22outf    );
end

% RX package
switch mele
    case 1
        % [ahealey] Add series L shunt C to the on-die termination model.
        % [ s11in, s12in, s21in, s22in ]= make_pkg(faxis,  Zp1(2), C_diepad(2), C_pkg_board(2), Zc1(2), param);
        [ s11in, s12in, s21in, s22in ]= make_pkg(faxis,  Zp1(2), C_diepad(2), C_pkg_board(2), Zc1(2), param, L_comp(2), C_bump(2));
    case 4
        % [ s11ina, s12ina, s21ina, s22ina ]= make_pkg(faxis, Zp1(2), C_diepad(2),0 ,Zc1(2), param));
        [ s11ina, s12ina, s21ina, s22ina ]= make_pkg(faxis, Zp1(2), C_diepad(2),0 ,Zc1(2), param, L_comp(2), C_bump(2));
        % [ahealey] End of modifications.
        %
        [ s11inb, s12inb, s21inb, s22inb ]= make_pkg(faxis, Zp2(2) , 0, 0 ,Zc2(2), param); % add pth
        [ s11inc, s12inc, s21inc, s22inc ]  = combines4p(  s11ina, s12ina, s21ina, s22ina, s11inb, s12inb, s21inb, s22inb    );
        %
        [ s11ind, s12ind, s21ind, s22ind ]= make_pkg(faxis, Zp3(2) , 0, Cv(2) ,Zc3(2), param); % add pth
        [ s11ine, s12ine, s21ine, s22ine ]  = combines4p(  s11inc, s12inc, s21inc, s22inc, s11ind, s12ind, s21ind, s22ind    );
        %
        [ s11inf, s12inf, s21inf, s22inf ]= make_pkg(faxis, Zp4(2) , 0, C_pkg_board(2), Zc4(2), param); % add pth
        [ s11in, s21in, s12in, s22in ]  = combines4p(  s11ine, s12ine, s21ine, s22ine, s11inf, s12inf, s21inf, s22inf    );
end


%     p(1 ,1, :)=s11in;
%     p(2 ,2, :)=s22in;
%     p(1 ,2, :)=s12in;
%     p(2 ,1, :)=s21in;
%
%     S=sparameters(p,faxis);
%     rfwrite(S,'temp.s4p');

if OP.IDEAL_TX_TERM || (OP.RX_CALIBRATION == 1 && channel_number == 2) || OP.include_pcb == 2
    gamma_tx=0;
else
    gamma_tx=(R_diepad(param.Tx_rd_sel)-Z0)/(R_diepad(param.Tx_rd_sel)+Z0);% equation 93A-17
end
if OP.IDEAL_RX_TERM
    gamma_rx=0;
else
    gamma_rx=(R_diepad(param.Rx_rd_sel)-Z0)/(R_diepad(param.Rx_rd_sel)+Z0);% equation 93A-17
end

if OP.INC_PACKAGE==0
    sdd21p= sdd21;
    warning('do not use INC_PACKAGE = 0. Instead use package parameters)');
else
    if OP.RX_CALIBRATION == 1 && channel_number == 2
        %   for calibration do not include the transmitter package
        [s11out_rx, s12out_rx, s21out_rx, s22out_rx ] = combines4p( sdd11, sdd21, sdd21, sdd22, s22out, s12out, s21out, s11out ); %#ok<ASGLU> % s22 is ball side of package
        SCH.Frequencies=faxis;
        SCH.Parameters(1,1,:)=s11out_rx;
        SCH.Parameters(2,2,:)=s22out_rx;
        SCH.Parameters(1,2,:)=s12out_rx;
        SCH.Parameters(2,1,:)=s21out_rx;
        SCH.NumPorts=2;
        SCH.Impedance=100;
        %% Equation 93A-18
        sdd21p= s21out_rx.*(1-gamma_tx).*(1+gamma_rx)./(1.- s11out_rx.*gamma_tx - s22out_rx.*gamma_rx  -s21out_rx.^2.*gamma_tx.*gamma_rx +s11out_rx.*s22out_rx.*gamma_tx.*gamma_rx);
    else
        %% Equations 93A-4 to 93A-7
        if ~OP.IDEAL_TX_TERM
            [sdd11, sdd12, sdd21, sdd22] = combines4p( s11out, s12out, s21out, s22out, sdd11, sdd12, sdd21, sdd22 ); %#ok<ASGLU>
        end
        H_t=ones(1,length(faxis)); % .3bj compatibility
        if OP.IDEAL_TX_TERM ||  OP.T_r_filter_type == 1
            % for RITT testing with good termination as in some instruments
            % and tx filter when required
            if OP.T_r_filter_type==0
                H_t = exp(-(pi*faxis/1e9*OP.transmitter_transition_time/1.6832).^2); %% Equation 93A-46 %%
            else
                tr=OP.transmitter_transition_time;
                f9=faxis/1e9;
                if OP.T_r_meas_point == 1
                    k=1.9466+7.12*sqrt(1-6.51e-3/tr);
                    H_t=105./(f9.^4*(k*tr)^4 - f9.^3*(k*tr)^3*10i - 45*f9.^2*(k*tr)^2 + f9*(k*tr)*105i + 105);
                else
                    H_t = exp( -2*(pi*f9*tr/1.6832).^2 ).*exp(-1j*2*pi*f9*tr*3);
                end
                
            end
        end
        if ~OP.IDEAL_RX_TERM
            [sdd11, sdd12, sdd21, sdd22] = combines4p( sdd11, sdd12, sdd21, sdd22, s22in, s12in, s21in, s11in ); %#ok<ASGLU> % s22 is ball side of package
        else
            warning('do not use IDEAL_RX_TERM. Instead hard code package and TR parameters')
        end
        %% Equation 93A-18 and part of 93A-1: Ht fix in V290 identified by Bill Kirkland and Ed Frlan
        sdd21p= H_t.*sdd21.*(1-gamma_tx).*(1+gamma_rx)./(1.- sdd11.*gamma_tx - sdd22.*gamma_rx  -sdd21.^2.*gamma_tx.*gamma_rx +sdd11.*sdd22.*gamma_tx.*gamma_rx);
    end
    %     % for debug
    %     p(1 ,1, :)=sdd11;
    %     p(2 ,2, :)=sdd22;
    %     p(1 ,2, :)=sdd12;
    %     p(2 ,1, :)=sdd21;
    %
    %     S=sparameters(p,faxis);
    %     rfwrite(S,'temp.s4p');
    
    SCH.Frequencies=faxis;
    SCH.Parameters(1,1,:)=sdd11;
    SCH.Parameters(2,2,:)=sdd22;
    SCH.Parameters(1,2,:)=sdd12;
    SCH.Parameters(2,1,:)=sdd21;
    SCH.NumPorts=2;
    SCH.Impedance=100;
    
end


function [voltage, t_base, causality_correction_dB, truncation_dB] = ...
    s21_to_impulse_DC(IL, freq_array, time_step, OP)
% Creates a time-domain impulse response from frequency-domain IL data.
% IL does not need to have DC but a corresponding frequency array
% (freq_array) is required.
%
% Causality is imposed using the Alternating Projections Method. See also:
% Quatieri and Oppenheim, "Iterative Techniques for Minimum Phase Signal
% Reconstruction from Phase or Magnitude", IEEE Trans. ASSP-29, December
% 1981 (http://ieeexplore.ieee.org/xpls/abs_all.jsp?arnumber=1163714)

ILin=IL;
fmax=1/time_step/2;
freq_step=(freq_array(3)-freq_array(2))/1;
fout=0:1/round(fmax/freq_step)*fmax:fmax;
IL=interp_Sparam(ILin,freq_array,fout, ...
    OP.interp_sparam_mag, OP.interp_sparam_phase,OP);
IL_nan = find(isnan(IL));
for in=IL_nan
    IL(in)=IL(in-1);
end
IL = IL(:);
% add padding for time steps
IL_symmetric = [IL(1:end-1);0; flipud(conj(IL(2:end-1)))];
impulse_response = real(ifft(IL_symmetric));
L = length(impulse_response);
t_base = (0:L-1)/(freq_step*L);

original_impulse_response=impulse_response;
% Correct non-causal effects frequently caused by extrapolation of IL
% Assumption: peak of impulse_response is in the first half, i.e. not anti-causal
abs_ir=abs(impulse_response);
a = find(abs_ir(1:L/2) > max(abs_ir(1:L/2))*OP.EC_PULSE_TOL);
start_ind = a(1);

err=inf;
while ~all(impulse_response==0)
    impulse_response(1:start_ind)=0;
    impulse_response(floor(L/2):end)=0;
    IL_modified=abs(IL_symmetric).*exp(1j*angle(fft(impulse_response)));
    ir_modified = real(ifft(IL_modified));
    delta = abs(impulse_response-ir_modified);
    
    err_prev = err;
    err=max(delta)/max(impulse_response);
    if err<OP.EC_REL_TOL || abs(err_prev-err)<OP.EC_DIFF_TOL
        break;
    end
    
    impulse_response=ir_modified;
end

causality_correction_dB=20*log10(norm(impulse_response-original_impulse_response)/norm(impulse_response));

if ~OP.ENFORCE_CAUSALITY
    impulse_response = original_impulse_response;
end
% truncate final samples smaller than 1e-3 of the peak
ir_peak = max(abs(impulse_response));
ir_last  = find(abs(impulse_response)>ir_peak*OP.impulse_response_truncation_threshold, 1, 'last');

voltage = impulse_response(1:ir_last);
t_base = t_base(1:ir_last);

truncation_dB=20*log10(norm(impulse_response(ir_last+1:end))/norm(voltage));


function S =s_for_c2(zref,f,cpad)
% S is 2 port s parameters out
S_Parameters(1,1,:) =  -1i*2*pi.*f*cpad*zref./(2+1i*2*pi.*f*cpad*zref);
S_Parameters(2,2,:) =  -1i*2*pi.*f*cpad*zref./(2+1i*2*pi.*f*cpad*zref);
S_Parameters(2,1,:) = 2./(2+1i*2*pi.*f*cpad*zref);
S_Parameters(1,2,:) = 2./(2+1i*2*pi.*f*cpad*zref);
S=sparameters(S_Parameters,f,zref);

function S =s_for_c4(zref,f,cpad)

S2 = s_for_c2(zref,f,cpad);
S4P=s2_to_s4(S2.Parameters);
S=sparameters(S4P,f,zref);
S.Parameters=snp2smp(S.Parameters,zref,[ 1 3 2 4]);




%%
function [ cmd_str ] = save_cmd_line( config_file, chdata, num_fext,num_next, cli_name )
% save commmend string
% for saving from interactive queries


cmd_str=[ cli_name '('  '''' config_file '''' ',' num2str(num_fext) ', ' num2str(num_next) ',' '''' chdata(1).filename ''''];
for i=1:num_next+num_fext
    cmd_str= [cmd_str  ',' '''' chdata(i+1).filename ''''];
end
cmd_str= [ cmd_str ')'];


%%%%% require the RF tool box
%%
function [ h ] = savefigs( param, OP )

%% find the figures
hw = waitbar(0,'Saving figures...');
h = findobj(0, 'Type', 'figure');
for ii=1:length(h)
    
    figname= get(h(ii), 'Name'); % use the figure name as file name
    if isempty(strfind(figname,param.base))
        figname = [figname ' ' OP.RUNTAG ' ' param.base ]; %#ok<AGROW>
    end
    if verLessThan('matlab', '8.4.0')
        figname = ['f_' num2str(h(ii)) '_' figname]; %#ok<AGROW>
    else
        figname = ['f_' num2str(h(ii).Number) '_' figname]; %#ok<AGROW>
    end
    figname = strrep(figname,':','-');
    figname = strrep(figname,' ','_');
    if OP.SAVE_FIGURES==1
        saveas(h(ii), fullfile(OP.RESULT_DIR, [figname '.fig']));
    end
    %% get x y data
    if OP.SAVE_FIGURE_to_CSV==1
        h_L = findobj(h(ii),'Type','line'); % find handles to all the lines
        M=[]; %ncol=1;
        for nk=1:length(h_L)
            % get x and data for a line.
            x_data=get(h_L(nk),'xdata')';
            y_data=get(h_L(nk),'ydata')';
            % .........>> need to get data in the line structure (legend or label) for headers
            M=[M; x_data; y_data]; %#ok<AGROW>
        end
        csvwrite([OP.RESULT_DIR figname '.csv'],M);
        %      clear M y x header h_L
    end
    waitbar(ii/length(h),hw)
    
end

close(hw)

%%
function [ h ] = savefigs_png( param, OP )

%% find the figures
hw = waitbar(0,'Saving figures...');
h = findobj(0, 'Type', 'figure');
for ii=1:length(h)
    
    figname= get(h(ii), 'Name'); % use the figure name as file name
    if isempty(strfind(figname,param.base))
        figname = [figname ' ' OP.RUNTAG ' ' param.base ]; %#ok<AGROW>
    end
    if verLessThan('matlab', '8.4.0')
        figname = ['f_' num2str(h(ii)) '_' figname]; %#ok<AGROW>
    else
        figname = ['f_' num2str(h(ii).Number) '_' figname]; %#ok<AGROW>
    end
    figname = strrep(figname,':','-');
    figname = strrep(figname,' ','_');
    if OP.SAVE_FIGURES==1
        saveas(h(ii), fullfile(OP.RESULT_DIR, [figname '.png']));
    end
    %% get x y data
    if OP.SAVE_FIGURE_to_CSV==1
        h_L = findobj(h(ii),'Type','line'); % find handles to all the lines
        M=[]; %ncol=1;
        for nk=1:length(h_L)
            % get x and data for a line.
            x_data=get(h_L(nk),'xdata')';
            y_data=get(h_L(nk),'ydata')';
            % .........>> need to get data in the line structure (legend or label) for headers
            M=[M; x_data; y_data]; %#ok<AGROW>
        end
        csvwrite([OP.RESULT_DIR figname '.csv'],M);
        %      clear M y x header h_L
    end
    waitbar(ii/length(h),hw)
    
end

close(hw)

%%
function [sdc21p,param,SCHDC] = sdc21_pkg(sdc21, sdc11, sdc22, faxis, param, OP, channel_type, channel_number,chdatalocal)
% concatenates package reflections with s21,s11,and s22 with spec return loss (gammas)
% faxis is the frequency array
% sdc21, sdc11, sdc22 are the corresponding array of differential parameters
% sdc121=ones(length(sdc21),1);
sdc12=sdc21;
Z0=param.Z0;

% The following three parameters have possibly different valuesF for TX and
% RX (so can be 2-element vectors).
R_diepad = param.R_diepad;
C_diepad = param.C_diepad;
C_pkg_board = param.C_pkg_board;
% [ahealey] Unpack optional compensating L and "bump" C model parameters.
L_comp = [0, 0];
C_bump = [0, 0];
if isfield(param, 'L_comp')
    L_comp = param.L_comp;
end
if isfield(param, 'C_bump')
    C_bump = param.C_bump;
end
if isscalar(L_comp), L_comp = L_comp*[1, 1]; end
if isscalar(C_bump), C_bump = C_bump*[1, 1]; end
% [ahealey] End of modifications.
param.Tx_rd_sel=1;
param.Rx_rd_sel=2;
% generate TX package according to channel type.
[ncases, mele]=size(param.z_p_next_cases);
switch upper(channel_type)
    case 'THRU'
        switch mele
            case 1
                Zc1(1)=param.pkg_Z_c(1);% tx impedance
                Zc1(2)=param.pkg_Z_c(2);% rx
                Zp1(1)=param.Pkg_len_TX;% tx mm (length)
                Zp1(2)=param.Pkg_len_RX;% rx
            case 4
                Zc1=[ param.pkg_Z_c(1,1) param.pkg_Z_c(2,1)] ;% Tx Rx impedance
                Zp1=[ param.Pkg_len_TX(1) param.Pkg_len_RX(1)] ;% Tx Rx length
                Zc2 =[ param.pkg_Z_c(1,2) param.pkg_Z_c(2,2)] ;
                Zp2 = [ param.Pkg_len_TX(2) param.Pkg_len_RX(2) ] ;
                Zc3 =[ param.pkg_Z_c(1,3) param.pkg_Z_c(2,3)] ;
                Zp3 = [ param.Pkg_len_TX(3) param.Pkg_len_RX(3) ] ;
                Zc4 = [ param.pkg_Z_c(1,4) param.pkg_Z_c(2,4)] ;
                Zp4 = [ param.Pkg_len_TX(4) param.Pkg_len_RX(4) ] ;
                Cv = [ param.C_v(1) param.C_v(2)  ];
            otherwise
                error('package symtax error')
        end
    case 'NEXT'
        switch mele
            case 1
                Zc1(1)=param.pkg_Z_c(1);% tx impedance
                Zc1(2)=param.pkg_Z_c(2);% rx
                Zp1(1)=param.Pkg_len_NEXT;% tx mm (length)
                Zp1(2)=param.Pkg_len_RX;% rx
            case 4
                Zc1=[ param.pkg_Z_c(1,1) param.pkg_Z_c(2,1)] ;% Tx Rx impedance
                Zp1=[ param.Pkg_len_NEXT(1) param.Pkg_len_RX(1)] ;% Tx Rx length
                Zc2 =[ param.pkg_Z_c(1,2) param.pkg_Z_c(2,2)] ;
                Zp2 = [ param.Pkg_len_NEXT(2) param.Pkg_len_RX(2) ] ;
                Zc3 =[ param.pkg_Z_c(1,3) param.pkg_Z_c(2,3)] ;
                Zp3 = [ param.Pkg_len_NEXT(3) param.Pkg_len_RX(3) ] ;
                Zc4 = [ param.pkg_Z_c(1,4) param.pkg_Z_c(2,4)] ;
                Zp4 = [ param.Pkg_len_NEXT(4) param.Pkg_len_RX(4) ] ;
                Cv = [ param.C_v(1) param.C_v(2)  ];
            otherwise
                error('package symtax error')
        end
    case 'FEXT'
        switch mele
            case 1
                Zc1(1)=param.pkg_Z_c(1);% tx impedance
                Zc1(2)=param.pkg_Z_c(2);% rx
                Zp1(1)=param.Pkg_len_FEXT;% tx mm (length)
                Zp1(2)=param.Pkg_len_RX;% rx
            case 4
                Zc1=[ param.pkg_Z_c(1,1) param.pkg_Z_c(2,1)] ;% Tx Rx impedance
                Zp1=[ param.Pkg_len_FEXT(1) param.Pkg_len_RX(1)] ;% Tx Rx length
                Zc2 =[ param.pkg_Z_c(1,2) param.pkg_Z_c(2,2)] ;
                Zp2 = [ param.Pkg_len_FEXT(2) param.Pkg_len_RX(2) ] ;
                Zc3 =[ param.pkg_Z_c(1,3) param.pkg_Z_c(2,3)] ;
                Zp3 = [ param.Pkg_len_FEXT(3) param.Pkg_len_RX(3) ] ;
                Zc4 = [ param.pkg_Z_c(1,4) param.pkg_Z_c(2,4)] ;
                Zp4 = [ param.Pkg_len_FEXT(4) param.Pkg_len_RX(4) ] ;
                Cv = [ param.C_v(1) param.C_v(2)  ];
            otherwise
                error('package symtax error')
        end
end
% tx package
%% change tx package to CC mode
param.Z0=param.Z0/2;
switch mele
    case 1
        % [ahealey] Add series L shunt C to the on-die termination model.
        % [ s11out, s12out, s21out, s22out ]= make_pkg(faxis, Zp1(1), C_diepad(1), C_pkg_board(1),Zc1(1), param);*2);
        [ s11out, s12out, s21out, s22out ]= make_pkg(faxis, Zp1(1), C_diepad(1)*2, C_pkg_board(1),Zc1(1)*2, param, L_comp(1)/2, C_bump(1)*2 );
    case 4
        % [ s11outa, s12outa, s21outa, s22outa ]= make_pkg(faxis, Zp1(1),  C_diepad(1),0 ,Zc1(1), param);
        [ s11outa, s12outa, s21outa, s22outa ]= make_pkg(faxis, Zp1(1),  C_diepad(1)*2,0 ,Zc1(1)*2, param, L_comp(1)/2, C_bump(1)*2);
        % [ahealey] End of modifications.
        %
        [ s11outb, s12outb, s21outb, s22outb ]= make_pkg(faxis, Zp2(1) , 0, 0 ,Zc2(1)*2, param); % add pth
        [ s11outc, s12outc, s21outc, s22outc ]  = combines4p(  s11outa, s12outa, s21outa, s22outa, s11outb, s12outb, s21outb, s22outb    );
        %
        %
        [ s11outd, s12outd, s21outd, s22outd ]= make_pkg(faxis, Zp3(1) , 0, Cv(1)*2 ,Zc3(1)*2, param); % add pth
        [ s11oute, s12oute, s21oute, s22oute ]  = combines4p(  s11outc, s12outc, s21outc, s22outc, s11outd, s12outd, s21outd, s22outd    );
        %
        [ s11outf, s12outf, s21outf, s22outf ]= make_pkg(faxis, Zp4(1) , 0, C_pkg_board(1)*2, Zc4(1)*2, param); % add pth
        [ s11out, s12out, s21out, s22out ]  = combines4p(  s11oute, s12oute, s21oute, s22oute, s11outf, s12outf, s21outf, s22outf    );
end
param.Z0=param.Z0*2;% change tx reference back for DD mode
% RX package
switch mele
    case 1
        % [ahealey] Add series L shunt C to the on-die termination model.
        % [ s11in, s12in, s21in, s22in ]= make_pkg(faxis,  Zp1(2), C_diepad(2), C_pkg_board(2), Zc1(2), param);
        [ s11in, s12in, s21in, s22in ]= make_pkg(faxis,  Zp1(2), C_diepad(2), C_pkg_board(2), Zc1(2), param, L_comp(2), C_bump(2));
    case 4
        % [ s11ina, s12ina, s21ina, s22ina ]= make_pkg(faxis, Zp1(2), C_diepad(2),0 ,Zc1(2), param));
        [ s11ina, s12ina, s21ina, s22ina ]= make_pkg(faxis, Zp1(2), C_diepad(2),0 ,Zc1(2), param, L_comp(2), C_bump(2));
        % [ahealey] End of modifications.
        %
        [ s11inb, s12inb, s21inb, s22inb ]= make_pkg(faxis, Zp2(2) , 0, 0 ,Zc2(2), param); % add pth
        [ s11inc, s12inc, s21inc, s22inc ]  = combines4p(  s11ina, s12ina, s21ina, s22ina, s11inb, s12inb, s21inb, s22inb    );
        %
        [ s11ind, s12ind, s21ind, s22ind ]= make_pkg(faxis, Zp3(2) , 0, Cv(2) ,Zc3(2), param); % add pth
        [ s11ine, s12ine, s21ine, s22ine ]  = combines4p(  s11inc, s12inc, s21inc, s22inc, s11ind, s12ind, s21ind, s22ind    );
        %
        [ s11inf, s12inf, s21inf, s22inf ]= make_pkg(faxis, Zp4(2) , 0, C_pkg_board(2), Zc4(2), param); % add pth
        [ s11in, s21in, s12in, s22in ]  = combines4p(  s11ine, s12ine, s21ine, s22ine, s11inf, s12inf, s21inf, s22inf    );
end


%     p(1 ,1, :)=s11in;
%     p(2 ,2, :)=s22in;
%     p(1 ,2, :)=s12in;
%     p(2 ,1, :)=s21in;
%
%     S=sparameters(p,faxis);
%     rfwrite(S,'temp.s4p');

if OP.IDEAL_TX_TERM || (OP.RX_CALIBRATION == 1 && channel_number == 2) || OP.include_pcb == 2
    gamma_tx=0;
else
    gamma_tx=(R_diepad(param.Tx_rd_sel)/2-Z0/2)/(R_diepad(param.Tx_rd_sel)/2+Z0/2);% equation 93A-17
end
if OP.IDEAL_RX_TERM
    gamma_rx=0;
else
    gamma_rx=(R_diepad(param.Rx_rd_sel)-Z0)/(R_diepad(param.Rx_rd_sel)+Z0);% equation 93A-17
end

if OP.INC_PACKAGE==0
    sdc21p= sdc21;
    warning('do not use INC_PACKAGE = 0. Instead use package parameters)');
else
    if OP.RX_CALIBRATION == 1 && channel_number == 2
        %   for calibration do not include the transmitter package
        [s11out_rx, s12out_rx, s21out_rx, s22out_rx ] = combines4p( sdc11, sdc12, sdc21, sdc22, s22out, s12out, s21out, s11out ); %#ok<ASGLU> % s22 is ball side of package
        SCHDC.Frequencies=faxis;
        SCHDC.Parameters(1,1,:)=s11out_rx;
        SCHDC.Parameters(2,2,:)=s22out_rx;
        SCHDC.Parameters(1,2,:)=s12out_rx;
        SCHDC.Parameters(2,1,:)=s21out_rx;
        SCHDC.NumPorts=2;
        SCHDC.Impedance=100;
        %% Equation 93A-18
        sdc21p= s21out_rx.*(1-gamma_tx).*(1+gamma_rx)./(1.- s11out_rx.*gamma_tx - s22out_rx.*gamma_rx  -s21out_rx.^2.*gamma_tx.*gamma_rx +s11out_rx.*s22out_rx.*gamma_tx.*gamma_rx);
    else
        %% Equations 93A-4 to 93A-7
        if ~OP.IDEAL_TX_TERM
            [sdc11, sdc12, sdc21, sdc22] = combines4p( s11out, s12out, s21out, s22out, sdc11, sdc12, sdc21, sdc22 ); %#ok<ASGLU>
        end
        H_t=ones(1,length(faxis)); % .3bj compatibility
        if OP.IDEAL_TX_TERM ||  OP.T_r_filter_type == 1
            % for RITT testing with good termination as in some instruments
            % and tx filter when required
            if OP.T_r_filter_type==0
                H_t = exp(-(pi*faxis/1e9*OP.transmitter_transition_time/1.6832).^2); %% Equation 93A-46 %%
            else
                tr=OP.transmitter_transition_time;
                f9=faxis/1e9;
                if OP.T_r_meas_point == 1
                    k=1.9466+7.12*sqrt(1-6.51e-3/tr);
                    H_t=105./(f9.^4*(k*tr)^4 - f9.^3*(k*tr)^3*10i - 45*f9.^2*(k*tr)^2 + f9*(k*tr)*105i + 105);
                else
                    H_t = exp( -2*(pi*f9*tr/1.6832).^2 ).*exp(-1j*2*pi*f9*tr*3);
                end
                
            end
        end
        %         H_t=ones(1,length(faxis)); % not sure if we need a H_t or not. Is the CM noise correlated to the edge rate?
        if ~OP.IDEAL_RX_TERM
            [sdc11, sdc12, sdc21, sdc22] = combines4p( sdc11, sdc12, sdc21, sdc22, s22in, s12in, s21in, s11in ); %#ok<ASGLU> % s22 is ball side of package
        else
            warning('do not use IDEAL_RX_TERM. Instead hard code package and TR parameters')
        end
        %% Equation 93A-18 and part of 93A-1: Ht fix in V290 identified by Bill Kirkland and Ed Frlan
        sdc21p= H_t.*sdc21.*(1-gamma_tx).*(1+gamma_rx)./(1.- sdc11.*gamma_tx - sdc22.*gamma_rx  -sdc21.^2.*gamma_tx.*gamma_rx +sdc11.*sdc22.*gamma_tx.*gamma_rx);
    end
    
    
    %     % for debug
    %     p(1 ,1, :)=sdc11;
    %     p(2 ,2, :)=sdc22;
    %     p(1 ,2, :)=sdc12;
    %     p(2 ,1, :)=sdc21;
    % compute AC_CM_RMS at tp0
    OP.TX_BesselThomson=1; %AC CM is measured in scopes with at BT filter
    if OP.TX_BesselThomson
        factor_3db=0.473037;
        param.fb_BT_cutoff=factor_3db*param.f_r;
        a = bessel( param.BTorder );
        acoef=fliplr( a );
        H_bt =a(1)./ polyval(acoef, (1i*faxis./(param.fb_BT_cutoff*param.fb)));
    else
        H_bt=ones(1,length(f));
    end
    if channel_number == 1
        f_int= faxis( faxis<=param.ACCM_MAX_Freq );
        H_cc= s21out.*(1-gamma_tx)./(1.- s11out.*gamma_tx ).*H_bt;
        param.sigma_ACCM_at_tp0= sqrt(2*param.AC_CM_RMS_TX^2*sum( abs( H_cc(2:length(f_int)) ).^2 .* diff(f_int))/f_int(end)) ;
        %     S=sparameters(p,faxis);
        %     rfwrite(S,'temp.s4p');
    end
    SCHDC.Frequencies=faxis;
    SCHDC.Parameters(1,1,:)=sdc11;
    SCHDC.Parameters(2,2,:)=sdc22;
    SCHDC.Parameters(1,2,:)=sdc12;
    SCHDC.Parameters(2,1,:)=sdc21;
    SCHDC.NumPorts=2;
    SCHDC.Impedance=25;
    
end





function t_params = stot(s_params)
% p 67 R. Mavaddat. (1996). Network scattering parameter. Singapore: World Scientific.
% ISBN 978-981-02-2305-2. http://books.google.com/?id=287g2NkRYxUC&lpg=PA65&dq=T-parameters+&pg=PA67.
[s11, s12, s21, s22] = deal(s_params(1,1,:), s_params(1,2,:), s_params(2,1,:), s_params(2,2,:));
delta = (s11.*s22-s12.*s21);
s21(s21==0)=eps;
t_params = [1./s21, -s22./s21; s11./s21, -delta./s21];

function csv_string = str2csv(c)
% convert a cell array of strings to a csv string
cell_tmp = cell(2, length(c));
cell_tmp(1,:)=c;
cell_tmp(2,:) = {','};
cell_tmp{2,end} = '';
csv_string=strcat(cell_tmp{:});

function [s11, s12, s21, s22] = synth_tline(f, Z_c, Z_0, gamma_coeff, tau, d)
f_GHz=f/1e9;
%% Equation 93A-10 %%
gamma_1 = gamma_coeff(2)*(1+1i);
%% Equation 93A-11 %%
gamma_2 = gamma_coeff(3)*(1-2i/pi*log(f_GHz)) + 2i*pi*tau;
%% Equation 93A-9 %%
gamma = gamma_coeff(1)+gamma_1.*sqrt(f_GHz)+gamma_2.*f_GHz;
gamma(f_GHz==0) = gamma_coeff(1);

%% Equation 93A-12 %%
rho_rl=(Z_c-2*Z_0)/(Z_c+2*Z_0);

exp_gamma_d = exp(-d*gamma);
%% Equations 93A-13 and 93A-14 %%
s11 = rho_rl*(1-exp_gamma_d.^2)./(1-rho_rl^2*exp_gamma_d.^2);
s21 = (1-rho_rl^2)*exp_gamma_d./(1-rho_rl^2*exp_gamma_d.^2);
s12 = s21;
s22 = s11;

function s_params = ttos(t_params)
% p 67 R. Mavaddat. (1996). Network scattering parameter. Singapore: World Scientific.
% ISBN 978-981-02-2305-2. http://books.google.com/?id=287g2NkRYxUC&lpg=PA65&dq=T-parameters+&pg=PA67.
[t11, t12, t21, t22] = deal(t_params(1,1,:), t_params(1,2,:), t_params(2,1,:), t_params(2,2,:));
delta = t11.*t22-t21.*t12;
t11(t11==0)=eps;
s_params = [t21./t11, delta./t11; 1./t11, -t12./t11];

function [out_var,varg_out]=varargin_extractor(varargin)

if isempty(varargin)
    out_var=[];
    varg_out={};
else
    out_var=varargin{1};
    varg_out=varargin;
    varg_out(1)=[];
end


function results= vma(PR, M)
% PR=sbr.Data;
% M=32;
% PR is the pulse response
% M is samples per UI
[ seq, syms, syms_nrz ] = PRBS13Q( );
% seq uses [ -1 -1/3 1/3 1] & syms uses [ 0 1 2 3 ]
symbols=seq;
imaxPR=find(PR==max(PR),1,'first'); % find index for peak
% start end symbols index for 7 3's and 6 0's
indx_S3x7_start=M*(strfind(syms,ones(1,7)*3)-1)+imaxPR;
indx_S3x7_end=M*(strfind(syms,ones(1,7)*3)+5)+imaxPR;
indx_S0x6_start=M*(strfind(syms,ones(1,6)*0))-1+imaxPR;
indx_S0x6_end=M*(strfind(syms,ones(1,6)*0)+4)+imaxPR;
% superposition code
shifting_vector=kron(symbols,[ 1  zeros(1,M-1) ]) ;
Bit_stream_response=filter(PR,1, shifting_vector);
% find center of 3's and 0's
icent3=floor((indx_S3x7_end-indx_S3x7_start)/2 + indx_S3x7_start);
icent0=floor((indx_S0x6_end-indx_S0x6_start)/2 + indx_S0x6_start);
% plot(Bit_stream_response(indx_S3x7_start:indx_S3x7_end))
% hold on
% plot(Bit_stream_response(indx_S0x6_start:indx_S0x6_end))
P_3 = mean( Bit_stream_response((icent3-M):(icent3+M) ) );
P_0 = mean( Bit_stream_response((icent0-M):(icent0+M) ) );
VMA= P_3 - P_0;
results.P_3=P_3;
results.P_0=P_0;
results.VMA=VMA;
function line_intersection=vref_intersect(eye_contour,x_in,vref)

%slope of the 2 sample points around vref crossing
m1=(eye_contour(x_in,1)-eye_contour(x_in-1,1));
%x-intercept for the line
b1=eye_contour(x_in,1)-m1*x_in;
% drawing a horizontal line through vref so slope = 0
m2=0;
%special case for horizontal line, b=y
b2=vref;
%the x-value of line intersection = (b2-b1)/(m1-m2)
%sinc m2 is always 0 and b2 is always vref, this could be stated as (vref-b1)/m1
%And usually vref is 0, so it further reduces to -b1/m1
line_intersection=(b2-b1)/(m1-m2);





function p=xls_parameter(param_sheet, param_name, eval_if_string, default_value)
% helper function to read parameter values from XLS file. Uses names to find values.
if nargin<3, eval_if_string=0; end
[row, col]=find(strcmpi(param_sheet, param_name)); % RIM 08-26-2020 make case insensitive
if numel(row)*numel(col)==0
    if nargin<4
        missingParameter(param_name);
    else
        p = default_value;
    end
elseif numel(row)*numel(col)>1
    % if there are several occurrences, use the first, but warn
    %     warning('COM:XLS_parameter:MultipleOccurrence', ...
    %         '%d occurrences of "%s" found. Using the first', numel(row), param_name);
    error('COM:XLS_parameter:MultipleOccurrence', ...
        '%d occurrences of "%s" found. Please recheck spreadsheet', numel(row), param_name);% RIM 01-0 8-20
    p=param_sheet{row(1), col(1)+1};
else
    p=param_sheet{row, col+1};
end
if ischar(p) && eval_if_string
    p=eval(p);
end

function zzz_list_of_changes
% structures:
% chdata(i)
%           i= 1 --> THRU index
%           i= 2, num_fext+1 --> FEXT channel index
%           i= num_fext+2, num_next+num_fext+1
%       base: name of THRU file
%          A: amplitude
%       type: 'THRU', 'NEXT', or 'FEXT'
%        ftr: Rise time frequency
%      fmaxi: max number of frequency points
%      faxis: frequency array [Hz]
%      sdd21: (Htot) rewritten as product of ,vtf based on pkg RL ,tx filter, Rx filter
%      sdd22: differential RL
%      sdd11: differential RL
%     sdd21p: vtf based on pkg RL , this an interim parameter and set to sdd21
%     sdd21f: raw differential IL not filtered use for FD plots
% added output_args.peak_uneq_pulse_mV
% added output_args.cable_loss when "Include PCB" is not 0 in the config file
% added: tap c(-2) c(2) and c(3)
% added: g_DC_HP and f_HP_PZ
% added: new value for "Include PCB" = 2 for cable Rx compliance test only Rx host added
% added: BREAD_CRUMBS is 1 then a mat file with the structures params and OP is created in the results directory
% added: T_r_filter_type for RITT testing when IDEAL_TX_TERM is 1:
% added T_r_meas_point for RITT, if 0, measurement was at tp0, if 1 measurement was tp0a
% 0 is for is for Gaussian filter and 1 is for a 4th order Bessel-Thomson filter
% fixed INCLUDE_CTLE=0 to really remove from computation
% r161a fixed matlab version issue when for OP.INCLUDE_CTLE=0
% r162 adjusting RITT rise time to Mike Dudek's recommendations also always enable risetime filter if T_r_filter_type=1
% r162 tx and rx package impedance {Zc)
% r162a Gaussian equation corrected
% r163 cast snr_tx with package test case
% r164 fix pdf for very low noise and lo pass filter enhancements
% r164 add zero gain at nqyist CTLE as in CL12e
% r165 add simpler congfig command called FORCE_TR (force risetime)
% r200 cm3 and cm4 added cp3 removed
% r200 fixed problem in s21_to_impulse_DC when s parameter have a DC entry
% r200 ILD_FOM updated to EQ93A-55  ERL adde
% r200 improved phase interpolation for return loss time conversion
% r200a db = @(x) 20*log10(abs(x)) added so sig processing toolbox won't be required
% r200b Fixed error in bifurcation of Tx/Rx Rd%
% r200c missed on fix for interpolation
% r210 new ERL with time gating function
% r224 update ERL with from D3.1
% r226  fix s2p reading problem
%    change SNR_ISI_XTK_normalized_1_sigma to SNR_ISI (with nulled noise)
%    Fix Rx calibration issue
%    added ERL limit and Nd
% r227 adding Pmax/Vf and peak of eq pulse, fixed issue with rx testing
%     INC_PACKAGE=0 not fully supported message
%     if N=0 use TDR_duration
%     red display text for fail ERL and COM
% r228 fixed ERL pass fail report, default Grr_limit to 1
% r230 add rx ffe
% r231 change crosstalk noise to icn like to speed things up
% r231 change default OP.impulse_response_truncation_threshold to 1e-3 from
% 1e-5mof-
% r232 fix default for Rx eq so old spead sheets work
% r234 fix inadvertent typo for clause 120e ctle and problem with TXffe loop time reduction
% r235 adding dfe quantization changed to normalized DFE taps reported
% r236 adding ffe gain loop and resample after RxFFE
% r240 added output for C2M and setting defaults for some FFE eq
% r241 force FFE main cursor to 1 and remove sum of taps = 1
% r250 adding more complex package
% r251 post cursor fix for DFE in force() and ffe backoff
% r251 remove TDR threshold noise filter
% r252 add rx FFE filter to receiver noise filter
% r252 change ICN in the xtk noise calculation to end at fb rather than fb/2
% r253 a few bug fixes in force from i indexing and for no ffe postcursors
% r254 precursor check fix in optimize_fom % mod fix in force
% r254 help to align columns in csv file
% r254 accept syntax for 2 tline flex package model
% r256 speed up optimize FOM
% r256 fix problem reading in config file from q/a
% r256 added code from Yasou Hidaka for reading in parameter an and printing out noise
% r257 fixed extrapolation of channel with lower bandwidths in s21_to_impulse_DC
% r257 in get_xtlk_noise in optimize_FOM: reomove crosstalk double counting and apply TXFFE is  FEXT
% r258 EXE_MODE switch 12/21 0:legacy 1:fast 2:superfast
% r258 CDR switch 'MM' or 'mod-MM'
% r258 correction for asymentirc tx/Rx packages
% r258 revamped display results display window
% r259 fix problem if Min_VEO is set in spreadsheet.
% r259 fix problem in optimize_FOM. get_xtlk_noise need to have 3 output
%      parameter else only FEXT is considered for FOM.zhilei huang 01/11/2019
% r259 putting COM_db and IL last in output to terminal
% r259 msgtext change to msg for C2C case other cases not vetted but not presently used
% r259 use N_bx for ERL rather than Nb (ndfe))
% r259 added TDR_W_TXPKG which performs TDR and ERL with the Tx package added
% r260 r259 used rd for the reciever to terminate the package. It was changed to the rd of the transmitter
% r260 used eta_0 PSD equation for sigma_n
% r260 fix IL graph legend to w/pkg and Tr
% r260 define tfx for each port
% r262 fx parameter passing parsing for mod_string revert to 2.57 no COM computational impact
% r262 Report estimate for DER for channel (Yasuo 2/30/19)
% r262 reset on exit default text interpreter to tex
% r262 localize run timer (John Buck 1/17/19)
% r262 set db as internal function in force to avoid tool box
% r262 changed loop for Grr and Gloss in get_tdr so that nbx and tfx works when beta x = 0
% r263 added to output_args RL structure and report "struct" in csv file
% r264 added EW estimate
% r266 using unequalized IR for Vf and Vf to compute ratio of Vp/Vf
% r267 added floating taps with param.N_bf, param.N_bg, param.N_bmax, param.bmaxg.  groups not used for ERL
% r268 added sequential/co-optimization switch for floating tap banks OP.FT_COOP default 0 i.e. sequential
% r269 changed  param.N_bmax to  param.N_f
% r270 implement JingBo Li's and Howard Heck's floating tap method
% r270 modification by Adam Healey for Ls and Cb termination (aka t-coil emulation)
% r270 added c_0 and c_1 for CA in add_brd
% r272 fixed version syntax problem in output_args RL report
% r272 fixed eye width computation problem crosstalk was missed in pervious versions
% r272 removed eye width report if doing a Rx calibration
% r273 better alignment and control for ICN reporting
% r273 fixed PSXTK graph
% r275 fixed delay adjustment for ERL/TDR in get_TDR (Adam Healey 09/06/2019)
% r276 go back to reporting channel IL results (output_args.IL_dB_channel_only_at_Fnq) with board added read_s4p_files (as in r270)
% r276 chdata(i).Aicn=param.a_icn_fext should have been chdata(i).Aicn=param.a_icn_next for the next selection. Since in most spec's they are the same there is little no impact in results
% r276 test for output_args for isfield(chdata(1),'sdd22_raw')
% r276 change divisor for ICN and FOM_ILD to param.f2 from param.fb, may raise ICN and ILD value reported in r275
% r276 C_1 was instantiated as C_0. This was fixed
% r276 fixed rounding problem in reporting of loss at f_nq
% r276 power limit (RSS) for tail DFE taps (B_float_RSS_MAX, N_tail_start)
% r277 added nv for deterining steady state voltage for fitting compatibility
% r278 added b_min to support asymmetric bmax
% r278 added kappa1 and kappa2 to scale package to channel reflection for ERL experiments
% r278 added keyword OP.SHOW_BRD which includes added board in TDR and ERL
% r292 speed up for FOM search (Adee Ran) implemented by Adam Gregory.
% r292 param.LOCAL_SEARCH set to is the heuristic step distance keyword is 'Local Search'
% r292 fixing TDR for different impedance references in get_TDR and s2p file compatibility
% r292 eq. 93A-19 and 93-20 code implementation  bug when include .3by change% to fix edge rate equation 93A-46 (h_T). no effect if Rd=50 or IL > 5 dB
% r292 H_t implemented  in s21_pkg
% r292 plot and report for die to die IL remove the Tr effect "IL with pkgs & Tr filter" goes to "IL with pkgs"
% r292 add GDC_MIN to optimize_FOM
% r293 fix if ndfe-0 and ERL only and s2p issue
% r293a investigate the Tukey filtering
% r293a if fix if bmin is missing
% r294 fix problems reading s2p files for ERL computation
% r294 align Tukey_Window with .3ck definition for ERL and TDR computations
% r294 add parameter param.Noise_Crest_Factor. Default is not to use
% r294 add gdc and gdc2 range limitations
% r295 add VEC Pass threshold
% r295 removed close force all. Tagged all figures with "COM"
% r295 consolidated print in new function "end_display_control"
% r295 report pre/pmax for Txffe
% r295 speed up test cases by not re-reading in s4p files
% r297 add  provisions for AC_CM_RMS for through CM (experimental)
% r299 add keyword T_O (param.T_O) and  samples_for_C2M (parsm.samples_for_C2M) for new C2M VEC and EH computations
% r310 refine VEC and EH for C2M from Adam Gregory in
%      cdf_to_ber_contour,COM_eye_width,combine_pdf_same_voltage_axis,
%      optimize_fom_for_C2M. pdf_to_cdf, conv_fct_MeanNotZero, and get_pdf_full
% r311 added RILN
% r314 when T_O is not zero 3 eyes are used to compute VEC and VEO
% r315 added keyword for Bessel_Thomson and Butterworth(default) filter.
% r315 Bessel_Thomson keyword is added mostly for measuring  Pmax, Vf, and SNDR
% r316 remove DC computation for RX Calibration loops
% r317 for SAVE_TD to include EQ and unEQ FIR
% r317 clean up bessel thomson and butterworth filter logic for ERL and normal COM
% r318 if min_VEO_test fails to find a solution the loop is restarted with min_VEO_test to near zero. Makes sure COM returns results
% r320 fixed RX_CALIBRATION which was broken in r310
% r320 speed up for C2M by moving managing optimize loop distribution of computations
% r320 for C2M added Gaussian window keyword, Gaussian_histogram_window, for T_O and keyword, QL which is  at Q limit at +/-T_O
% r320 removed external feature and replace with TDMODE
% r320 added TDMODE which allows for the use of pulse resonance files (CSV) instead of s4p files
% r330 added tfx to N for ERL

